#pragma once

#include "core/core.h"
#include "window_manager/wmanager.h"
#include "filesystem/v3nus_fs.h"

typedef int sample_array_t;

#define JPGE_OUT_BUF_SIZE 2048

// JPEG chroma subsampling factors. Y_ONLY (grayscale images) and H2V2 (color images) are the most common.
enum subsampling_t 
{ 
    Y_ONLY = 0, 
    H1V1, 
    H2V1, 
    H2V2 
};

// JPEG compression parameters structure.
struct je_params
{
    // Quality: 1-100, higher is better. Typical values are around 50-95.
    int m_quality;

    // m_subsampling:
    // 0 = Y (grayscale) only
    // 1 = YCbCr, no subsampling (H1V1, YCbCr 1x1x1, 3 blocks per MCU)
    // 2 = YCbCr, H2V1 subsampling (YCbCr 2x1x1, 4 blocks per MCU)
    // 3 = YCbCr, H2V2 subsampling (YCbCr 4x1x1, 6 blocks per MCU-- very common)
    subsampling_t m_subsampling;

    // Disables CbCr discrimination - only intended for testing.
    // If true, the Y quantization table is also used for the CbCr channels.
    bool m_no_chroma_discrim_flag;

    bool m_two_pass_flag;
};

void init_je_params( je_params* jp );
bool check_je_params( je_params* jp );

struct jpeg_encoder
{
    je_params m_params;
    uchar m_num_components;
    uchar m_comp_h_samp[ 3 ], m_comp_v_samp[ 3 ];
    int m_image_x, m_image_y, m_image_bpp, m_image_bpl;
    int m_image_x_mcu, m_image_y_mcu;
    int m_image_bpl_xlt, m_image_bpl_mcu;
    int m_mcus_per_row;
    int m_mcu_x, m_mcu_y;
    uchar* m_mcu_lines[ 16 ];
    uchar m_mcu_y_ofs;
    sample_array_t m_sample_array[ 64 ];
    int16 m_coefficient_array[ 64 ];
    int m_quantization_tables[ 2 ][ 64 ];
    uint m_huff_codes[ 4 ][ 256 ];
    uchar m_huff_code_sizes[ 4 ][ 256 ];
    uchar m_huff_bits[ 4 ][ 17 ];
    uchar m_huff_val[ 4 ][ 256 ];
    uint m_huff_count[ 4 ][ 256 ];
    int m_last_dc_val[ 3 ];
    uchar m_out_buf[ JPGE_OUT_BUF_SIZE ];
    uchar* m_pOut_buf;
    uint m_out_buf_left;
    uint m_bit_buffer;
    uint m_bits_in;
    uchar m_pass_num;
    bool m_all_stream_writes_succeeded;  
    V3_FILE f;
};

// Initializes the compressor.
void je_init( jpeg_encoder* je );

// Set compression parameters.
// f: The stream object to use for writing compressed data.
// params - Compression parameters structure, defined above.
// width, height  - Image dimensions.
// channels - May be 1, or 3. 1 indicates grayscale, 3 indicates RGB source data.
// Returns false on out of memory or if a stream write fails.
bool je_set_params( V3_FILE f, int width, int height, int src_channels, je_params* comp_params, jpeg_encoder* je );

// Deinitializes the compressor, freeing any allocated memory. May be called at any time.
void je_deinit( jpeg_encoder* je );

// Call this method with each source scanline.
// width * src_channels bytes per scanline is expected (RGB or Y format).
// You must call with NULL after all scanlines are processed to finish compression.
// Returns false on out of memory or if a stream write fails.
bool je_process_scanline( const void* pScanline, jpeg_encoder* je );

void je_optimize_huffman_table( int table_num, int table_len, jpeg_encoder* je );
void je_emit_byte( uchar i, jpeg_encoder* je );
void je_emit_word( uint i, jpeg_encoder* je );
void je_emit_marker( int marker, jpeg_encoder* je );
void je_emit_jfif_app0( jpeg_encoder* je );
void je_emit_dqt( jpeg_encoder* je );
void je_emit_sof( jpeg_encoder* je );
void je_emit_dht( uchar* bits, uchar* val, int index, bool ac_flag, jpeg_encoder* je );
void je_emit_dhts( jpeg_encoder* je );
void je_emit_sos( jpeg_encoder* je );
void je_emit_markers( jpeg_encoder* je );
void je_compute_huffman_table( uint* codes, uchar* code_sizes, uchar* bits, uchar* val, jpeg_encoder* je );
void je_compute_quant_table( int* dst, int16* src, jpeg_encoder* je );
void je_adjust_quant_table( int* dst, int* src, jpeg_encoder* je );
void je_first_pass_init( jpeg_encoder* je );
bool je_second_pass_init( jpeg_encoder* je );
bool je_jpg_open( int p_x_res, int p_y_res, int src_channels, jpeg_encoder* je );
void je_load_block_8_8_grey( int x, jpeg_encoder* je );
void je_load_block_8_8( int x, int y, int c, jpeg_encoder* je );
void je_load_block_16_8( int x, int c, jpeg_encoder* je );
void je_load_block_16_8_8( int x, int c, jpeg_encoder* je );
void je_load_quantized_coefficients( int component_num, jpeg_encoder* je );
void je_flush_output_buffer( jpeg_encoder* je );
void je_put_bits( uint bits, uint len, jpeg_encoder* je );
void je_code_coefficients_pass_one( int component_num, jpeg_encoder* je );
void je_code_coefficients_pass_two( int component_num, jpeg_encoder* je );
void je_code_block( int component_num, jpeg_encoder* je );
void je_process_mcu_row( jpeg_encoder* je );
bool je_terminate_pass_one( jpeg_encoder* je );
bool je_terminate_pass_two( jpeg_encoder* je );
bool je_process_end_of_image( jpeg_encoder* je );
void je_load_mcu( const void* src, jpeg_encoder* je );
void je_clear( jpeg_encoder* je );

bool compress_image_to_jpeg_file( const char* pFilename, int width, int height, int num_channels, const uchar* pImage_data, je_params* comp_params );
