/*
    debug.cpp. Debug (log) functions (thread safe)
    This file is part of the SunDog engine.
    Copyright (C) 2002 - 2013 Alexander Zolotov <nightradio@gmail.com>
    www.warmplace.ru
*/

#include "core/core.h"
#include "memory/memory.h"
#include "utils/utils.h"
#include "filesystem/v3nus_fs.h"
#include "main/user_code.h"
#include "../debug.h"

#include <stdarg.h>

#ifndef PALMOS
    #include <stdlib.h>
#endif

#ifdef ANDROID
    #include <android/log.h>
#endif

#define Y_LIMIT 50

sundog_mutex debug_mutex;
UTF8_CHAR* debug_output_file = 0;
int hide_debug_counter = 0;
UTF8_CHAR debug_buf[ 2048 ];

#ifdef PALMOS
int debug_count = 0;
int debug_y = 10;
#endif

void debug_init( void )
{
    debug_output_file = 0;
    hide_debug_counter = 0;
#ifdef PALMOS
    debug_count = 0;
    debug_y = 10;
#endif
    sundog_mutex_init( &debug_mutex, 0 );
#ifndef PALMOS
    debug_output_file = (UTF8_CHAR*)malloc( 16 );
    sprintf( debug_output_file, "log.txt" );
#endif
}

void debug_deinit( void )
{
    sundog_mutex_destroy( &debug_mutex );
#ifndef PALMOS
    free( debug_output_file );
    debug_output_file = 0;
#endif
}

void hide_debug( void )
{
    if( sundog_mutex_lock( &debug_mutex ) ) return;
    hide_debug_counter++;
    sundog_mutex_unlock( &debug_mutex );
}

void show_debug( void )
{
    if( sundog_mutex_lock( &debug_mutex ) ) return;
    if( hide_debug_counter > 0 )
	hide_debug_counter--;
    sundog_mutex_unlock( &debug_mutex );
}

void debug_set_output_file( const UTF8_CHAR* filename )
{
#ifdef PALMOS
    return;
#endif
    if( sundog_mutex_lock( &debug_mutex ) ) return;
    free( debug_output_file );
    debug_output_file = 0;
    UTF8_CHAR* name = v3_make_filename( filename );
    if( name == 0 )
    {
	name = (char*)filename;
	debug_output_file = (UTF8_CHAR*)malloc( mem_strlen( name ) + 1 );
	debug_output_file[ 0 ] = 0;
	mem_strcat( debug_output_file, name );
    }
    else 
    {
	debug_output_file = (UTF8_CHAR*)malloc( mem_strlen( name ) + 1 );
	debug_output_file[ 0 ] = 0;
	mem_strcat( debug_output_file, name );
	mem_free( name );
    }
    sundog_mutex_unlock( &debug_mutex );
}

const UTF8_CHAR* debug_get_output_file( void )
{
    return debug_output_file;
}

void debug_reset( void )
{
#ifdef NODEBUG
    return;
#endif

    if( sundog_mutex_lock( &debug_mutex ) ) return;
    
    if( debug_output_file )
    {
#ifndef PALMOS
	v3_remove( debug_output_file );
#endif
    }
    
    sundog_mutex_unlock( &debug_mutex );
}

void dprint( const UTF8_CHAR* str, ... )
{
#ifdef NODEBUG
    return;
#endif

    if( hide_debug_counter ) return;

#ifndef PALMOS
    if( debug_output_file == 0 ) return;
#endif

    if( sundog_mutex_lock( &debug_mutex ) ) return;
    
    va_list p;
    va_start( p, str );

#ifdef PALMOS
    
    int ptr = 0;
    int ptr2 = 0;
    UTF8_CHAR num_str[ 16 ];
    int len;

    //Make a number:
    int_to_string( debug_count, num_str );
    len = mem_strlen( num_str );
    mem_copy( debug_buf, num_str, len );
    debug_buf[ len ] = ':';
    debug_buf[ len + 1 ] = ' ';
    ptr2 += len + 2;
    debug_count++;

    //Make a string:
    while( 1 )
    {
	if( str[ ptr ] == 0 ) break;
	if( str[ ptr ] == '%' )
	{
	    if( str[ ptr + 1 ] == 'd' || str[ ptr + 1 ] == 'x' )
	    {
		int arg = va_arg( p, int );
		if( str[ ptr + 1 ] == 'd' )
		    int_to_string( arg, num_str );
		else
		    hex_int_to_string( arg, num_str );
		len = mem_strlen( num_str );
		mem_copy( debug_buf + ptr2, num_str, len );
		ptr2 += len;
		ptr++;
	    }
	    else
	    if( str[ ptr + 1 ] == 's' )
	    {
		//ASCII string:
		UTF8_CHAR *arg2 = va_arg( p, UTF8_CHAR* );
		if( arg2 )
		{
		    len = mem_strlen( arg2 );
		    if( len )
		    {
			mem_copy( debug_buf + ptr2, arg2, len );
			ptr2 += len;
		    }
		}
		ptr++;
	    }
	}
	else
	{
	    debug_buf[ ptr2 ] = str[ ptr ];
	    ptr2++;
	}
	ptr++;
    }
    debug_buf[ ptr2 ] = 0;
    
#else
    
    vsprintf( debug_buf, str, p );
    
#endif

    va_end( p );

    //Save result:
#if !defined(PALMOS)
    FILE* f = fopen( debug_output_file, "ab" );
    if( f )
    {
	fwrite( debug_buf, 1, mem_strlen( debug_buf ), f );
	fclose( f );
    }
#endif
#ifdef PALMOS
    //PalmOS:
    int a;
    for( a = 0; a < 128; a++ )
    {
	if( debug_buf[ a ] == 0 ) break;
	if( debug_buf[ a ] == 0xA ) break;
	if( debug_buf[ a ] == 0xD ) break;
    }
    WinDrawChars( "                                                                                ", 80, 0, debug_y );
    WinDrawChars( debug_buf, a, 0, debug_y );
    debug_y += 10;
    if( debug_y >= Y_LIMIT ) debug_y = 0;
#else
    printf( "%s", debug_buf );
#ifdef ANDROID
    __android_log_print( ANDROID_LOG_INFO, "native-activity", debug_buf );
#endif
#endif
    
    sundog_mutex_unlock( &debug_mutex );
}

void debug_show_error_report( void )
{
#ifdef IPHONE
    char* log = 0;
    const UTF8_CHAR* log_file = debug_get_output_file();
    size_t log_size = 0;
    size_t log_msize = 0;
    FILE* f = fopen( log_file, "rb" );
    if( f )
    {
	while( 1 )
	{
	    int c = fgetc( f );
	    if( c < 0 )
	    {
		c = 0;
	    }
	    if( log == 0 )
	    {
		log = (char*)MEM_NEW2( 1024 );
		log_msize = 1024;
	    }
	    if( log_size >= log_msize )
	    {
		log_msize += 1024;
		log = (char*)mem_resize( log, log_msize );
	    }
	    log[ log_size ] = c;
	    log_size++;
	    if( c == 0 ) break;
	}
	fclose( f );
    }
    if( log )
    {
	UTF8_CHAR ts[ 256 ];
	sprintf( ts, "%s ERROR REPORT", user_window_name );
	send_mail( "nightradio@gmail.com", ts, log );
	mem_free( log );
    }
#endif
}
