#ifndef __MEMORY__
#define __MEMORY__

#include "../core/core.h"
#include "../core/debug.h"
#include "../filesystem/v3nus_fs.h"
#include "../utils/utils.h"

#ifndef PALMOS
    #include <stdlib.h>
    #include <string.h>
#endif
    
#if defined(WIN) || defined(WINCE)
    #include <windows.h>
#endif

enum 
{
    HEAP_DYNAMIC = 0,
    HEAP_STORAGE
};
#define HEAP_MASK		3

#define MEM_FLAG_NOCACHE	4

#ifndef MEM_FAST_MODE
    #ifndef PALMOS
	#define USE_NAMES
    #endif
#endif

#define MEM_MAX_NAME_SIZE 16

struct mem_block_struct 
{
    uint heap;
    size_t size;
#ifdef USE_NAMES
    UTF8_CHAR name[ MEM_MAX_NAME_SIZE ];
#endif
#ifndef MEM_FAST_MODE
    mem_block_struct* next;
    mem_block_struct* prev;
#endif
};

#define MEM_NEW( heap, size ) mem_new( heap, size, (UTF8_CHAR*)__FUNCTION__ )
#define MEM_NEW2( size ) mem_new( HEAP_DYNAMIC, size, (UTF8_CHAR*)__FUNCTION__ )

extern size_t g_mem_error;

//mem_new():
//heap - heap number:
//HEAP_DYNAMIC = dynamic heap for small blocks
//HEAP_STORAGE = storage heap for large static blocks
//size - block size
void mem_init();
void mem_deinit();
int mem_free_all();
void* mem_new( unsigned int heap, size_t size, const UTF8_CHAR* name ); //Each memory block has its own name
void simple_mem_free( void* ptr );
void mem_free( void* ptr );
inline void mem_set( void* ptr, size_t size, unsigned char value )
{
    if( ptr == 0 ) return;
#ifdef PALMOS
    MemSet( ptr, size, value );
#else
    memset( ptr, value, size );
#endif
}
void mem_zero( void* ptr );
void* mem_resize( void* ptr, size_t size );
inline void mem_copy( void* dest, const void* src, size_t size )
{
    if( dest == 0 || src == 0 ) return;
#ifdef PALMOS
    MemMove( dest, src, size ); //It's for dynamic heap only!!
#else
    memmove( dest, src, size );
#endif
}
inline int mem_cmp( const char* p1, const char* p2, size_t size )
{
    if( p1 == 0 || p2 == 0 ) return 0;
#ifdef PALMOS
    return MemCmp( p1, p2, size ); //It's for dynamic heap only!!
#else
    return memcmp( p1, p2, size );
#endif
}
inline void mem_strcat( char* dest, const char* src )
{
    if( dest == 0 || src == 0 ) return;
#ifdef PALMOS
    StrCat( dest, src );
#else
    strcat( dest, src );
#endif
}
int mem_strcat_s( char* dest, size_t dest_size, const char* src );
inline int mem_strcmp( const char* s1, const char* s2 )
{
    if( s1 == 0 || s2 == 0 ) return 0;
#ifdef PALMOS
    return StrCompare( s1, s2 );
#else
    return strcmp( s1, s2 );
#endif
}
inline const char* mem_strstr( const char* s1, const char* s2 )
{
    if( s1 == 0 || s2 == 0 ) return 0;
#ifdef PALMOS
    return StrStr( s1, s2 );
#else
    return strstr( s1, s2 );
#endif
}
size_t mem_strlen( const char* s );
size_t mem_strlen_utf16( const UTF16_CHAR* s );
size_t mem_strlen_utf32( const UTF32_CHAR* s );
char* mem_strdup( const char* s1 );

//Get info:
uint mem_get_heap( void* ptr );
uint mem_get_flags( void* ptr );
size_t mem_get_size( void* ptr );
UTF8_CHAR* mem_get_name( void* ptr );
size_t mem_get_usage( void );

//PalmOS specific:
void mem_on( void ); //Storage protection ON
void mem_off( void ); //Storage protection OFF
void mem_palm_normal_mode( void ); //Switch to normal mode (Storage protection ON)
void mem_palm_our_mode( void ); //Switch back to our mode (Storage protection is ON or OFF)

#ifdef PALMOS
#include <PalmOS.h>
extern SysAppInfoPtr ai1, ai2, appInfo;
extern unsigned short ownID;
#endif

#endif
