//Main sound header

#ifndef __SOUND__
#define __SOUND__

#include "../core/core.h"
#include "../time/timemanager.h"
#include "../utils/utils.h"

//Structures and defines:

#define SOUND_MODE_INT16		    	1
#define SOUND_MODE_FLOAT32		    	2

#define SOUND_SLOTS			    	4

#define SOUND_SCOPE_FLAG_WRITE_AND_LOCK	    	1

#define SOUND_STREAM_FLAG_USER_CONTROLLED	( 1 << 0 )
#define SOUND_STREAM_FLAG_ONE_THREAD		( 1 << 1 )

#if CPUMARK >= 10
#define SOUND_SCOPE_TYPE signed short
#define SOUND_SCOPE_MAX 32768
#else
#define SOUND_SCOPE_TYPE signed char
#define SOUND_SCOPE_MAX 128
#endif

#define FLOAT32_TO_INT16( res, val ) \
{ \
    float temp = val; \
    temp *= 32767; \
    if( temp > 32767 ) \
	temp = 32767; \
    if( temp < -32767 ) \
	temp = -32767; \
    res = (signed short)temp; \
}

#define FLOAT32_TO_INT32( res, val ) \
{ \
    float temp = val; \
    temp *= 0x80000000; \
    if( temp > 0x80000000 - 1 ) \
    temp = 0x80000000 - 1; \
    if( temp < -( 0x80000000 - 1 ) ) \
    temp = -( 0x80000000 - 1 ); \
    res = (signed int)temp; \
}

#define INT16_TO_FLOAT32( res, val ) { res = (float)(val) / 32768.0F; }

//MIDI flags:
#define MIDI_PORT_READ ( 1 << 0 )
#define MIDI_PORT_WRITE ( 1 << 1 )
#define MIDI_NO_DEVICE ( 1 << 2 )

enum sound_buffer_type
{
    sound_buffer_int16,
    sound_buffer_float32
};

struct sound_slot
{
    void*		callback;
    void*		user_data;

    int			status;

    volatile int	event;
    volatile int	event_answer;

    void*		in_buffer;
    void*		buffer;
    int 		frames;
    ticks_hr_t 		time; //output time
    
    SOUND_SCOPE_TYPE**	scope_buffer;
    bool		scope_buffer_locked;
    int			scope_buffer_frames;
    int			scope_buffer_channels;
    int			scope_buffer_flags;
};

struct sound_struct
{
    int			mode;
    int 		flags;
    int			freq;
    int			channels;
    int                 latency; //delay between the sound rendering and its output (in frames)

    sound_slot		slots[ SOUND_SLOTS ];
    void*		slot_buffer;
    int			slot_buffer_size;

    void*		in_buffer;
    sound_buffer_type	in_type;
    int			in_channels;
    int			in_enabled;
    sundog_mutex	in_mutex;
    int                 in_request; // >0 - enable INPUT; ==0 - disable INPUT.
    int                 in_request_answer; //Must be equal to input_request.
    
    void*		out_buffer;
    int 		out_frames;
    ticks_hr_t 		out_time; //output time
    
    sundog_mutex	mutex;

    //Flags:
    volatile bool 	main_sound_callback_working;
};

struct sundog_midi_event
{
    ticks_hr_t		t;
    size_t		size;
    uchar*		data;
};

struct sundog_midi_port
{
    bool		active;
    int			flags;
    UTF8_CHAR*		port_name;
    UTF8_CHAR*		dev_name;
    void*		device_specific;
};

struct sundog_midi_client
{
    sundog_midi_port*	ports;
    UTF8_CHAR*		name;
    void*		device_specific;
};

//Variables:

extern sound_struct g_snd;
extern bool g_snd_initialized;

#ifndef NOMIDI
extern ticks_hr_t g_last_midi_in_activity;
extern ticks_hr_t g_last_midi_out_activity;
extern sundog_mutex g_sundog_midi_mutex;
#endif

//Functions (sound):

void user_controlled_audio_callback( void* buffer, int frames, int latency, ticks_hr_t output_time );
int main_sound_output_callback( sound_struct* ss, int flags );

int sound_stream_init( int mode, int freq, int channels, int flags );
void sound_stream_set_slot_callback( int slot, void* callback, void* user_data );
void sound_stream_remove_slot_callback( int slot );
void sound_stream_create_scope_buffer( int slot, int frames, int channels, int flags );
void sound_stream_remove_scope_buffer( int slot );
SOUND_SCOPE_TYPE* sound_stream_get_scope_buffer( int slot, int channel );
bool sound_stream_is_scope_buffer_locked( int slot );
void sound_stream_scope_buffer_unlock( int slot );
void sound_stream_lock( int slot );
void sound_stream_unlock( int slot );
void sound_stream_play( int slot );
void sound_stream_stop( int slot );
void sound_stream_input( bool enable ); //Call this function in the main app thread, where the sound stream is not locked. Or...
void sound_stream_input_request( bool enable ); //... or use this function + sound_stream_handle_input_requests() in the main thread
void sound_stream_handle_input_requests( void );
const UTF8_CHAR* sound_stream_get_driver_name( void );
const UTF8_CHAR* sound_stream_get_driver_info( void );
int sound_stream_get_drivers( UTF8_CHAR*** names, UTF8_CHAR*** infos );
int sound_stream_get_devices( const UTF8_CHAR* driver, UTF8_CHAR*** names, UTF8_CHAR*** infos, bool input );
void sound_stream_close( void );

int device_sound_stream_init( int mode, int freq, int channels );
void device_sound_stream_input( bool enable );
const UTF8_CHAR* device_sound_stream_get_driver_name( void );
void device_sound_stream_close( void );

//Functions (MIDI):

int sundog_midi_init( void );
int sundog_midi_deinit( void );
int sundog_midi_client_open( sundog_midi_client* c, const UTF8_CHAR* name );
int sundog_midi_client_close( sundog_midi_client* c );
int sundog_midi_client_get_devices( sundog_midi_client* c, UTF8_CHAR*** devices, int flags );
int sundog_midi_client_open_port( sundog_midi_client* c, const UTF8_CHAR* port_name, const UTF8_CHAR* dev_name, int flags );
int sundog_midi_client_reopen_port( sundog_midi_client* c, int pnum );
int sundog_midi_client_close_port( sundog_midi_client* c, int pnum );
sundog_midi_event* sundog_midi_client_get_event( sundog_midi_client* c, int pnum );
int sundog_midi_client_next_event( sundog_midi_client* c, int pnum );
int sundog_midi_client_send_event( sundog_midi_client* c, int pnum, sundog_midi_event* evt );

int device_midi_client_open( sundog_midi_client* c, const UTF8_CHAR* name );
int device_midi_client_close( sundog_midi_client* c );
int device_midi_client_get_devices( sundog_midi_client* c, UTF8_CHAR*** devices, int flags );
int device_midi_client_open_port( sundog_midi_client* c, int pnum, const UTF8_CHAR* port_name, const UTF8_CHAR* dev_name, int flags );
int device_midi_client_close_port( sundog_midi_client* c, int pnum );
sundog_midi_event* device_midi_client_get_event( sundog_midi_client* c, int pnum );
int device_midi_client_next_event( sundog_midi_client* c, int pnum );
int device_midi_client_send_event( sundog_midi_client* c, int pnum, sundog_midi_event* evt );

#endif
