/*
    utils.cpp. Various functions: string list; random generator; ...
    This file is part of the SunDog engine.
    Copyright (C) 2002 - 2013 Alexander Zolotov <nightradio@gmail.com>
    www.warmplace.ru
*/

#include "../../core/core.h"
#include "../../core/debug.h"
#include "../../memory/memory.h"
#include "../../filesystem/v3nus_fs.h"
#include "../../main/user_code.h"
#include "../utils.h"

#include <stdarg.h>

#ifndef PALMOS
#include <stdlib.h>
#endif

#ifdef OSX
#include <CoreFoundation/CFBundle.h>
#include <ApplicationServices/ApplicationServices.h>
#endif

//LIST OF STRINGS:

#ifndef NOLIST

void list_init( list_data* data )
{
    data->items = (UTF8_CHAR*)MEM_NEW2( 4096 );
    data->items_ptr = (int*)MEM_NEW2( 64 * sizeof( int* ) );
    list_clear( data );
    list_reset_selection( data );
}

void list_close( list_data* data )
{
    if( data->items )
    {
	mem_free( data->items );
	data->items = 0;
    }
    if( data->items_ptr )
    {
	mem_free( data->items_ptr );
	data->items_ptr = 0;
    }
}

void list_clear( list_data* data )
{
    data->items_num = 0;
    data->start_item = 0;
}

void list_reset_items( list_data* data )
{
    data->items_num = 0;
}

void list_reset_selection( list_data* data )
{
    data->selected_item = -1;
}

void list_add_item( const UTF8_CHAR* item, char attr, list_data* data )
{
    if( data->items == 0 ) return;
    if( data->items_ptr == 0 ) return;
    
    int ptr, p;

    if( data->items_num == 0 )
    {
	data->items_ptr[ 0 ] = 0;
	ptr = 0;
    }
    else
    {
	ptr = data->items_ptr[ data->items_num ];
    }
    
    int len = (int)mem_strlen( item );
    int old_size = (int)mem_get_size( data->items );
    if( ptr + len + 2 >= old_size )
    {
	data->items = (UTF8_CHAR*)mem_resize( data->items, old_size * 2 );
	if( data->items == 0 ) return;
    }
    data->items[ ptr++ ] = attr | 128;
    for( p = 0; ; p++, ptr++ )
    {
	data->items[ ptr ] = item[ p ];
	if( item[ p ] == 0 ) break;
    }
    ptr++;
    
    data->items_num++;
    old_size = (int)mem_get_size( data->items_ptr ) / sizeof( int* );
    if( data->items_num >= old_size )
	data->items_ptr = (int*)mem_resize( data->items_ptr, old_size * 2 * sizeof( int* ) );
    if( data->items_ptr == 0 ) return;
    data->items_ptr[ data->items_num ] = ptr;
}

void list_delete_item( int item_num, list_data* data )
{
    if( data->items == 0 ) return;
    if( data->items_ptr == 0 ) return;
    
    int ptr, p;

    if( item_num < data->items_num && item_num >= 0 )
    {
	ptr = data->items_ptr[ item_num ]; //Offset of our item (in the "items")
	
	//Get item size (in chars):
	int size;
	for( size = 0; ; size++ )
	{
	    if( data->items[ ptr + size ] == 0 ) break;
	}
	
	//Delete it:
	int items_size = (int)mem_get_size( data->items );
	int items_ptr_size = (int)mem_get_size( data->items_ptr ) / sizeof( int* );
	for( p = ptr; p < items_size - size - 1 ; p++ )
	{
	    data->items[ p ] = data->items[ p + size + 1 ];
	}
	for( p = 0; p < data->items_num; p++ )
	{
	    if( data->items_ptr[ p ] > ptr ) data->items_ptr[ p ] -= size + 1;
	}
	for( p = item_num; p < data->items_num; p++ )
	{
	    if( p + 1 < items_ptr_size ) 
		data->items_ptr[ p ] = data->items_ptr[ p + 1 ];
	    else
		data->items_ptr[ p ] = 0;
	}
	data->items_num--;
	if( data->items_num < 0 ) data->items_num = 0;

	if( data->selected_item >= data->items_num ) data->selected_item = data->items_num - 1;
    }
}

void list_move_item_up( int item_num, list_data* data )
{
    if( data->items == 0 ) return;
    if( data->items_ptr == 0 ) return;
    
    if( item_num < data->items_num && item_num >= 0 )
    {
	if( item_num != 0 )
	{
	    int temp = data->items_ptr[ item_num - 1 ];
	    data->items_ptr[ item_num - 1 ] = data->items_ptr[ item_num ];
	    data->items_ptr[ item_num ] = temp;
	    if( item_num == data->selected_item ) data->selected_item--;
	}
    }
}	

void list_move_item_down( int item_num, list_data* data )
{
    if( data->items == 0 ) return;
    if( data->items_ptr == 0 ) return;
    
    if( item_num < data->items_num && item_num >= 0 )
    {
	if( item_num != data->items_num - 1 )
	{
	    int temp = data->items_ptr[ item_num + 1 ];
	    data->items_ptr[ item_num + 1 ] = data->items_ptr[ item_num ];
	    data->items_ptr[ item_num ] = temp;
	    if( item_num == data->selected_item ) data->selected_item++;
	}
    }
}

UTF8_CHAR* list_get_item( int item_num, list_data* data )
{
    if( data->items == 0 ) return 0;
    if( data->items_ptr == 0 ) return 0;
    
    if( item_num >= data->items_num ) return 0;
    if( item_num >= 0 )
	return data->items + data->items_ptr[ item_num ] + 1;
    else 
	return 0;
}

char list_get_attr( int item_num, list_data* data )
{
    if( data->items == 0 ) return 0;
    if( data->items_ptr == 0 ) return 0;
    
    if( item_num >= data->items_num ) return 0;
    if( item_num >= 0 )
	return data->items[ data->items_ptr[ item_num ] ] & 127;
    else
	return 0;
}

int list_get_selected_num( list_data* data )
{
    return data->selected_item;
}

void list_set_selected_num( int sel, list_data* data )
{
    data->selected_item = sel;
}

//Return values:
//1 - item1 > item2
//0 - item1 <= item2
int list_compare_items( int item1, int item2, list_data* data )
{
    UTF8_CHAR* i1 = data->items + data->items_ptr[ item1 ];
    UTF8_CHAR* i2 = data->items + data->items_ptr[ item2 ];
    UTF8_CHAR a1 = i1[ 0 ] & 127;
    UTF8_CHAR a2 = i2[ 0 ] & 127;
    i1++;
    i2++;
    
    int retval = 0;
    
    //Compare:
    if( a1 != a2 )
    {
	if( a1 == 1 ) retval = 0;
	if( a2 == 1 ) retval = 1;
    }
    else
    {
	for( int a = 0; ; a++ )
	{
	    if( i1[ a ] == 0 ) break;
	    if( i2[ a ] == 0 ) break;
	    if( i1[ a ] < i2[ a ] ) { break; }             //item1 < item2
	    if( i1[ a ] > i2[ a ] ) { retval = 1; break; } //item1 > item2
	}
    }
    
    return retval;
}

void list_sort( list_data* data )
{
    if( data->items == 0 ) return;
    if( data->items_ptr == 0 ) return;
    
    for(;;)
    {
	int s = 0;
	for( int a = 0; a < data->items_num - 1; a++ )
	{
	    if( list_compare_items( a, a + 1, data ) )
	    {
		s = 1;
		int temp = data->items_ptr[ a + 1 ];
		data->items_ptr[ a + 1 ] = data->items_ptr[ a ];
		data->items_ptr[ a ] = temp;
	    }
	}
	if( s == 0 ) break;
    }
}

#endif

//MUTEXES:

#ifdef PALMOS
#include "../../sound/sound.h"
#endif

//For PalmOS programmers: 
//sundog_mutex_lock/unlock functions not works in the sound callback.

int sundog_mutex_init( sundog_mutex* mutex, int attr )
{
    int retval = 0;
#ifdef UNIX
    pthread_mutexattr_t mutexattr;
    pthread_mutexattr_init( &mutexattr );
    pthread_mutexattr_settype( &mutexattr, PTHREAD_MUTEX_RECURSIVE );
    retval = pthread_mutex_init( &mutex->mutex, &mutexattr );
    pthread_mutexattr_destroy( &mutexattr );
#endif
#if defined(WIN) || defined(WINCE)
    InitializeCriticalSection( &mutex->mutex );
#endif
#ifdef PALMOS
    mutex->mutex_cnt = 0;
    mutex->main_sound_callback_working = &g_snd.main_sound_callback_working;
#endif
    return retval;
}

int sundog_mutex_destroy( sundog_mutex* mutex )
{
    int retval = 0;
#ifdef UNIX
    retval = pthread_mutex_destroy( &mutex->mutex );
#endif
#if defined(WIN) || defined(WINCE)
    DeleteCriticalSection( &mutex->mutex );
#endif
#ifdef PALMOS
    mutex->mutex_cnt = 0;
#endif
    return retval;
}

int sundog_mutex_lock( sundog_mutex* mutex )
{
    int retval = 0;
#ifdef UNIX
    retval = pthread_mutex_lock( &mutex->mutex );
#endif
#if defined(WIN) || defined(WINCE)
    EnterCriticalSection( &mutex->mutex );
#endif
#ifdef PALMOS
    if( *mutex->main_sound_callback_working )
    {
	//We are in the main sound callback.
	if( mutex->mutex_cnt <= 0 )
	{
	    mutex->mutex_cnt--;
	}
	else
	{
	    //Mutex is locked. But we can't wait for unlocking here,
	    //because all another "threads" are frozen (PalmOS is not multitask).
	    return 1;
	}
    }
    else
    {
	mutex->mutex_cnt++;
    }
#endif
    return retval;
}

int sundog_mutex_trylock( sundog_mutex* mutex )
{
    int retval = 0;
#ifdef UNIX
    retval = pthread_mutex_trylock( &mutex->mutex );
#endif
#if defined(WIN) || defined(WINCE)
    if( TryEnterCriticalSection( &mutex->mutex ) != 0 )
	retval = 0;
    else 
	retval = 1;
#endif
#ifdef PALMOS
    if( *mutex->main_sound_callback_working )
    {
	//We are in the main sound callback.
	if( mutex->mutex_cnt <= 0 )
	{
	    mutex->mutex_cnt--;
	}
	else
	{
	    //Mutex is locked. But we can't wait for unlocking here,
	    //because all another "threads" are frozen (PalmOS is not multitask).
	    return 1;
	}
    }
    else
    {
	mutex->mutex_cnt++;
    }
#endif
    return retval;
}

int sundog_mutex_unlock( sundog_mutex* mutex )
{
    int retval = 0;
#ifdef UNIX
    retval = pthread_mutex_unlock( &mutex->mutex );
#endif
#if defined(WIN) || defined(WINCE)
    LeaveCriticalSection( &mutex->mutex );
#endif
#ifdef PALMOS
    if( *mutex->main_sound_callback_working )
    {
	//We are in the main sound callback.
	mutex->mutex_cnt++;
    }
    else 
    {
	mutex->mutex_cnt--;
    }
#endif
    return retval;
}

//THREADS:

#ifndef PALMOS

#if defined(UNIX)
void* sundog_thread_handler( void* arg )
#endif
#if defined(WIN) || defined(WINCE)
DWORD __stdcall sundog_thread_handler( void* arg )
#endif
{
    sundog_thread* th = (sundog_thread*)arg;
    th->proc( th->arg );
    th->finished = 1;
#if defined(UNIX)
    pthread_exit( NULL );
#endif
#if defined(WIN) || defined(WINCE)
    ExitThread( 0 );
#endif
    return 0;
}

#endif

int sundog_thread_create( sundog_thread* th, void* (*proc)(void*), void* arg, int flags )
{
    int rv = 0;
    th->arg = arg;
    th->proc = proc;
    th->finished = 0;
#if defined(UNIX)
    int err = pthread_create( &th->th, 0, &sundog_thread_handler, (void*)th );
    if( err )
    {
	dprint( "pthread_create error %d\n", err );
	return 1;
    }
#endif
#if defined(WIN) || defined(WINCE)
    th->th = CreateThread( NULL, 0, &sundog_thread_handler, (void*)th, 0, NULL );
#endif
    return rv;
}

int sundog_thread_destroy( sundog_thread* th, int timeout )
{
    int rv = 2; //0 - successful; 1 - timeout; 2 - some error
    
    int err;

    bool dont_destroy_after_timeout = 0;
    if( timeout < 0 ) 
    {
	timeout = -timeout;
	dont_destroy_after_timeout = 1;
    }
    
    if( timeout == 0x7FFFFFFF )
    {
	//Infinite:
#ifdef UNIX
	err = pthread_join( th->th, 0 );
	if( err ) 
	{ 
	    dprint( "pthread_join() error %d\n", err ); 
	}
	else rv = 0;
#endif
#if defined(WIN) || defined(WINCE)
	err = WaitForSingleObject( th->th, INFINITE );
	if( err == WAIT_OBJECT_0 )
	{
	    //Thread terminated:
	    CloseHandle( th->th );
	    rv = 0;
	}
	else
	{
	    dprint( "WaitForSingleObject() error %d\n", err );
	}
#endif
	return rv;
    }

    int timeout_counter = timeout; //Timeout in milliseconds
    int step = 20; //ms
    while( timeout_counter > 0 )
    {
	if( th->finished )
	{
#ifdef UNIX
	    err = pthread_join( th->th, 0 );
	    if( err ) dprint( "pthread_join() error %d\n", err );
#endif
#if defined(WIN) || defined(WINCE)
	    CloseHandle( th->th );
#endif
	    rv = 0;
	    break;
	}
#ifdef UNIX
	struct timespec delay;
	delay.tv_sec = 0;
	delay.tv_nsec = 1000000 * step;
	nanosleep( &delay, NULL ); //Sleep for delay time
#endif
#if defined(WIN) || defined(WINCE)
	err = WaitForSingleObject( th->th, step );
	if( err == WAIT_OBJECT_0 )
	{
	    //Thread terminated:
	    CloseHandle( th->th );
	    rv = 0;
	    break;
	}
	if( err == WAIT_FAILED )
	{
	    return 2;
	}
#endif
	timeout_counter -= step;
    }
    if( timeout_counter <= 0 )
    {
	if( dont_destroy_after_timeout ) return 1;
#ifdef UNIX
#ifndef ANDROID
	err = pthread_cancel( th->th );
	if( err ) dprint( "pthread_cancel() error %d\n", err );
#endif
#endif
#if defined(WIN) || defined(WINCE)
	TerminateThread( th->th, 0 );
	CloseHandle( th->th );
#endif
	rv = 1;
    }

    return rv;
}

//PROFILES:

profile_data g_profile;

void profile_new( profile_data* p )
{
    if( p == 0 ) p = &g_profile;
    
    mem_set( p, sizeof( profile_data ), 0 );

    p->file_num = -1;
    p->num = 0;
    p->keys = (profile_key*)MEM_NEW2( sizeof( profile_key ) * 4 );
    mem_zero( p->keys );
}

int profile_resize( int new_num, profile_data* p )
{
    if( p == 0 ) p = &g_profile;

    int old_size = mem_get_size( p->keys ) / sizeof( profile_key );
    int new_size = new_num + 4;
    if( new_num > old_size )
    {
	p->keys = (profile_key*)mem_resize( p->keys, sizeof( profile_key ) * new_size );
	if( p->keys == 0 ) return -1;
	mem_set( &p->keys[ old_size ], ( new_size - old_size ) * sizeof( profile_key ), 0 );
    }
    p->num = new_num;
    
    return 0;
}

int profile_add_key( const UTF8_CHAR* key, const UTF8_CHAR* value, int line_num, profile_data* p )
{
    int rv = -1;
    
    if( p == 0 ) p = &g_profile;
    
    if( key && p->keys )
    {
	for( rv = 0; rv < p->num; rv++ )
	{
	    if( p->keys[ rv ].key == 0 ) break;
	}
	if( rv >= p->num )
	{
	    //Free item not found.
	    if( profile_resize( p->num + 1, p ) ) return -1;
	}
	if( value )
	{
	    p->keys[ rv ].value = (UTF8_CHAR*)MEM_NEW2( mem_strlen( value ) + 1 );
	    p->keys[ rv ].value[ 0 ] = 0;
	    mem_strcat( p->keys[ rv ].value, value );
	}
	p->keys[ rv ].key = (UTF8_CHAR*)MEM_NEW2( mem_strlen( key ) + 1 );
	if( p->keys[ rv ].key == 0 ) return -1;
	p->keys[ rv ].key[ 0 ] = 0;
	mem_strcat( p->keys[ rv ].key, key );
	p->keys[ rv ].line_num = line_num;
    }

    return rv;
}

void profile_remove_key( const UTF8_CHAR* key, profile_data* p )
{
    if( p == 0 ) p = &g_profile;
    
    if( key && p->keys )
    {
	int i;
	for( i = 0; i < p->num; i++ )
	{
	    if( p->keys[ i ].key )
	    {
		if( mem_strcmp( p->keys[ i ].key, key ) == 0 ) 
		    break;
	    }
	}
	if( i < p->num )
	{
	    mem_free( p->keys[ i ].value );
	    p->keys[ i ].value = 0;
	    p->keys[ i ].deleted = 1;
	}
    }
}

void profile_set_str_value( const UTF8_CHAR* key, const UTF8_CHAR* value, profile_data* p )
{
    if( p == 0 ) p = &g_profile;
    
    if( key && p->keys )
    {
	int i;
	for( i = 0; i < p->num; i++ )
	{
	    if( p->keys[ i ].key )
	    {
		if( mem_strcmp( p->keys[ i ].key, key ) == 0 ) 
		    break;
	    }
	}
	if( i < p->num )
	{
	    //Already exists:
	    mem_free( p->keys[ i ].value );
	    p->keys[ i ].value = 0;
	    p->keys[ i ].deleted = 0;
	    if( value )
	    {
		p->keys[ i ].value = (UTF8_CHAR*)MEM_NEW2( mem_strlen( value ) + 1 );
		p->keys[ i ].value[ 0 ] = 0;
		mem_strcat( p->keys[ i ].value, value );
	    }
	}
	else 
	{
	    //Not exists:
	    profile_add_key( key, value, -1, p );
	}
    }
}

void profile_set_int_value( const UTF8_CHAR* key, int value, profile_data* p )
{
    UTF8_CHAR ts[ 16 ];
    sprintf( ts, "%d", value );
    profile_set_str_value( key, ts, p );
}

int profile_get_int_value( const UTF8_CHAR* key, int default_value, profile_data* p )
{
    int rv = default_value;

    if( p == 0 ) p = &g_profile;

    if( key && p->keys )
    {
	int i;
	for( i = 0; i < p->num; i++ )
	{
	    if( p->keys[ i ].key )
	    {
		if( mem_strcmp( p->keys[ i ].key, key ) == 0 ) 
		    break;
	    }
	}
	if( i < p->num && p->keys[ i ].value )
	{
	    rv = string_to_int( p->keys[ i ].value );
	}
    }
    return rv;
}

UTF8_CHAR* profile_get_str_value( const UTF8_CHAR* key, const UTF8_CHAR* default_value, profile_data* p )
{
    UTF8_CHAR* rv = (UTF8_CHAR*)default_value;

    if( p == 0 ) p = &g_profile;

    if( key && p->keys )
    {
	int i;
	for( i = 0; i < p->num; i++ )
	{
	    if( p->keys[ i ].key )
		if( mem_strcmp( p->keys[ i ].key, key ) == 0 ) 
		    break;
	}
	if( i < p->num && p->keys[ i ].value )
	{
	    rv = p->keys[ i ].value;
	}
    }
    return rv;
}

void profile_close( profile_data* p )
{
    if( p == 0 ) p = &g_profile;

    mem_free( p->file_name );
    mem_free( p->source );
    p->source = 0;
    p->file_name = 0;
    
    if( p->num && p->keys )
    {
	for( int i = 0; i < p->num; i++ )
	{
	    if( p->keys[ i ].key ) mem_free( p->keys[ i ].key );
	    if( p->keys[ i ].value ) mem_free( p->keys[ i ].value );
	}
    }
    mem_free( p->keys );
    p->keys = 0;
    p->num = 0;
}

#define PROFILE_KEY_CHAR( cc ) ( !( cc < 0x21 || ptr >= size ) )

void profile_load( const UTF8_CHAR* filename, profile_data* p )
{
    UTF8_CHAR* str1 = (UTF8_CHAR*)MEM_NEW2( 1025 );
    if( str1 == 0 ) return;
    UTF8_CHAR* str2 = (UTF8_CHAR*)MEM_NEW2( 1025 );
    if( str2 == 0 ) return;
    str1[ 1024 ] = 0;
    str2[ 1024 ] = 0;
    str1[ 0 ] = 0;
    str2[ 0 ] = 0;
    int i = 0;
    
    int size = 0;
    UTF8_CHAR* f = 0;
    V3_FILE fp = 0;
    
    int ptr = 0;
    int c = 0;
    char comment_mode = 0;
    char key_mode = 0;
    int line_num = 0;
    
    if( p == 0 ) p = &g_profile;

    int pn = -1;
    if( filename == 0 )
    {
	pn = 0;
	while( user_profile_names[ pn ][ 0 ] != '0' )
	{
	    V3_FILE pf = v3_open( user_profile_names[ pn ], "rb" );
	    if( pf )
	    {
		v3_close( pf );
		filename = user_profile_names[ pn ];
		break;
	    }
	    else 
	    {
		filename = 0;
	    }
	    pn++;
	}
	if( filename == 0 ) pn = -1;
    }

    profile_close( p );
    profile_new( p );
    
    p->file_num = pn;
    p->file_name = (UTF8_CHAR*)MEM_NEW2( mem_strlen( filename ) + 1 );
    p->file_name[ 0 ] = 0;
    mem_strcat( p->file_name, filename );
    
    size = v3_get_file_size( filename );
    if( size == 0 ) goto load_end;
    f = (UTF8_CHAR*)MEM_NEW2( size );
    if( f == 0 ) goto load_end;
    fp = v3_open( filename, "rb" );
    if( fp )
    {
	v3_read( f, 1, size, fp );
	v3_close( fp );
    }
    else
    {
	mem_free( f );
	goto load_end;
    }
    
    while( ptr < size )
    {
        c = f[ ptr ];
        if( c == 0xD || c == 0xA )
        {
    	    comment_mode = 0; //Reset comment mode at the end of line
	    if( key_mode > 0 )
	    {
		profile_add_key( str1, str2, line_num, p );
	    }
	    key_mode = 0;
	    line_num++;
	}
	if( comment_mode == 0 )
	{
	    if( f[ ptr ] == '/' && f[ ptr + 1 ] == '/' )
	    {
	        comment_mode = 1; //Comments
		ptr += 2;
	        continue;
	    }
	    if( PROFILE_KEY_CHAR( c ) )
	    {
	        if( key_mode == 0 )
	        {
	    	    //Get key name:
		    str2[ 0 ] = 0;
		    for( i = 0; i < 1024; i++ )
		    {
			if( !PROFILE_KEY_CHAR( f[ ptr ] ) ) 
			{ 
			    str1[ i ] = 0;
			    ptr--;
			    break; 
			}
		        str1[ i ] = f[ ptr ];
			ptr++;
		    }
		    key_mode = 1;
		}
		else if( key_mode == 1 )
		{
		    //Get value:
		    str2[ 0 ] = 0;
		    if( f[ ptr ] == '"' )
		    {
			ptr++;
			for( i = 0; i < 1024; i++ )
			{
			    if( ptr >= size || f[ ptr ] == '"' ) 
			    { 
			        str2[ i ] = 0;
			        break; 
			    }
			    str2[ i ] = f[ ptr ];
			    ptr++;
			}
		    }
		    else
		    {
			for( i = 0; i < 1024; i++ )
			{
			    if( ptr >= size || f[ ptr ] < 0x21 ) 
			    { 
			        str2[ i ] = 0;
				ptr--;
			        break; 
			    }
			    str2[ i ] = f[ ptr ];
			    ptr++;
			}
		    }
		    key_mode = 2;
		}
	    }
	}
	ptr++;
    }
    if( key_mode > 0 )
    {
	profile_add_key( str1, str2, line_num, p );
    }

    p->source = f;
    
load_end:

    mem_free( str1 );
    mem_free( str2 );
}

void profile_save_key( int key, V3_FILE f, profile_data* p )
{
    if( ( p->keys[ key ].key ) && ( p->keys[ key ].deleted == 0 ) )
    {
	v3_write( p->keys[ key ].key, 1, mem_strlen( p->keys[ key ].key ), f );
	if( p->keys[ key ].value )
	{
	    int vsize = mem_strlen( p->keys[ key ].value );
	    bool q = 0;
	    for( int i2 = 0; i2 < vsize; i2++ )
	    {
		int cc = p->keys[ key ].value[ i2 ];
		if( cc < 0x21 || cc == '/' ) 
		{
		    q = 1;
		    break;
		}
	    }
	    v3_putc( ' ', f );
	    if( q ) v3_putc( '"', f );
	    v3_write( p->keys[ key ].value, 1, vsize, f );
	    if( q ) v3_putc( '"', f );
	    v3_putc( '\n', f );
	}
    }
}

int profile_save( profile_data* p )
{
    int rv = 0;
    
    if( p == 0 ) p = &g_profile;
    
    //Get file name for writing:
    V3_FILE f = 0;
    if( p->file_name )
	f = v3_open( p->file_name, "wb" );
    if( f )
    {
	v3_close( f );
    }
    else 
    {
	if( p->file_num < 0 )
	{
	    for( p->file_num = 0; ; p->file_num++ )
	    {
		if( user_profile_names[ p->file_num ][ 0 ] == '0' ) break;
	    }
	    p->file_num--;
	}
	else 
	{
	    p->file_num--;
	}
	if( p->file_num < 0 ) return 1;
	while( 1 )
	{
	    mem_free( p->file_name );
	    p->file_name = 0;
	    p->file_name = (UTF8_CHAR*)MEM_NEW2( mem_strlen( user_profile_names[ p->file_num ] ) + 1 );
	    p->file_name[ 0 ] = 0;
	    mem_strcat( p->file_name, user_profile_names[ p->file_num ] );
	    f = v3_open( p->file_name, "wb" );
	    if( f )
	    {
		v3_close( f );
		break;
	    }
	    p->file_num--;
	    if( p->file_num < 0 ) return 1;
	}
    }
    
    //Save profile:
    f = v3_open( p->file_name, "wb" );
    if( f == 0 ) return 1;
    int line_num = 0;
    bool new_line = 1;
    int size = mem_get_size( p->source );
    for( int i = 0; i < size; i++ )
    {
	int c = p->source[ i ];
	if( c == 0xA || c == 0xD )
	{
	    //New line:
	    line_num++;
	    new_line = 1;
	    v3_putc( c, f );
	}
	else 
	{
	    int key = -1;
	    if( new_line )
	    {
		new_line = 0;
		for( int k = 0; k < p->num; k++ )
		{
		    if( p->keys[ k ].line_num == line_num )
		    {
			//Key found for this line:
			key = k;
			break;
		    }
		}
	    }
	    if( key >= 0 )
	    {
		//Ignore this line:
		while( i < size )
		{
		    c = p->source[ i ];
		    if( c == 0xA || c == 0xD )
		    {
			if( i + 1 < size )
			{
			    if( c == 0xA && p->source[ i + 1 ] == 0xD ) i++;
			    if( c == 0xD && p->source[ i + 1 ] == 0xA ) i++;
			}
			line_num++;
			new_line = 1;
			break;
		    }
		    i++;
		}
		//Save value:
		profile_save_key( key, f, p );
	    }
	    else 
	    {
		v3_putc( c, f );
	    }
	}
    }
    for( int k = 0; k < p->num; k++ )
    {
	if( p->keys[ k ].line_num == -1 )
	{
	    profile_save_key( k, f, p );
	}
    }
    v3_close( f );
    
    return rv;
}

//WORKING WITH STRINGS:

void int_to_string( int value, UTF8_CHAR* str )
{
    int n;
    UTF8_CHAR ts[ 10 ];
    int ts_ptr = 0;
    
    if( value < 0 )
    {
    	*str = '-';
	str++;
	value = -value;
    }

    n = value % 10; ts[ ts_ptr++ ] = (UTF8_CHAR) n + 48; value /= 10; if( !value ) goto int_finish;
    n = value % 10; ts[ ts_ptr++ ] = (UTF8_CHAR) n + 48; value /= 10; if( !value ) goto int_finish;
    n = value % 10; ts[ ts_ptr++ ] = (UTF8_CHAR) n + 48; value /= 10; if( !value ) goto int_finish;
    n = value % 10; ts[ ts_ptr++ ] = (UTF8_CHAR) n + 48; value /= 10; if( !value ) goto int_finish;
    n = value % 10; ts[ ts_ptr++ ] = (UTF8_CHAR) n + 48; value /= 10; if( !value ) goto int_finish;
    n = value % 10; ts[ ts_ptr++ ] = (UTF8_CHAR) n + 48; value /= 10; if( !value ) goto int_finish;
    n = value % 10; ts[ ts_ptr++ ] = (UTF8_CHAR) n + 48; value /= 10; if( !value ) goto int_finish;
    n = value % 10; ts[ ts_ptr++ ] = (UTF8_CHAR) n + 48; value /= 10; if( !value ) goto int_finish;
    n = value % 10; ts[ ts_ptr++ ] = (UTF8_CHAR) n + 48; value /= 10; if( !value ) goto int_finish;
    n = value; ts[ ts_ptr++ ] = (UTF8_CHAR) n + 48;
int_finish:

    while( ts_ptr )
    {
	ts_ptr--;
	*str = ts[ ts_ptr ];
	str++;
    }
    
    *str = 0;
}

void int_to_string_ext( int value, UTF8_CHAR* str, int chars )
{
    if( value < 0 )
    {
    	*str = '-';
	str++;
	value = -value;
    }
    str += chars - 1;
    *( str + 1 ) = 0;
    for( int i = 0; i < chars; i++ )
    {
	*str = '0' + value % 10;
	value /= 10;
	str--;
    }
}
 
void hex_int_to_string( uint value, UTF8_CHAR* str )
{
    int n;
    UTF8_CHAR ts[ 8 ];
    int ts_ptr = 0;

    n = value & 15; if( n > 9 ) ts[ ts_ptr++ ] = (UTF8_CHAR) ( n - 10 ) + 'A'; else ts[ ts_ptr++ ] = (UTF8_CHAR) n + '0'; value >>= 4; if( !value ) goto hex_int_finish;
    n = value & 15; if( n > 9 ) ts[ ts_ptr++ ] = (UTF8_CHAR) ( n - 10 ) + 'A'; else ts[ ts_ptr++ ] = (UTF8_CHAR) n + '0'; value >>= 4; if( !value ) goto hex_int_finish;
    n = value & 15; if( n > 9 ) ts[ ts_ptr++ ] = (UTF8_CHAR) ( n - 10 ) + 'A'; else ts[ ts_ptr++ ] = (UTF8_CHAR) n + '0'; value >>= 4; if( !value ) goto hex_int_finish;
    n = value & 15; if( n > 9 ) ts[ ts_ptr++ ] = (UTF8_CHAR) ( n - 10 ) + 'A'; else ts[ ts_ptr++ ] = (UTF8_CHAR) n + '0'; value >>= 4; if( !value ) goto hex_int_finish;
    n = value & 15; if( n > 9 ) ts[ ts_ptr++ ] = (UTF8_CHAR) ( n - 10 ) + 'A'; else ts[ ts_ptr++ ] = (UTF8_CHAR) n + '0'; value >>= 4; if( !value ) goto hex_int_finish;
    n = value & 15; if( n > 9 ) ts[ ts_ptr++ ] = (UTF8_CHAR) ( n - 10 ) + 'A'; else ts[ ts_ptr++ ] = (UTF8_CHAR) n + '0'; value >>= 4; if( !value ) goto hex_int_finish;
    n = value & 15; if( n > 9 ) ts[ ts_ptr++ ] = (UTF8_CHAR) ( n - 10 ) + 'A'; else ts[ ts_ptr++ ] = (UTF8_CHAR) n + '0'; value >>= 4; if( !value ) goto hex_int_finish;
    n = value & 15; if( n > 9 ) ts[ ts_ptr++ ] = (UTF8_CHAR) ( n - 10 ) + 'A'; else ts[ ts_ptr++ ] = (UTF8_CHAR) n + '0';
hex_int_finish:

    while( ts_ptr )
    {
	ts_ptr--;
	*str = ts[ ts_ptr ];
	str++;
    }
    
    *str = 0;
}

void hex_int_to_string_ext( uint value, UTF8_CHAR* str, int chars )
{
    str += chars - 1;
    *( str + 1 ) = 0;
    for( int i = 0; i < chars; i++ )
    {
	int v = value & 15;
	if( v < 10 )
	{
	    *str = '0' + v;
	}
	else
	{
	    *str = 'A' + ( v - 10 );
	}
	value /= 16;
	str--;
    }
}

int string_to_int( const UTF8_CHAR* str )
{
    int res = 0;
    int d = 1;
    int minus = 1;
    for( int a = mem_strlen( str ) - 1; a >= 0; a-- )
    {
	int v = str[ a ];
	if( v >= '0' && v <= '9' )
	{
	    v -= '0';
	    res += v * d;
	    d *= 10;
	}
	else
	if( v == '-' ) minus = -1;
    }
    return res * minus;
}

int hex_string_to_int( const UTF8_CHAR* str )
{
    int res = 0;
    int d = 1;
    int minus = 1;
    for( int a = mem_strlen( str ) - 1; a >= 0; a-- )
    {
	int v = str[ a ];
	if( ( v >= '0' && v <= '9' ) || ( v >= 'A' && v <= 'F' ) || ( v >= 'a' && v <= 'f' ) )
	{
	    if( v >= '0' && v <= '9' ) v -= '0';
	    else
	    if( v >= 'A' && v <= 'F' ) { v -= 'A'; v += 10; }
	    else
	    if( v >= 'a' && v <= 'f' ) { v -= 'a'; v += 10; }
	    res += v * d;
	    d *= 16;
	}
	else
	if( v == '-' ) minus = -1;
    }
    return res * minus;
}

char int_to_hchar( int value )
{
    if( value < 10 ) return value + '0';
	else return ( value - 10 ) + 'A';
}

#if HEAPSIZE <= 4
    #define TEMP_STR_SIZE_DWORDS 256
#else
    #define TEMP_STR_SIZE_DWORDS 1024
#endif

int g_temp_str[ TEMP_STR_SIZE_DWORDS * 4 ];
int g_str_num = 0;

UTF16_CHAR* utf8_to_utf16( UTF16_CHAR* dest, int dest_chars, const UTF8_CHAR* s )
{
    unsigned char* src = (unsigned char*)s;
    if( dest == 0 )
    {
	char *t = (char*)&g_temp_str[ TEMP_STR_SIZE_DWORDS * g_str_num ];
	dest = (UTF16_CHAR*)t;
	dest_chars = TEMP_STR_SIZE_DWORDS * 2;
    }
    UTF16_CHAR* dest_begin = dest;
    UTF16_CHAR* dest_end = dest + dest_chars;
    while( *src != 0 )
    {
	if( *src < 128 ) 
	{
	    *dest = (UTF16_CHAR)(*src);
	    src++;
	    dest++;
	}
	else
	{
	    if( *src & 64 )
	    {
		int res;
		if( ( *src & 32 ) == 0 )
		{
		    //Two bytes:
		    res = ( *src & 31 ) << 6;
		    src++;
		    res |= ( *src & 63 );
		    src++;
		    *dest = (UTF16_CHAR)res;
		    dest++;
		}
		else if( ( *src & 16 ) == 0 )
		{
		    //Three bytes:
		    res = ( *src & 15 ) << 12;
		    src++;
		    res |= ( *src & 63 ) << 6;
		    src++;
		    res |= ( *src & 63 );
		    src++;
		    *dest = (UTF16_CHAR)res;
		    dest++;
		}
		else if( ( *src & 8 ) == 0 )
		{
		    //Four bytes:
		    res = ( *src & 7 ) << 18;
		    src++;
		    res |= ( *src & 63 ) << 12;
		    src++;
		    res |= ( *src & 63 ) << 6;
		    src++;
		    res |= ( *src & 63 );
		    src++;
		    *dest = 0xD800 | (UTF16_CHAR)( res & 1023 );
		    dest++;
		    if( dest >= dest_end )
		    {
			dest--;
			break;
		    }
		    *dest = 0xDC00 | (UTF16_CHAR)( ( res >> 10 ) & 1023 );
		    dest++;
		}
		else
		{
		    //Unknown byte:
		    src++;
		    continue;
		}
	    }
	    else
	    {
		//Unknown byte:
		src++;
		continue;
	    }
	}
	if( dest >= dest_end )
	{
	    dest--;
	    break;
	}
    }
    *dest = 0;
    g_str_num++; g_str_num &= 3;
    return dest_begin;
}

UTF32_CHAR* utf8_to_utf32( UTF32_CHAR* dest, int dest_chars, const UTF8_CHAR* s )
{
    unsigned char* src = (unsigned char*)s;
    if( dest == 0 )
    {
	char* t = (char*)&g_temp_str[ TEMP_STR_SIZE_DWORDS * g_str_num ];
	dest = (UTF32_CHAR*)t;
	dest_chars = TEMP_STR_SIZE_DWORDS;
    }
    UTF32_CHAR* dest_begin = dest;
    UTF32_CHAR* dest_end = dest + dest_chars;
    while( *src != 0 )
    {
	if( *src < 128 ) 
	{
	    *dest = *src;
	    src++;
	    dest++;
	}
	else
	{
	    if( *src & 64 )
	    {
		int res;
		if( ( *src & 32 ) == 0 )
		{
		    //Two bytes:
		    res = ( *src & 31 ) << 6;
		    src++;
		    res |= ( *src & 63 );
		    src++;
		    *dest = (UTF32_CHAR)res;
		    dest++;
		}
		else if( ( *src & 16 ) == 0 )
		{
		    //Three bytes:
		    res = ( *src & 15 ) << 12;
		    src++;
		    res |= ( *src & 63 ) << 6;
		    src++;
		    res |= ( *src & 63 );
		    src++;
		    *dest = (UTF32_CHAR)res;
		    dest++;
		}
		else if( ( *src & 8 ) == 0 )
		{
		    //Four bytes:
		    res = ( *src & 7 ) << 18;
		    src++;
		    res |= ( *src & 63 ) << 12;
		    src++;
		    res |= ( *src & 63 ) << 6;
		    src++;
		    res |= ( *src & 63 );
		    src++;
		    *dest = (UTF32_CHAR)res;
		    dest++;
		}
		else
		{
		    //Unknown byte:
		    src++;
		    continue;
		}
	    }
	    else
	    {
		//Unknown byte:
		src++;
		continue;
	    }
	}
	if( dest >= dest_end )
	{
	    dest--;
	    break;
	}
    }
    *dest = 0;
    g_str_num++; g_str_num &= 3;
    return dest_begin;
}

UTF8_CHAR* utf16_to_utf8( UTF8_CHAR* dst, int dest_chars, const UTF16_CHAR* src )
{
    unsigned char* dest = (unsigned char*)dst;
    if( dest == 0 )
    {
	dest = (unsigned char*)&g_temp_str[ TEMP_STR_SIZE_DWORDS * g_str_num ];
	dest_chars = TEMP_STR_SIZE_DWORDS * 4;
    }
    unsigned char* dest_begin = dest;
    unsigned char* dest_end = dest + dest_chars;
    while( *src != 0 )
    {
	int res;
	if( ( *src & ~1023 ) != 0xD800 )
	{
	    res = *src;
	    src++;
	}
	else
	{
	    res = *src & 1023;
	    src++;
	    res |= ( *src & 1023 ) << 10;
	    src++;
	}
	if( res < 128 )
	{
	    *dest = (unsigned char)res;
	    dest++;
	}
	else if( res < 0x800 )
	{
	    if( dest >= dest_end - 2 ) break;
	    *dest = 0xC0 | ( ( res >> 6 ) & 31 );
	    dest++;
	    *dest = 0x80 | ( ( res >> 0 ) & 63 );
	    dest++;
	}
	else if( res < 0x10000 )
	{
	    if( dest >= dest_end - 3 ) break;
	    *dest = 0xE0 | ( ( res >> 12 ) & 15 );
	    dest++;
	    *dest = 0x80 | ( ( res >> 6 ) & 63 );
	    dest++;
	    *dest = 0x80 | ( ( res >> 0 ) & 63 );
	    dest++;
	}
	else
	{
	    if( dest >= dest_end - 4 ) break;
	    *dest = 0xF0 | ( ( res >> 18 ) & 7 );
	    dest++;
	    *dest = 0x80 | ( ( res >> 12 ) & 63 );
	    dest++;
	    *dest = 0x80 | ( ( res >> 6 ) & 63 );
	    dest++;
	    *dest = 0x80 | ( ( res >> 0 ) & 63 );
	    dest++;
	}
	if( dest >= dest_end )
	{
	    dest--;
	    break;
	}
    }
    *dest = 0;
    g_str_num++; g_str_num &= 3;
    return (UTF8_CHAR*)dest_begin;
}

UTF8_CHAR* utf32_to_utf8( UTF8_CHAR* dst, int dest_chars, const UTF32_CHAR* src )
{
    unsigned char* dest = (unsigned char*)dst;
    if( dest == 0 )
    {
	dest = (unsigned char*)&g_temp_str[ TEMP_STR_SIZE_DWORDS * g_str_num ];
	dest_chars = TEMP_STR_SIZE_DWORDS * 4;
    }
    unsigned char* dest_begin = dest;
    unsigned char* dest_end = dest + dest_chars;
    while( *src != 0 )
    {
	int res = (int)*src;
	src++;
	if( res < 128 )
	{
	    *dest = (unsigned char)res;
	    dest++;
	}
	else if( res < 0x800 )
	{
	    if( dest >= dest_end - 2 ) break;
	    *dest = 0xC0 | ( ( res >> 6 ) & 31 );
	    dest++;
	    *dest = 0x80 | ( ( res >> 0 ) & 63 );
	    dest++;
	}
	else if( res < 0x10000 )
	{
	    if( dest >= dest_end - 3 ) break;
	    *dest = 0xE0 | ( ( res >> 12 ) & 15 );
	    dest++;
	    *dest = 0x80 | ( ( res >> 6 ) & 63 );
	    dest++;
	    *dest = 0x80 | ( ( res >> 0 ) & 63 );
	    dest++;
	}
	else
	{
	    if( dest >= dest_end - 4 ) break;
	    *dest = 0xF0 | ( ( res >> 18 ) & 7 );
	    dest++;
	    *dest = 0x80 | ( ( res >> 12 ) & 63 );
	    dest++;
	    *dest = 0x80 | ( ( res >> 6 ) & 63 );
	    dest++;
	    *dest = 0x80 | ( ( res >> 0 ) & 63 );
	    dest++;
	}
	if( dest >= dest_end )
	{
	    dest--;
	    break;
	}
    }
    *dest = 0;
    g_str_num++; g_str_num &= 3;
    return (UTF8_CHAR*)dest_begin;
}

int utf8_to_utf32_char( const UTF8_CHAR* str, UTF32_CHAR* res )
{
    *res = 0;
    unsigned char* src = (unsigned char*)str;
    while( *src != 0 )
    {
	if( *src < 128 ) 
	{
	    *res = (UTF32_CHAR)*src;
	    return 1;
	}
	else
	{
	    if( *src & 64 )
	    {
		if( ( *src & 32 ) == 0 )
		{
		    //Two bytes:
		    *res = ( *src & 31 ) << 6;
		    src++;
		    *res |= ( *src & 63 );
		    return 2;
		}
		else if( ( *src & 16 ) == 0 )
		{
		    //Three bytes:
		    *res = ( *src & 15 ) << 12;
		    src++;
		    *res |= ( *src & 63 ) << 6;
		    src++;
		    *res |= ( *src & 63 );
		    return 3;
		}
		else if( ( *src & 8 ) == 0 )
		{
		    //Four bytes:
		    *res = ( *src & 7 ) << 18;
		    src++;
		    *res |= ( *src & 63 ) << 12;
		    src++;
		    *res |= ( *src & 63 ) << 6;
		    src++;
		    *res |= ( *src & 63 );
		    return 4;
		}
		else
		{
		    //Unknown byte:
		    *res = '?';
		    return 1;
		}
	    }
	    else
	    {
		//Unknown byte:
		*res = '?';
		return 1;
	    }
	}
    }
    return 0;
}

int utf8_to_utf32_char_safe( UTF8_CHAR* str, size_t str_size, UTF32_CHAR* res )
{
    *res = 0;
    uchar* src = (uchar*)str;
    uchar* src_end = src + str_size;
    while( *src != 0 && src != src_end )
    {
	if( *src < 128 ) 
	{
	    *res = (UTF32_CHAR)*src;
	    return 1;
	}
	else
	{
	    if( *src & 64 )
	    {
		if( ( *src & 32 ) == 0 )
		{
		    //Two bytes:
		    *res = ( *src & 31 ) << 6;
		    if( src == src_end ) return 1;
		    src++;
		    *res |= ( *src & 63 );
		    return 2;
		}
		else if( ( *src & 16 ) == 0 )
		{
		    //Three bytes:
		    *res = ( *src & 15 ) << 12;
		    if( src == src_end ) return 1;
		    src++;
		    *res |= ( *src & 63 ) << 6;
		    if( src == src_end ) return 2;
		    src++;
		    *res |= ( *src & 63 );
		    return 3;
		}
		else if( ( *src & 8 ) == 0 )
		{
		    //Four bytes:
		    *res = ( *src & 7 ) << 18;
		    if( src == src_end ) return 1;
		    src++;
		    *res |= ( *src & 63 ) << 12;
		    if( src == src_end ) return 2;
		    src++;
		    *res |= ( *src & 63 ) << 6;
		    if( src == src_end ) return 3;
		    src++;
		    *res |= ( *src & 63 );
		    return 4;
		}
		else
		{
		    //Unknown byte:
		    *res = '?';
		    return 1;
		}
	    }
	    else
	    {
		//Unknown byte:
		*res = '?';
		return 1;
	    }
	}
    }
    return 0;
}

void utf8_unix_slash_to_windows( UTF8_CHAR* str )
{
    while( *str != 0 )
    {
	if( *str == 0x2F ) *str = 0x5C;	    
	str++;
    }
}

void utf16_unix_slash_to_windows( UTF16_CHAR* str )
{
    while( *str != 0 )
    {
	if( *str == 0x2F ) *str = 0x5C;	    
	str++;
    }
}

void utf32_unix_slash_to_windows( UTF32_CHAR* str )
{
    while( *str != 0 )
    {
	if( *str == 0x2F ) *str = 0x5C;	    
	str++;
    }
}

#ifdef PALMOS

int sprintf( char* dest_str, const char* str, ... )
{
    va_list p;
    va_start( p, str );
    
    int ptr = 0;
    int ptr2 = 0;
    UTF8_CHAR num_str[ 16 ];
    int len;
    
    //Make a string:
    for(;;)
    {
	if( str[ ptr ] == 0 ) break;
	if( str[ ptr ] == '%' )
	{
	    char next_c = str[ ptr + 1 ];
	    int chars = 0;
	    if( next_c >= '0' && next_c <= '9' )
	    {
		chars = ( next_c - '0' ) * 10;
		ptr++;
		next_c = str[ ptr + 1 ];
		chars += next_c - '0';
		ptr++;
		next_c = str[ ptr + 1 ];
	    }
	    if( next_c == '%' )
	    {
		dest_str[ ptr2 ] = '%';
                ptr2++;
	    }
	    if( next_c == 'd' || next_c == 'x' )
	    {
		//Integer value:
		int arg = va_arg( p, int );
		if( str[ ptr + 1 ] == 'd' )
		{
		    if( chars == 0 )
    			int_to_string( arg, num_str );
    		    else
    			int_to_string_ext( arg, num_str, chars );
		}
		else
		{
		    if( chars == 0 )
			hex_int_to_string( arg, num_str );
		    else
			hex_int_to_string_ext( arg, num_str, chars );
		}
		len = mem_strlen( num_str );
		mem_copy( dest_str + ptr2, num_str, len );
		ptr2 += len;
		ptr++;
	    }
	    else
	    {
		if( next_c == 's' )
		{
		    //ASCII string:
		    UTF8_CHAR* arg2 = va_arg( p, UTF8_CHAR* );
		    if( arg2 )
		    {
			len = mem_strlen( arg2 );
			if( len )
			{
			    mem_copy( dest_str + ptr2, arg2, len );
			    ptr2 += len;
			}
		    }
		    ptr++;
		}
	    }
	}
	else
	{
	    dest_str[ ptr2 ] = str[ ptr ];
	    ptr2++;
	}
	ptr++;
    }
    dest_str[ ptr2 ] = 0;
    va_end( p );
    
    return ptr2;
}

#endif

//UNDO MANAGER:

#define UNDO_DEBUG

void undo_init( size_t size_limit, int (*action_handler)( UNDO_HANDLER_PARS ), void* user_data, undo_data* u )
{
    mem_set( u, sizeof( undo_data ), 0 );
    u->data_size_limit = size_limit / 2;
    u->actions_num_limit = round_to_power_of_two( ( size_limit / 2 ) / sizeof( undo_action ) );
    u->action_handler = action_handler;
    u->user_data = user_data;
}

void undo_remove_action( size_t n, undo_data* u )
{
    undo_action* a = &u->actions[ ( u->first_action + n ) & ( u->actions_num_limit - 1 ) ];
    u->data_size -= mem_get_size( a->data );
    //printf( "remove %d. size: %d. limit: %d\n", mem_get_size( a->data ), u->data_size, u->data_size_limit );
    mem_free( a->data );
    a->data = 0;
}

void undo_deinit( undo_data* u )
{
    if( u->actions )
    {
	for( size_t i = 0; i < u->actions_num; i++ )
	{
	    undo_remove_action( i, u );
	}
	mem_free( u->actions );
    }
}

void undo_reset( undo_data* u )
{
#ifdef UNDO_DEBUG
    dprint( "undo_reset(). data_size: %d\n", u->data_size );
#endif
    if( u->actions )
    {
	for( size_t i = 0; i < u->actions_num; i++ )
	{
	    undo_remove_action( i, u );
	}
	mem_free( u->actions );
	u->actions = 0;
    }
    u->status = 0;
    u->data_size = 0;
    u->level = 0;
    u->first_action = 0;
    u->cur_action = 0;
    u->actions_num = 0;
}

void undo_remove_first_actions( size_t level_bound, bool use_level_bound, undo_data* u )
{
    while( ( u->data_size > u->data_size_limit ) && ( u->actions_num > 0 ) )
    {
	//Not enough memory.
	//Remove the first action.
	undo_action* a = &u->actions[ ( u->first_action + 0 ) & ( u->actions_num_limit - 1 ) ];
	size_t first_level = a->level;
	if( use_level_bound )
	{
	    if( first_level == level_bound ) break; //Can't remove current level!
	}
	undo_remove_action( 0, u );
	u->first_action++;
	u->first_action &= u->actions_num_limit - 1;
	u->cur_action--;
	u->actions_num--;
	
	while( u->actions_num > 0 )
	{
	    a = &u->actions[ ( u->first_action + 0 ) & ( u->actions_num_limit - 1 ) ];
	    size_t next_level = a->level;
	    if( next_level == first_level )
	    {
		//Remove all actions with this level:
		undo_remove_action( 0, u );
		u->first_action++;
		u->first_action &= u->actions_num_limit - 1;
		u->cur_action--;
		u->actions_num--;
	    }
	    else
	    {
		break;
	    }
	}
    }
}

int undo_add_action( undo_action* action, undo_data* u )
{
    //Execute action:
    action->level = u->level;
    u->status = UNDO_STATUS_ADD_ACTION;
    int action_error = u->action_handler( 1, action, u );
    u->status = UNDO_STATUS_NONE;
    if( action_error )
	return action_error;
    
    if( u->actions == 0 )
    {
	u->actions = (undo_action*)MEM_NEW2( sizeof( undo_action ) * u->actions_num_limit );
    }
    
    if( u->cur_action >= u->actions_num_limit )
    {
	//Overflow in the table with actions.
	//Remove the first action.
	undo_remove_action( 0, u );
	u->first_action++;
	u->first_action &= u->actions_num_limit - 1;
	u->cur_action--;
	u->actions_num--;
    }
    
    //Remove previous history.
    for( size_t i = u->cur_action; i < u->actions_num; i++ )
    {
	undo_remove_action( i, u );
    }
    u->actions_num -= u->actions_num - u->cur_action;
    
    //Copy action:
    undo_action* a = &u->actions[ ( u->first_action + u->cur_action ) & ( u->actions_num_limit - 1 ) ];
    mem_copy( a, action, sizeof( undo_action ) );

    //Increase number of actions:
    u->actions_num++;
    
    //Go to the next action:
    u->cur_action++;
    
    //Get data size:
    u->data_size += mem_get_size( a->data );
    
    undo_remove_first_actions( 0, 0, u );

    return 0;
}

void undo_next_level( undo_data* u )
{
    u->level++;
}

void execute_undo( undo_data* u )
{
    bool l = 0;
    size_t level = 0;
    
    while( u->cur_action > 0 )
    {
	u->cur_action--;
	undo_action* a = &u->actions[ ( u->first_action + u->cur_action ) & ( u->actions_num_limit - 1 ) ];
	
	if( l == 0 )
	{
	    level = a->level;
	    l = 1;
	}
	else 
	{
	    if( a->level != level )
	    {
		u->cur_action++;
		break;
	    }
	}
	
	size_t old_size = mem_get_size( a->data );
	if( u->action_handler( 0, a, u ) == 0 )
	{
	    size_t new_size = mem_get_size( a->data );
	    u->data_size -= old_size - new_size;
	    undo_remove_first_actions( level, 1, u );
	    if( u->data_size > u->data_size_limit )
	    {
		//Error:
#ifdef UNDO_DEBUG
		dprint( "execute_undo(). mem.error. data_size: %d\n", (int)u->data_size );
#endif
		undo_reset( u );
		break;
	    }
	}
	else 
	{
	    //Error:
#ifdef UNDO_DEBUG
	    dprint( "execute_undo(). action error. data_size: %d\n", (int)u->data_size );
#endif
	    undo_reset( u );
	    break;
	}
    }
}

void execute_redo( undo_data* u )
{
    bool l = 0;
    size_t level = 0;
    
    while( u->cur_action < u->actions_num )
    {
	undo_action* a = &u->actions[ ( u->first_action + u->cur_action ) & ( u->actions_num_limit - 1 ) ];
	
	if( l == 0 )
	{
	    level = a->level;
	    l = 1;
	}
	else 
	{
	    if( a->level != level ) break;
	}
	
	size_t old_size = mem_get_size( a->data );
	if( u->action_handler( 1, a, u ) == 0 )
	{
	    size_t new_size = mem_get_size( a->data );
	    u->data_size -= old_size - new_size;
	    undo_remove_first_actions( level, 1, u );
	    if( u->data_size > u->data_size_limit )
	    {
		//Error:
#ifdef UNDO_DEBUG
		dprint( "execute_redo(). mem.error. data_size: %d\n", (int)u->data_size );
#endif
		undo_reset( u );
		break;
	    }
	}
	else 
	{
	    //Error:
#ifdef UNDO_DEBUG
	    dprint( "execute_redo(). action error. data_size: %d\n", (int)u->data_size );
#endif
	    undo_reset( u );
	    break;
	}
	
	u->cur_action++;
    }
}

//COPY / PASTE:

#ifndef NOCOPYPASTE

#ifdef IPHONE
extern "C" int iphone_sundog_copy( char* filename );
extern "C" char* iphone_sundog_paste( void );
#endif

int system_copy( char* filename, int flags )
{
    int rv = 0;
    UTF8_CHAR* new_filename = v3_make_filename( filename );
    if( new_filename ) filename = new_filename;
#ifdef IPHONE
    rv = iphone_sundog_copy( filename );
#endif
    if( new_filename ) mem_free( new_filename );
    return rv;
}

UTF8_CHAR* system_paste( int flags )
{
    UTF8_CHAR* rv = 0;
#ifdef IPHONE
    rv = iphone_sundog_paste();
    if( rv )
    {
	UTF8_CHAR* name = (UTF8_CHAR*)MEM_NEW2( mem_strlen( rv ) + 1 );
	name[ 0 ] = 0;
	mem_strcat( name, rv );
	free( rv );
	rv = name;
    }
#endif
    return rv;
}

#endif

//URL:

#ifdef IPHONE
extern "C" void iphone_sundog_open_url( const char* url );
extern "C" void iphone_sundog_send_mail( const char* email_text, const char* subj_text, const char* body_text );
#endif

#ifdef ANDROID
#ifndef NOMAIN
extern void android_sundog_open_url( const char* url );
#endif
#endif

void open_url( const UTF8_CHAR* url )
{
#ifdef LINUX
    UTF8_CHAR* ts = (UTF8_CHAR*)MEM_NEW2( mem_strlen( url ) + 256 );
    if( ts )
    {
	sprintf( ts, "xdg-open %s", url );
	if( system( ts ) != 0 )
	{
	    sprintf( ts, "sensible-browser %s", url );
	    system( ts );
	}
	mem_free( ts );
    }
#endif
#ifdef WIN
    ShellExecute( NULL, "open", url, NULL, NULL, SW_SHOWNORMAL );
#endif
#ifdef OSX
    CFURLRef u = CFURLCreateWithBytes (
	NULL,                   // allocator
        (UInt8*)url,     	// URLBytes
        mem_strlen( url ),      // length
        kCFStringEncodingASCII, // encoding
        NULL                    // baseURL
    );
    LSOpenCFURLRef( u, 0 );
    CFRelease( u );
#endif
#ifdef IPHONE
    iphone_sundog_open_url( url );
#endif
#ifdef ANDROID
#ifndef NOMAIN
    android_sundog_open_url( url );
#endif
#endif
}

void send_mail( const UTF8_CHAR* email_text, const UTF8_CHAR* subj_text, const UTF8_CHAR* body_text )
{
#ifdef IPHONE
    iphone_sundog_send_mail( email_text, subj_text, body_text );
#endif
}

//RANDOM GENERATOR:

unsigned int rand_next = 1;

void set_seed( unsigned int seed )
{
    rand_next = seed;
}

unsigned int pseudo_random( void )
{
    rand_next = rand_next * 1103515245 + 12345;
    return ( (unsigned int)( rand_next / 65536 ) % 32768 );
}

unsigned int pseudo_random_with_seed( unsigned int* s )
{
    *s = *s * 1103515245 + 12345;
    return ( (unsigned int)( *s / 65536 ) % 32768 );
}

//INT SQRT:

uint sqrt_newton( uint l )
{
    uint temp, div; 
    uint rslt = l; 
    
    if( l <= 0 ) return 0;
    
    if( l & 0xFFFF0000 )
    {
	if( l & 0xFF000000L )
	    div = 0x3FFF; 
	else
	    div = 0x3FF; 
    }
    else 
    {
    	if( l & 0x0FF00 ) 
    	    div = 0x3F; 
	else
	    div = ( l > 4 ) ? 0x7 : l; 
    }
    
    while( l )
    {
	temp = l / div + div; 
	div = temp >> 1; 
	div += temp & 1; 
	if( rslt > div ) 
	{
	    rslt = (unsigned)div; 
	}
	else 
	{
	    if( l / rslt == rslt - 1 && l % rslt == 0 )
		rslt--; 
            return rslt; 
        }
    }
    
    return 0;
}

//VARIOUS:

int round_to_power_of_two( int v )
{
    int b = 1;
    while( 1 )
    {
        if( b >= v )
        {
	    break;
	}
	b <<= 1;
    }
    return b;
}
