#ifndef __UTILS_H__
#define __UTILS_H__

#ifdef UNIX
    #include <pthread.h>
#endif

#if defined(WIN) || defined(WINCE)
    #include <windows.h>
#endif

//LIST:

struct list_data
{
    UTF8_CHAR* items; //Items data
    int* items_ptr; //Item pointers
    int items_num; //Number of items; 
    int selected_item;
    int start_item;
};

void list_init( list_data* data );
void list_close( list_data* data );
void list_clear( list_data* data );
void list_reset_items( list_data* data );
void list_reset_selection( list_data* data );
void list_add_item( const UTF8_CHAR* item, char attr, list_data* data );
void list_delete_item( int item_num, list_data* data );
void list_move_item_up( int item_num, list_data* data );
void list_move_item_down( int item_num, list_data* data );
UTF8_CHAR* list_get_item( int item_num, list_data* data );
char list_get_attr( int item_num, list_data* data );
int list_get_selected_num( list_data* data );
void list_set_selected_num( int sel, list_data* data );
int list_compare_items( int item1, int item2, list_data* data );
void list_sort( list_data* data );

//MUTEXES:

struct sundog_mutex
{
#ifdef UNIX
    pthread_mutex_t mutex;
#endif
#if defined(WIN) || defined(WINCE)
    CRITICAL_SECTION mutex;
#endif
#ifdef PALMOS
    volatile int mutex_cnt;
    volatile bool* main_sound_callback_working;
#endif
};

// mutex attributes:
// 0 - recursive.

int sundog_mutex_init( sundog_mutex* mutex, int attr ); 
int sundog_mutex_destroy( sundog_mutex* mutex );
int sundog_mutex_lock( sundog_mutex* mutex );
int sundog_mutex_trylock( sundog_mutex* mutex );
int sundog_mutex_unlock( sundog_mutex* mutex );

//THREADS:

struct sundog_thread
{
#ifdef UNIX
    pthread_t th;
#endif
#if defined(WIN) || defined(WINCE)
    HANDLE th;
#endif
    void* arg;
    void* (*proc)(void*);
    volatile bool finished;
};

int sundog_thread_create( sundog_thread* th, void* (*proc)(void*), void* arg, int flags );
int sundog_thread_destroy( sundog_thread* th, int timeout );

//PROFILES:

#define KEY_SCREENX		"width"
#define KEY_SCREENY		"height"
#define KEY_ROTATE		"rotate"
#define KEY_FULLSCREEN		"fullscreen"
#define KEY_ZOOM		"zoom"
#define KEY_PPI			"ppi"
#define KEY_SCALE		"scale"
#define KEY_NOCURSOR		"nocursor"
#define KEY_NOAUTOLOCK		"noautolock"
#define KEY_NOBORDER		"noborder"
#define KEY_WINDOWNAME		"windowname"
#define KEY_VIDEODRIVER		"videodriver"
#define KEY_TOUCHCONTROL	"touchcontrol"
#define KEY_PENCONTROL		"pencontrol"
#define KEY_MAXFPS              "maxfps"
#define KEY_VSYNC		"vsync"
#define KEY_SHOW_VIRT_KBD	"show_virt_kbd"
#define KEY_HIDE_VIRT_KBD	"hide_virt_kbd"
#define KEY_COLOR_THEME		"builtin_theme"
#define KEY_SOUNDBUFFER		"buffer"
#define KEY_AUDIODEVICE		"audiodevice"
#define KEY_AUDIODEVICE_IN	"audiodevice_in"
#define KEY_AUDIODRIVER		"audiodriver"
#define KEY_FREQ		"frequency"

struct profile_key
{
    UTF8_CHAR* key;
    UTF8_CHAR* value;
    int line_num;
    bool deleted;
};

struct profile_data
{
    int file_num;
    UTF8_CHAR* file_name;
    UTF8_CHAR* source;
    profile_key* keys;
    int num;
};

void profile_new( profile_data* p );
int profile_resize( int new_num, profile_data* p );
int profile_add_key( const UTF8_CHAR* key, const UTF8_CHAR* value, int line_num, profile_data* p );
void profile_remove_key( const UTF8_CHAR* key, profile_data* p );
void profile_set_str_value( const UTF8_CHAR* key, const UTF8_CHAR* value, profile_data* p );
void profile_set_int_value( const UTF8_CHAR* key, int value, profile_data* p );
int profile_get_int_value( const UTF8_CHAR* key, int default_value, profile_data* p );
UTF8_CHAR* profile_get_str_value( const UTF8_CHAR* key, const UTF8_CHAR* default_value, profile_data* p );
void profile_close( profile_data* p );
void profile_load( const UTF8_CHAR* filename, profile_data* p );
int profile_save( profile_data* p );

//STRINGS:

void int_to_string( int value, UTF8_CHAR* str );
void int_to_string_ext( int value, UTF8_CHAR* str, int chars );
void hex_int_to_string( uint value, UTF8_CHAR* str );
void hex_int_to_string_ext( uint value, UTF8_CHAR* str, int chars );
int string_to_int( const UTF8_CHAR* str );
int hex_string_to_int( const UTF8_CHAR* str );
char int_to_hchar( int value );

UTF16_CHAR* utf8_to_utf16( UTF16_CHAR* dest, int dest_chars, const UTF8_CHAR* s );
UTF32_CHAR* utf8_to_utf32( UTF32_CHAR* dest, int dest_chars, const UTF8_CHAR* s );
UTF8_CHAR* utf16_to_utf8( UTF8_CHAR* dst, int dest_chars, const UTF16_CHAR* src );
UTF8_CHAR* utf32_to_utf8( UTF8_CHAR* dst, int dest_chars, const UTF32_CHAR* src );

int utf8_to_utf32_char( const UTF8_CHAR* str, UTF32_CHAR* res );
int utf8_to_utf32_char_safe( UTF8_CHAR* str, size_t str_size, UTF32_CHAR* res );

void utf8_unix_slash_to_windows( UTF8_CHAR* str );
void utf16_unix_slash_to_windows( UTF16_CHAR* str );
void utf32_unix_slash_to_windows( UTF32_CHAR* str );

//UNDO MANAGER:

#define UNDO_HANDLER_PARS bool redo, undo_action* action, undo_data* u

enum 
{
    UNDO_STATUS_NONE = 0,
    UNDO_STATUS_ADD_ACTION,
};

struct undo_action
{
    size_t level;
    uint type; //Action type
    uint par[ 5 ]; //User defined parameters...
    void* data;
};

struct undo_data
{
    int status;
    
    size_t data_size;
    size_t data_size_limit;
    
    size_t actions_num_limit;
    
    size_t level;
    size_t first_action;
    size_t cur_action; //Relative to first_action
    size_t actions_num;
    undo_action* actions;
    int (*action_handler)( UNDO_HANDLER_PARS );
    
    void* user_data;
};

void undo_init( size_t size_limit, int (*action_handler)( UNDO_HANDLER_PARS ), void* user_data, undo_data* u );
void undo_deinit( undo_data* u );
void undo_reset( undo_data* u );
int undo_add_action( undo_action* action, undo_data* u );
void undo_next_level( undo_data* u );
void execute_undo( undo_data* u );
void execute_redo( undo_data* u );

//COPY / PASTE:

//Content types:
#define CT_AUDIO 1
#define CT_TEXT 2

int system_copy( char* filename, int flags );
UTF8_CHAR* system_paste( int flags );

//URL:

void open_url( const UTF8_CHAR* url );
void send_mail( const UTF8_CHAR* email_text, const UTF8_CHAR* subj_text, const UTF8_CHAR* body_text );

//RANDOM GENERATOR:

void set_seed( unsigned int seed );
unsigned int pseudo_random( void );
unsigned int pseudo_random_with_seed( unsigned int* s );

//INT SQRT:

uint sqrt_newton( uint l );

//VARIOUS:

int round_to_power_of_two( int v );

#define INT32_SWAP(n) ( ( (((unsigned int) n) << 24 ) & 0xFF000000 ) | \
                        ( (((unsigned int) n) << 8 ) & 0x00FF0000 ) | \
                        ( (((unsigned int) n) >> 8 ) & 0x0000FF00 ) | \
                        ( (((unsigned int) n) >> 24 ) & 0x000000FF ) )

#define INT16_SWAP(n) ( ( (((unsigned short) n) << 8 ) & 0xFF00 ) | \
                        ( (((unsigned short) n) >> 8 ) & 0x00FF ) )

#endif
