#include <jni.h>
#include <errno.h>
#include <pthread.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/stat.h>

#include <EGL/egl.h>
#include <GLES/gl.h>
#include <GLES/glext.h>

#include <android/sensor.h>
#include <android/log.h>
#include <android_native_app_glue.h>

#include "core/core.h"
#include "time/timemanager.h"
#include "window_manager/wmanager.h"

#define LOGI(...) ((void)__android_log_print(ANDROID_LOG_INFO, "native-activity", __VA_ARGS__))
#define LOGW(...) ((void)__android_log_print(ANDROID_LOG_WARN, "native-activity", __VA_ARGS__))

#define MAX_TOUCHES 10

int engine_init_display( void );
void engine_term_display( void );

extern int sundog_main( int argc, char* argv[] );
extern void android_sundog_exit_request( void );
extern void android_sundog_touches( int x, int y, bool dont_draw, int mode, int touch_id );
extern void android_sundog_key_event( bool down, uint16 key, uint16 scancode, uint16 flags );

extern char* g_android_cache_int_path;
extern char* g_android_cache_ext_path;
extern char* g_android_files_int_path;
extern char* g_android_files_ext_path;

int g_android_sundog_screen_xsize = 0;
int g_android_sundog_screen_ysize = 0;
int g_android_sundog_screen_ppi = 0;
bool g_android_sundog_gl_buffer_preserved = 0;
void* g_android_sundog_framebuffer = 0;
volatile int g_android_sundog_resize_request = 0;
volatile int g_android_sundog_resize_request_rp = 0;
EGLDisplay g_android_sundog_display;
EGLSurface g_android_sundog_surface;
EGLContext g_android_sundog_context;

struct sundog_info_struct
{
    int initialized;
    pthread_t pth;
    bool pth_finished;
    bool exit_request;
    int32_t touch_id[ MAX_TOUCHES ];
    bool touch_busy[ MAX_TOUCHES ];
};

// Shared state for our app.
struct engine 
{
    struct android_app* app;

    ASensorManager* sensorManager;
    const ASensor* accelerometerSensor;
    ASensorEventQueue* sensorEventQueue;

    EGLDisplay display;
    EGLSurface surface;
    EGLContext context;
    
    sundog_info_struct sd;
};

engine g_engine;

volatile bool g_eventloop_stop_request = 0;
volatile bool g_eventloop_stop_answer = 0;

int call_int_java_method( const char* method_name, const char* str_par )
{
    int rv = -1;
    
    JNIEnv* jni;
    g_engine.app->activity->vm->AttachCurrentThread( &jni, NULL );
    jclass na_class = jni->FindClass( "android/app/NativeActivity" );
    if( na_class == 0 )
    {
        LOGW( "NativeActivity class not found!" );
    }
    else
    {
        jmethodID getClassLoader = jni->GetMethodID( na_class, "getClassLoader", "()Ljava/lang/ClassLoader;" );
        jobject class_loader_obj = jni->CallObjectMethod( g_engine.app->activity->clazz, getClassLoader );
        jclass class_loader = jni->FindClass( "java/lang/ClassLoader" );
        jmethodID loadClass = jni->GetMethodID( class_loader, "loadClass", "(Ljava/lang/String;)Ljava/lang/Class;" );
        jstring class_name = jni->NewStringUTF( PKGNAME "/MyNativeActivity" );
        jclass my_class = (jclass)jni->CallObjectMethod( class_loader_obj, loadClass, class_name );
        
        jmethodID m;
        if( str_par )
    	    m = jni->GetMethodID( my_class, method_name, "(Ljava/lang/String;)I" );
    	else
    	    m = jni->GetMethodID( my_class, method_name, "()I" );
	if( m )
    	{
    	    if( str_par )
		rv = jni->CallIntMethod( g_engine.app->activity->clazz, m, jni->NewStringUTF( str_par ) );
	    else
		rv = jni->CallIntMethod( g_engine.app->activity->clazz, m );
    	}
    }
    g_engine.app->activity->vm->DetachCurrentThread();
    
    return rv;
}

void* sundog_thread( void* arg )
{
    sundog_info_struct* sd = &g_engine.sd;
    
    LOGI( "sundog_thread() ..." );

    if( engine_init_display() ) exit( -1 );
    call_int_java_method( "WakeLock", 0 );

    LOGI( "sundog_main() ..." );

    sundog_main( 0, 0 );

    LOGI( "sundog_main() finished" );
    
    call_int_java_method( "WakeUnlock", 0 );
    engine_term_display();
    
    if( sd->exit_request == 0 )
    {
	exit( 0 );
    }

    LOGI( "sundog_thread() finished" );

    sd->pth_finished = 1;
    pthread_exit( NULL );
}

void sundog_init( void )
{
    int err;
    sundog_info_struct* sd = &g_engine.sd;

    LOGI( "sundog_init() ..." );

    if( sd->initialized == 1 ) return;
    
    g_android_sundog_resize_request = 0;
    g_android_sundog_resize_request_rp = 0;
    
    g_eventloop_stop_request = 0;
    g_eventloop_stop_answer = 0;

    memset( sd, 0, sizeof( sundog_info_struct ) );
 
    err = pthread_create( &sd->pth, 0, &sundog_thread, 0 );
    if( err == 0 )
    {
	//The pthread_detach() function marks the thread identified by thread as
	//detached. When a detached thread terminates, its resources are 
	//automatically released back to the system.
	err = pthread_detach( sd->pth );
	if( err != 0 )
	{
	    LOGW( "sundog_init(): pthread_detach error %d", err );
	    return;
	}
    }
    else
    {
	LOGW( "sundog_init(): pthread_create error %d", err );
	return;
    }
    
    LOGI( "sundog_init(): done" );
    
    sd->initialized = 1;
}

void sundog_deinit( void )
{
    int err;
    sundog_info_struct* sd = &g_engine.sd;
    
    if( sd->initialized == 0 ) return;

    LOGI( "sundog_deinit() ..." );

    //Stop the thread:
    sd->exit_request = 1;
    android_sundog_exit_request();
    g_eventloop_stop_request = 0; //Resume SunDog eventloop if it was in Pause state
    int timeout_counter = 1000 * 6; //Timeout in milliseconds
    while( timeout_counter > 0 )
    {
	struct timespec delay;
	delay.tv_sec = 0;
	delay.tv_nsec = 1000000 * 20; //20 milliseconds
	if( sd->pth_finished ) break;
	nanosleep( &delay, NULL ); //Sleep for delay time
	timeout_counter -= 20;
    }
    if( timeout_counter <= 0 )
    {
	LOGW( "sundog_deinit(): thread timeout" );
    }
    
    LOGI( "sundog_deinit(): done" );

    sd->initialized = 0;
}

void android_sundog_screen_redraw( void )
{
    eglSwapBuffers( g_engine.display, g_engine.surface );
}

void android_sundog_event_handler( void )
{
    HANDLE_THREAD_EVENTS;
    if( g_eventloop_stop_request )
    {
        g_eventloop_stop_answer = 1;
        while( g_eventloop_stop_request )
        {
            time_sleep( 100 );
        }
        g_eventloop_stop_answer = 0;
    }
}

int android_sundog_copy_resources( void )
{
    return call_int_java_method( "CopyResources", 0 );
}

void android_sundog_open_url( const char* url_text )
{
    call_int_java_method( "OpenURL", url_text );
}

// Initialize an EGL context for the current display.
int engine_init_display( void ) 
{
    // initialize OpenGL ES and EGL
    
    bool buf_preserved;
#ifdef GLNORETAIN
    buf_preserved = 0;
#else
    buf_preserved = 1;
#endif
    char* ts = (char*)malloc( strlen( g_android_files_ext_path ) + 64 );
    if( ts )
    {
	ts[ 0 ] = 0;
	strcat( ts, g_android_files_ext_path );
	strcat( ts, "option_glnoretain" );
	FILE* f = fopen( ts, "rb" );
	if( f == 0 )
	{
	    ts[ 0 ] = 0;
	    strcat( ts, g_android_files_ext_path );
	    strcat( ts, "option_glnoretain.txt" );
	    f = fopen( ts, "rb" );
	}
	if( f )
	{
	    LOGI( "option_glnoretain found. OpenGL Buffer Preserved = NO" );
	    buf_preserved = 0;	
	    fclose( f );
	}
	free( ts );
    }

    /*
     * Here specify the attributes of the desired configuration.
     * Below, we select an EGLConfig with at least 8 bits per color
     * component compatible with on-screen windows
     */
    const EGLint attribs1[] = 
    {
	EGL_SURFACE_TYPE, EGL_WINDOW_BIT | EGL_SWAP_BEHAVIOR_PRESERVED_BIT,
	EGL_DEPTH_SIZE, 16,
        EGL_NONE
    };
    const EGLint attribs2[] = 
    {
	EGL_SURFACE_TYPE, EGL_WINDOW_BIT,
	EGL_DEPTH_SIZE, 16,
        EGL_NONE
    };
    EGLint w, h, dummy, format;
    EGLint numConfigs = 0;
    EGLint num = 0;
    EGLConfig config[ 32 ];
    EGLSurface surface = EGL_NO_SURFACE;
    EGLContext context = EGL_NO_CONTEXT;

    g_engine.display = eglGetDisplay( EGL_DEFAULT_DISPLAY );
    if( g_engine.display == EGL_NO_DISPLAY )
    {
	LOGW( "eglGetDisplay() error %d", eglGetError() );
	return -1;
    }

    if( eglInitialize( g_engine.display, 0, 0 ) != EGL_TRUE )
    {
	LOGW( "eglInitialize() error %d", eglGetError() );
    	return -1;
    }

    if( buf_preserved )
    {
	if( eglChooseConfig( g_engine.display, attribs1, &config[ numConfigs ], 1, &num ) != EGL_TRUE )
	{
    	    LOGW( "eglChooseConfig() error %d", eglGetError() );
	}
	LOGI( "EGL Configs with EGL_SWAP_BEHAVIOR_PRESERVED_BIT: %d", num );
	numConfigs += num;
    }
    num = 0;
    if( eglChooseConfig( g_engine.display, attribs2, &config[ numConfigs ], 1, &num ) != EGL_TRUE )
    {
	LOGW( "eglChooseConfig() error %d", eglGetError() );
    }    
    LOGI( "EGL Configs without EGL_SWAP_BEHAVIOR_PRESERVED_BIT: %d", num );
    numConfigs += num;
    
    if( numConfigs == 0 )
    {
	LOGW( "ERROR: No matching configs (numConfigs == 0)" );
	return -1;	
    }    

    for( EGLint cnum = 0; cnum < numConfigs; cnum++ )
    {
	//Get config info:
	EGLint cv, r, g, b;
	eglGetConfigAttrib( g_engine.display, config[ cnum ], EGL_RED_SIZE, &r );
	eglGetConfigAttrib( g_engine.display, config[ cnum ], EGL_GREEN_SIZE, &g );
	eglGetConfigAttrib( g_engine.display, config[ cnum ], EGL_BLUE_SIZE, &b );
	eglGetConfigAttrib( g_engine.display, config[ cnum ], EGL_SURFACE_TYPE, &cv );
	char stype[ 256 ];
	stype[ 0 ] = 0;
	if( cv & EGL_PBUFFER_BIT ) strcat( stype, "PBUF " );
	if( cv & EGL_PIXMAP_BIT ) strcat( stype, "PIXMAP " );
	if( cv & EGL_WINDOW_BIT ) strcat( stype, "WIN " );
	if( cv & EGL_VG_COLORSPACE_LINEAR_BIT ) strcat( stype, "CL " );
	if( cv & EGL_VG_ALPHA_FORMAT_PRE_BIT ) strcat( stype, "AFP " );
	if( cv & EGL_MULTISAMPLE_RESOLVE_BOX_BIT ) strcat( stype, "MRB " );
	if( cv & EGL_SWAP_BEHAVIOR_PRESERVED_BIT ) strcat( stype, "PRESERVED " );
	LOGI( "EGL Config %d. RGBT: %d %d %d %x %s", cnum, r, g, b, cv, stype );
    }
    
    bool configErr = 1;
    for( EGLint cnum = 0; cnum < numConfigs; cnum++ )
    {
	LOGI( "EGL Config %d", cnum );
	
	if( surface != EGL_NO_SURFACE ) eglDestroySurface( g_engine.display, surface );
	if( context != EGL_NO_CONTEXT ) eglDestroyContext( g_engine.display, surface );
	surface = EGL_NO_SURFACE;
	context = EGL_NO_CONTEXT;

	/* EGL_NATIVE_VISUAL_ID is an attribute of the EGLConfig that is
         * guaranteed to be accepted by ANativeWindow_setBuffersGeometry().
         * As soon as we picked a EGLConfig, we can safely reconfigure the
         * ANativeWindow buffers to match, using EGL_NATIVE_VISUAL_ID. */
        if( eglGetConfigAttrib( g_engine.display, config[ cnum ], EGL_NATIVE_VISUAL_ID, &format ) != EGL_TRUE )
        {
    	    LOGW( "eglGetConfigAttrib() error %d", eglGetError() );
	    continue;
	}
	
        ANativeWindow_setBuffersGeometry( g_engine.app->window, 0, 0, format );
	
        surface = eglCreateWindowSurface( g_engine.display, config[ cnum ], g_engine.app->window, NULL );
	if( surface == EGL_NO_SURFACE )
        {
    	    LOGW( "eglCreateWindowSurface() error %d", eglGetError() );
	    continue;
	}
	context = eglCreateContext( g_engine.display, config[ cnum ], NULL, NULL );
	if( surface == EGL_NO_CONTEXT )
        {
	    LOGW( "eglCreateContext() error %d", eglGetError() );
	    continue;
	}

	if( eglMakeCurrent( g_engine.display, surface, surface, context ) != EGL_TRUE ) 
	{
    	    LOGW( "eglMakeCurrent error %d", eglGetError() );
    	    continue;
	}

	if( buf_preserved )
	{
	    EGLint sb;
    	    if( eglSurfaceAttrib( g_engine.display, surface, EGL_SWAP_BEHAVIOR, EGL_BUFFER_PRESERVED ) != EGL_TRUE )
    	    {
    		LOGW( "eglSurfaceAttrib error %d", eglGetError() );
    	    }
    	    if( eglQuerySurface( g_engine.display, surface, EGL_SWAP_BEHAVIOR, &sb ) != EGL_TRUE )
    	    {
    		LOGW( "eglQuerySurface error %d", eglGetError() );
    	    }
    	    if( sb == EGL_BUFFER_PRESERVED )
	    {
		g_android_sundog_gl_buffer_preserved = 1;
		LOGI( "EGL_BUFFER_PRESERVED" );
    	    }
	    else
    	    {
    	        g_android_sundog_gl_buffer_preserved = 0;
    		LOGI( "EGL_BUFFER_DESTROYED" );
	    }
	}
	else
	{
    	    g_android_sundog_gl_buffer_preserved = 0;
    	}

	configErr = 0;
	break;
    }
    if( configErr )
    {
	LOGW( "ERROR: No matching configs." );
	return -1;
    }    

    eglQuerySurface( g_engine.display, surface, EGL_WIDTH, &w );
    eglQuerySurface( g_engine.display, surface, EGL_HEIGHT, &h );

    g_engine.context = context;    
    g_engine.surface = surface;
    g_android_sundog_display = g_engine.display;
    g_android_sundog_surface = surface;
    g_android_sundog_context = context;
    
    g_android_sundog_screen_xsize = w;
    g_android_sundog_screen_ysize = h;

    // Initialize GL state.
    glHint( GL_PERSPECTIVE_CORRECTION_HINT, GL_FASTEST );
    glShadeModel( GL_SMOOTH );
    glDisable( GL_DEPTH_TEST );
#ifndef GLNORETAIN
    if( g_android_sundog_gl_buffer_preserved == 0 )
    {
	glDisable( GL_DITHER );
    }
#endif

    return 0;
}

// Tear down the EGL context currently associated with the display.
void engine_term_display( void ) 
{
    if( g_engine.display != EGL_NO_DISPLAY ) 
    {
        eglMakeCurrent( g_engine.display, EGL_NO_SURFACE, EGL_NO_SURFACE, EGL_NO_CONTEXT );
        if( g_engine.context != EGL_NO_CONTEXT ) 
        {
            eglDestroyContext( g_engine.display, g_engine.context );
        }
        if( g_engine.surface != EGL_NO_SURFACE ) 
        {
            eglDestroySurface( g_engine.display, g_engine.surface );
        }
        eglTerminate( g_engine.display );
    }
    g_engine.display = EGL_NO_DISPLAY;
    g_engine.context = EGL_NO_CONTEXT;
    g_engine.surface = EGL_NO_SURFACE;
}

uint16 convert_key( int32_t vk )
{
    uint16 rv = 0xFFFF;
    
    switch( vk )
    {
	case AKEYCODE_BACK: rv = KEY_ESCAPE; break;
	case AKEYCODE_MENU: rv = KEY_MENU; break;

	case AKEYCODE_SOFT_LEFT: rv = KEY_LEFT; break;
	case AKEYCODE_SOFT_RIGHT: rv = KEY_RIGHT; break;
	case AKEYCODE_0: case 0x90: rv = '0'; break;
	case AKEYCODE_1: case 0x91: rv = '1'; break;
	case AKEYCODE_2: case 0x92: rv = '2'; break;
        case AKEYCODE_3: case 0x93: rv = '3'; break;
	case AKEYCODE_4: case 0x94: rv = '4'; break;
	case AKEYCODE_5: case 0x95: rv = '5'; break;
	case AKEYCODE_6: case 0x96: rv = '6'; break;
	case AKEYCODE_7: case 0x97: rv = '7'; break;
        case AKEYCODE_8: case 0x98: rv = '8'; break;
	case AKEYCODE_9: case 0x99: rv = '9'; break;
        case AKEYCODE_STAR: rv = '*'; break;
        case AKEYCODE_POUND: rv = '#'; break;
	case AKEYCODE_DPAD_UP: rv = KEY_UP; break;
	case AKEYCODE_DPAD_DOWN: rv = KEY_DOWN; break;
	case AKEYCODE_DPAD_LEFT: rv = KEY_LEFT; break;
	case AKEYCODE_DPAD_RIGHT: rv = KEY_RIGHT; break;
	case AKEYCODE_DPAD_CENTER: rv = KEY_SPACE; break;
	case AKEYCODE_A: rv = 'a'; break;
	case AKEYCODE_B: rv = 'b'; break;
	case AKEYCODE_C: rv = 'c'; break;
	case AKEYCODE_D: rv = 'd'; break;
        case AKEYCODE_E: rv = 'e'; break;
	case AKEYCODE_F: rv = 'f'; break;
	case AKEYCODE_G: rv = 'g'; break;
        case AKEYCODE_H: rv = 'h'; break;
	case AKEYCODE_I: rv = 'i'; break;
        case AKEYCODE_J: rv = 'j'; break;
	case AKEYCODE_K: rv = 'k'; break;
	case AKEYCODE_L: rv = 'l'; break;
        case AKEYCODE_M: rv = 'm'; break;
	case AKEYCODE_N: rv = 'n'; break;
        case AKEYCODE_O: rv = 'o'; break;
	case AKEYCODE_P: rv = 'p'; break;
	case AKEYCODE_Q: rv = 'q'; break;
        case AKEYCODE_R: rv = 'r'; break;
	case AKEYCODE_S: rv = 's'; break;
	case AKEYCODE_T: rv = 't'; break;
	case AKEYCODE_U: rv = 'u'; break;
	case AKEYCODE_V: rv = 'v'; break;
	case AKEYCODE_W: rv = 'w'; break;
        case AKEYCODE_X: rv = 'x'; break;
        case AKEYCODE_Y: rv = 'y'; break;
        case AKEYCODE_Z: rv = 'z'; break;
        case AKEYCODE_COMMA: rv = ','; break;
        case AKEYCODE_PERIOD: rv = '.'; break;
	case AKEYCODE_ALT_LEFT: rv = KEY_ALT; break;
	case AKEYCODE_ALT_RIGHT: rv = KEY_ALT; break;
	case AKEYCODE_SHIFT_LEFT: rv = KEY_SHIFT; break;
        case AKEYCODE_SHIFT_RIGHT: rv = KEY_SHIFT; break;
        case AKEYCODE_TAB: rv = KEY_TAB; break;
        case AKEYCODE_SPACE: rv = ' '; break;
	case AKEYCODE_ENTER: rv = KEY_ENTER; break;
        case AKEYCODE_DEL: rv = KEY_BACKSPACE; break;
        //case 112: rv = KEY_DELETE; break; //AKEYCODE_FORWARD_DEL
        case AKEYCODE_GRAVE: rv = '`'; break;
	case AKEYCODE_MINUS: rv = '-'; break;
        case AKEYCODE_EQUALS: rv = '='; break;
        case AKEYCODE_LEFT_BRACKET: rv = '['; break;
	case AKEYCODE_RIGHT_BRACKET: rv = ']'; break;
	case AKEYCODE_BACKSLASH: rv = '\\'; break;
	case AKEYCODE_SEMICOLON: rv = ';'; break;
	case AKEYCODE_APOSTROPHE: rv = '\''; break;
	case AKEYCODE_SLASH: rv = '/'; break;
	case AKEYCODE_AT: rv = '@'; break;
	case AKEYCODE_PLUS: rv = '+'; break;
	case AKEYCODE_PAGE_UP: rv = KEY_PAGEUP; break;
	case AKEYCODE_PAGE_DOWN: rv = KEY_PAGEDOWN; break;
	case AKEYCODE_BUTTON_A: rv = 'a'; break;
	case AKEYCODE_BUTTON_B: rv = 'b'; break;
        case AKEYCODE_BUTTON_C: rv = 'c'; break;
        case AKEYCODE_BUTTON_X: rv = 'x'; break;
        case AKEYCODE_BUTTON_Y: rv = 'y'; break;
        case AKEYCODE_BUTTON_Z: rv = 'z'; break;
    }
    
    if( rv == 0xFFFF ) rv = KEY_UNKNOWN + vk;
    return rv;
}

// Process the next input event.
static int32_t engine_handle_input( struct android_app* app, AInputEvent* event ) 
{
    int32_t rv = 0;
    sundog_info_struct* sd = &g_engine.sd;
    int x, y;
    int32_t id;
    int32_t evt_type = AInputEvent_getType( event );
    if( evt_type == AINPUT_EVENT_TYPE_KEY )
    {
	int32_t sc = AKeyEvent_getScanCode( event );
	int32_t vk = AKeyEvent_getKeyCode( event );
	uint16 sundog_key = convert_key( vk );
	int32_t ms = AKeyEvent_getMetaState( event );
	uint16 sundog_flags = 0;
	if( ms & AMETA_ALT_ON ) sundog_flags |= EVT_FLAG_ALT;
	if( ms & AMETA_SHIFT_ON ) sundog_flags |= EVT_FLAG_SHIFT;
	if( ms & 4096 ) sundog_flags |= EVT_FLAG_CTRL;
	int32_t evt = AKeyEvent_getAction( event );
	switch( evt )
	{
	    case AKEY_EVENT_ACTION_DOWN:
		//LOGI( "AKEY_EVENT_ACTION_DOWN VK:%d SD:%d F:%d", vk, sundog_key, sundog_flags );
		android_sundog_key_event( 1, sundog_key, (uint16)sc, sundog_flags );
		break;
	    case AKEY_EVENT_ACTION_UP:
		//LOGI( "AKEY_EVENT_ACTION_UP VK:%d SD:%d F:%d", vk, sundog_key, sundog_flags );
		android_sundog_key_event( 0, sundog_key, (uint16)sc, sundog_flags );
		break;
	}
	switch( vk )
	{
	    case AKEYCODE_BACK:
	    case AKEYCODE_MENU:
		rv = 1; //prevent default handler
		break;
	}
    }
    if( evt_type == AINPUT_EVENT_TYPE_MOTION )
    {
	size_t pcount = AMotionEvent_getPointerCount( event );
	int32_t evt = AMotionEvent_getAction( event );
	switch( evt & AMOTION_EVENT_ACTION_MASK )
	{
	    case AMOTION_EVENT_ACTION_DOWN:
		//First touch:
		memset( sd->touch_id, 0, sizeof( sd->touch_id ) );
		memset( sd->touch_busy, 0, sizeof( sd->touch_busy ) );
		x = AMotionEvent_getX( event, 0 );
		y = AMotionEvent_getY( event, 0 );
		id = AMotionEvent_getPointerId( event, 0 );
		sd->touch_id[ 0 ] = id;
		sd->touch_busy[ 0 ] = 1;
		android_sundog_touches( x, y, 0, 0, 0 );
		android_sundog_touches( 0, 0, 0, -1, 0 );
		rv = 1;
		break;
	    case AMOTION_EVENT_ACTION_MOVE:
		{
		    for( size_t i = 0; i < pcount; i++ )
    		    {
			id = AMotionEvent_getPointerId( event, i );
			x = AMotionEvent_getX( event, i );
			y = AMotionEvent_getY( event, i );
			for( int t = 0; t < MAX_TOUCHES; t++ )
			{
			    if( sd->touch_busy[ t ] && sd->touch_id[ t ] == id )
			    {
				bool dont_draw = 0;
				if( i < pcount - 1 ) dont_draw = 1;
				android_sundog_touches( x, y, dont_draw, 1, t );
				android_sundog_touches( 0, 0, dont_draw, -1, t );
				break;
			    }
			}
    		    }
		}
		rv = 1;
		break;
	    case AMOTION_EVENT_ACTION_UP:
	    case AMOTION_EVENT_ACTION_CANCEL:
		{
		    for( size_t i = 0; i < pcount; i++ )
    		    {
			id = AMotionEvent_getPointerId( event, i );
			x = AMotionEvent_getX( event, i );
			y = AMotionEvent_getY( event, i );
			for( int t = 0; t < MAX_TOUCHES; t++ )
			{
			    if( sd->touch_busy[ t ] && sd->touch_id[ t ] == id )
			    {
				android_sundog_touches( x, y, 0, 2, t );
				android_sundog_touches( 0, 0, 0, -1, t );
				break;
			    }
			}
    		    }
		}
		rv = 1;
		break;
	    case AMOTION_EVENT_ACTION_POINTER_DOWN:
		{
		    int i = ( evt & AMOTION_EVENT_ACTION_POINTER_INDEX_MASK ) >> AMOTION_EVENT_ACTION_POINTER_INDEX_SHIFT;
		    id = AMotionEvent_getPointerId( event, i );
		    x = AMotionEvent_getX( event, i );
		    y = AMotionEvent_getY( event, i );
		    for( int t = 0; t < MAX_TOUCHES; t++ )
		    {
			if( sd->touch_busy[ t ] == 0 )
			{
			    sd->touch_busy[ t ] = 1;
			    sd->touch_id[ t ] = id;
		    	    android_sundog_touches( x, y, 0, 0, t );
			    android_sundog_touches( 0, 0, 0, -1, t );
			    break;
			}
		    }
		}
		rv = 1;
		break;
	    case AMOTION_EVENT_ACTION_POINTER_UP:
		{
		    int i = ( evt & AMOTION_EVENT_ACTION_POINTER_INDEX_MASK ) >> AMOTION_EVENT_ACTION_POINTER_INDEX_SHIFT;
		    id = AMotionEvent_getPointerId( event, i );
		    x = AMotionEvent_getX( event, i );
		    y = AMotionEvent_getY( event, i );
		    for( int t = 0; t < MAX_TOUCHES; t++ )
		    {
			if( sd->touch_busy[ t ] && sd->touch_id[ t ] == id )
			{
			    sd->touch_busy[ t ] = 0;
		    	    android_sundog_touches( x, y, 0, 2, t );
			    android_sundog_touches( 0, 0, 0, -1, t );
			    break;
			}
		    }
		}
		rv = 1;
		break;
	}
    }
    return rv;
}

// Process the next main command.
static void engine_handle_cmd( struct android_app* app, int32_t cmd ) 
{
    switch( cmd ) 
    {
	case APP_CMD_PAUSE:
	    if( g_engine.sd.initialized )
	    {
		g_eventloop_stop_request = 1;
		int i = 0;
		int timeout = 1000;
		int step = 20;
		for( i = 0; i < timeout; i += step )
		{
		    if( g_eventloop_stop_answer == 1 ) break;
		    time_sleep( step );
		}
		if( i >= timeout )
		{
		    LOGI( "SunDog PAUSE TIMEOUT" );
		}
	    }
	    break;
        case APP_CMD_STOP:
    	    LOGI( "SunDog STOP" );
            sundog_deinit();
    	    break;
	case APP_CMD_RESUME:
	    if( g_engine.sd.initialized )
	    {
		g_eventloop_stop_request = 0;
		int i = 0;
		int timeout = 1000;
		int step = 20;
		for( i = 0; i < timeout; i += step )
		{
		    if( g_eventloop_stop_answer == 0 ) break;
		    time_sleep( step );
		}
		if( i >= timeout )
		{
		    LOGI( "SunDog RESUME TIMEOUT" );
		}
	    }
	    break;
        case APP_CMD_START:
    	    break;
    	    
        case APP_CMD_SAVE_STATE:
            break;
        case APP_CMD_INIT_WINDOW:
            // The window is being shown, get it ready.
            if( g_engine.app->window != NULL ) 
	    {
		sundog_init();
            }
            break;
        case APP_CMD_TERM_WINDOW:
            // The window is being hidden or closed, clean it up.
            sundog_deinit();
            break;
        case APP_CMD_GAINED_FOCUS:
            // When our app gains focus, we start monitoring the accelerometer.
            if( g_engine.accelerometerSensor != NULL) 
	    {
                // We'd like to get 60 events per second (in us).
                //ASensorEventQueue_enableSensor( engine->sensorEventQueue, engine->accelerometerSensor );
                //ASensorEventQueue_setEventRate( engine->sensorEventQueue, engine->accelerometerSensor, (1000L/60)*1000 );
            }
            break;
        case APP_CMD_LOST_FOCUS:
            // When our app loses focus, we stop monitoring the accelerometer.
            // This is to avoid consuming battery while not being used.
            if( g_engine.accelerometerSensor != NULL ) 
	    {
                //ASensorEventQueue_disableSensor( engine->sensorEventQueue, engine->accelerometerSensor );
            }
            break;
	case APP_CMD_CONFIG_CHANGED:
	    g_android_sundog_resize_request++;
	    break;
    }
}

void android_main( struct android_app* state ) 
{
    LOGI( "ANDROID MAIN " ARCH_NAME " " __DATE__ " " __TIME__ );

    // Make sure glue isn't stripped.
    app_dummy();
    
    memset( &g_engine, 0, sizeof( g_engine ) );
    state->userData = &g_engine;
    state->onAppCmd = engine_handle_cmd;
    state->onInputEvent = engine_handle_input;
    g_engine.app = state;

    g_android_sundog_screen_ppi = AConfiguration_getDensity( state->config );

    {
	JNIEnv* jni;
	state->activity->vm->AttachCurrentThread( &jni, NULL );
	jclass na_class = jni->FindClass( "android/app/NativeActivity" ); 
	if( na_class == 0 ) 
	{
	    LOGW( "NativeActivity class not found!" );
	    return;
	}
	else
	{
	    jmethodID getClassLoader = jni->GetMethodID( na_class, "getClassLoader", "()Ljava/lang/ClassLoader;" );
	    jobject class_loader_obj = jni->CallObjectMethod( state->activity->clazz, getClassLoader );
	    if( class_loader_obj == 0 )
	    {
		LOGW( "ClassLoader object not found!" );
		return;
	    }
	    jclass class_loader = jni->FindClass( "java/lang/ClassLoader" );
	    if( class_loader == 0 ) 
	    {
	        LOGW( "ClassLoader class not found!" );
	        return;
	    }
	    jmethodID loadClass = jni->GetMethodID( class_loader, "loadClass", "(Ljava/lang/String;)Ljava/lang/Class;" );
	    jstring class_name = jni->NewStringUTF( PKGNAME "/MyNativeActivity" );
	    jclass my_class = (jclass)jni->CallObjectMethod( class_loader_obj, loadClass, class_name );
	    if( my_class == 0 )
	    {
		LOGW( "My Class not found!" );
		return;
	    }

	    jmethodID m = jni->GetMethodID( my_class, "GetInternalCacheDir", "()Ljava/lang/String;" );
	    if( m )
	    {
		jstring s = (jstring)jni->CallObjectMethod( state->activity->clazz, m );
		const char* str = "";
		if( s )
		{
		    str = jni->GetStringUTFChars( s, 0 );
		    mkdir( str, 0770 );
		}
		LOGI( "InternalCacheDir: %s", str );
		size_t len = strlen( str );
		g_android_cache_int_path = (char*)malloc( len + 2 );
		memcpy( g_android_cache_int_path, str, len + 1 );
		strcat( g_android_cache_int_path, "/" );
		if( s )
		{
		    jni->ReleaseStringUTFChars( s, str );  
		}
	    }
	    
	    m = jni->GetMethodID( my_class, "GetInternalFilesDir", "()Ljava/lang/String;" );
	    if( m )
	    {
		jstring s = (jstring)jni->CallObjectMethod( state->activity->clazz, m );
		const char* str = "";
		if( s )
		{
		    str = jni->GetStringUTFChars( s, 0 );
		    mkdir( str, 0770 );
		}
		LOGI( "InternalFilesDir: %s", str );
		size_t len = strlen( str );
		g_android_files_int_path = (char*)malloc( len + 2 );
		memcpy( g_android_files_int_path, str, len + 1 );
		strcat( g_android_files_int_path, "/" );
		if( s )
		{
		    jni->ReleaseStringUTFChars( s, str );  
		}
	    }

	    m = jni->GetMethodID( my_class, "GetExternalCacheDir", "()Ljava/lang/String;" );
	    if( m )
	    {
		jstring s = (jstring)jni->CallObjectMethod( state->activity->clazz, m );
		const char* str = "";
		if( s )
		{
		    str = jni->GetStringUTFChars( s, 0 );
		    mkdir( str, 0777 );
		}
		LOGI( "ExternalCacheDir: %s", str );
		size_t len = strlen( str );
		g_android_cache_ext_path = (char*)malloc( len + 2 );
		memcpy( g_android_cache_ext_path, str, len + 1 );
		strcat( g_android_cache_ext_path, "/" );
		if( s )
		{
		    jni->ReleaseStringUTFChars( s, str );  
		}
	    }

	    m = jni->GetMethodID( my_class, "GetExternalFilesDir", "()Ljava/lang/String;" );
	    if( m )
	    {
		jstring s = (jstring)jni->CallObjectMethod( state->activity->clazz, m );
		const char* str = "";
		if( s )
		{
		    str = jni->GetStringUTFChars( s, 0 );
		    mkdir( str, 0777 );
		}
		LOGI( "ExternalFilesDir: %s", str );
		size_t len = strlen( str );
		g_android_files_ext_path = (char*)malloc( len + 2 );
		memcpy( g_android_files_ext_path, str, len + 1 );
		strcat( g_android_files_ext_path, "/" );
		if( s )
		{
		    jni->ReleaseStringUTFChars( s, str );  
		}
	    }
	}

	state->activity->vm->DetachCurrentThread();
	
    }
    
    // Prepare to monitor accelerometer
    g_engine.sensorManager = ASensorManager_getInstance();
    g_engine.accelerometerSensor = ASensorManager_getDefaultSensor( g_engine.sensorManager, ASENSOR_TYPE_ACCELEROMETER );
    g_engine.sensorEventQueue = ASensorManager_createEventQueue( g_engine.sensorManager, state->looper, LOOPER_ID_USER, NULL, NULL );

    // loop waiting for stuff to do.

    while( 1 ) 
    {
        // Read all pending events.
        int ident;
        int events;
        struct android_poll_source* source;

        // If not animating, we will block forever waiting for events.
        // If animating, we loop until all events are read, then continue
        // to draw the next frame of animation.
        while( ( ident = ALooper_pollAll( -1, NULL, &events, (void**)&source ) ) >= 0 ) 
        {

            // Process this event.
            if( source != NULL ) 
            {
                source->process( state, source );
            }

            // If a sensor has data, process it now.
            if( ident == LOOPER_ID_USER ) 
            {
                if( g_engine.accelerometerSensor != NULL ) 
                {
                    ASensorEvent event;
                    while( ASensorEventQueue_getEvents( g_engine.sensorEventQueue, &event, 1 ) > 0 ) 
                    {
                        //LOGI( "accelerometer: x=%f y=%f z=%f", event.acceleration.x, event.acceleration.y, event.acceleration.z );
                    }
                }
            }

            // Check if we are exiting.
            if( state->destroyRequested != 0 ) 
            {
                sundog_deinit();
                return;
            }
        }
    }
}
