/*
    handlers.cpp. Standart window handlers
    This file is part of the SunDog engine.
    Copyright (C) 2002 - 2013 Alexander Zolotov <nightradio@gmail.com>
    www.warmplace.ru
*/

#include "../../core/core.h"
#include "../../core/debug.h"
#include "../../time/timemanager.h"
#include "../../utils/utils.h"
#include "../../main/user_code.h"
#include "../wmanager.h"
#include "dsp.h"

#ifdef IPHONE
#include <string.h>
#endif

int window_handler_check_data( sundog_event* evt, window_manager* wm )
{
    if( evt->type != EVT_GETDATASIZE )
    {
	if( evt->win->data == 0 )
	{
	    dprint( "ERROR: Event to window (%s) without data\n", evt->win->name );
	    return 1;
	}
    }
    return 0;
}

#include "handlers_dialog.h"
#include "handlers_files.h"
#include "handlers_list.h"
#include "handlers_popup.h"
#include "handlers_virtual_keyboard.h"

int null_handler( sundog_event* evt, window_manager* wm )
{
    return 0;
}

int desktop_handler( sundog_event* evt, window_manager* wm )
{
    int retval = 0;
    WINDOWPTR win = evt->win;
    switch( evt->type )
    {
	case EVT_SCREENRESIZE:
	    win->x = 0;
	    win->y = 0;
	    win->xsize = wm->screen_xsize;
	    win->ysize = wm->screen_ysize;
	    break;
	case EVT_MOUSEBUTTONDOWN:
	case EVT_MOUSEMOVE:
	case EVT_MOUSEBUTTONUP:
	case EVT_TOUCHBEGIN:
	case EVT_TOUCHEND:
	case EVT_TOUCHMOVE:
	    retval = 1;
	    break;
    }
    return retval;
}

#define DIVIDER_BINDS 2

int DIVIDER_VSCROLL_MIN = 0;
int DIVIDER_VSCROLL_MAX = 0;
bool VERTICAL_DIVIDER = 0;
bool DIVIDER_WITH_TIME = 0;

struct divider_data
{
    int start_x;
    int start_y;
    int start_wx;
    int start_wy;
    int start_vscroll;
    int pushed;
    COLOR color;
    int vscroll_min;
    int vscroll_max;
    int vscroll_cur;
    WINDOWPTR binds[ DIVIDER_BINDS ];
    int binds_x[ DIVIDER_BINDS ];
    int binds_y[ DIVIDER_BINDS ];
    WINDOWPTR push_win;
    char push_right;
    char push_bottom;
    
    bool vert;
    bool time;
    
    WINDOWPTR prev_focus_win;
};

void divider_push_start( divider_data *data, WINDOWPTR win )
{
    if( data->push_win )
    {
        if( data->push_win->x >= win->x + win->xsize )
    	    data->push_right = 1;
	else
	    data->push_right = 0;
	if( data->push_win->y >= win->y + win->ysize )
	    data->push_bottom = 1;
	else
	    data->push_bottom = 0;
    }
}

void divider_push( divider_data *data, WINDOWPTR win )
{
    if( data->push_win )
    {
        if( data->push_bottom )
        {
    	    if( win->y + win->ysize > data->push_win->y )
	    {
	        //Push down:
	        data->push_win->y = win->y + win->ysize;
	    }
	}
	else
	{
	    if( win->y < data->push_win->y + data->push_win->ysize )
	    {
		//Push up:
		data->push_win->y = win->y - data->push_win->ysize;
	    }
	}
    }
}

int divider_handler( sundog_event *evt, window_manager *wm )
{
    int retval = 0;
    WINDOWPTR win = evt->win;
    divider_data* data = (divider_data*)win->data;
    switch( evt->type )
    {
	case EVT_GETDATASIZE:
	    retval = sizeof( divider_data );
	    break;
	case EVT_AFTERCREATE:
	    data->pushed = 0;
	    data->color = win->color;
	    data->vscroll_min = DIVIDER_VSCROLL_MIN;
	    data->vscroll_max = DIVIDER_VSCROLL_MAX;
	    data->vert = VERTICAL_DIVIDER;
	    data->time = DIVIDER_WITH_TIME;
	    data->vscroll_cur = 0;
	    {
		for( int a = 0; a < DIVIDER_BINDS;a++ )
		    data->binds[ a ] = 0;
	    }
	    data->push_win = 0;
	    data->prev_focus_win = 0;
	    DIVIDER_VSCROLL_MIN = 0;
	    DIVIDER_VSCROLL_MAX = 0;
	    VERTICAL_DIVIDER = 0;
	    DIVIDER_WITH_TIME = 0;
	    retval = 1;
	    break;
	case EVT_MOUSEBUTTONDOWN:
	    if( evt->key == MOUSE_BUTTON_LEFT )
	    {
		data->start_x = evt->x;
		data->start_y = evt->y;
		data->start_wx = win->x;
		data->start_wy = win->y;
		data->start_vscroll = data->vscroll_cur;
		for( int a = 0; a < DIVIDER_BINDS; a++ )
		{
		    if( data->binds[ a ] )
		    {
			data->binds_x[ a ] = data->binds[ a ]->x;
			data->binds_y[ a ] = data->binds[ a ]->y;
			divider_push_start( (divider_data*)data->binds[ a ]->data, data->binds[ a ] );
		    }
		}
		divider_push_start( data, win );
		data->pushed = 1;
		win->color = PUSHED_COLOR( data->color );
		draw_window( win, wm );
		retval = 1;
	    }
	    break;
	case EVT_MOUSEMOVE:
	    if( evt->key == MOUSE_BUTTON_LEFT )
	    {
		int dx = evt->x - data->start_x;
		int dy = evt->y - data->start_y;

		win->x = data->start_wx + dx;
		win->y = data->start_wy + dy;
		for( int a = 0; a < DIVIDER_BINDS; a++ )
		{
		    if( data->binds[ a ] )
		    {
			data->binds[ a ]->x = data->binds_x[ a ] + dx;
			data->binds[ a ]->y = data->binds_y[ a ] + dy;
			divider_push( (divider_data*)data->binds[ a ]->data, data->binds[ a ] );
		    }
		}
		divider_push( data, win );

		if( data->vscroll_min != data->vscroll_max )
		{
		    data->vscroll_cur = data->start_vscroll + dy;
		    if( data->vscroll_cur < data->vscroll_min )
			data->vscroll_cur = data->vscroll_min;
		    if( data->vscroll_cur > data->vscroll_max )
			data->vscroll_cur = data->vscroll_max;
		    if( win->action_handler )
		    {
			win->action_handler( win->handler_data, win, wm );
		    }
		}
		
		recalc_regions( wm );
		draw_window( wm->root_win, wm );
		
		retval = 1;
	    }
	    break;
	case EVT_MOUSEBUTTONUP:
	    if( evt->key == MOUSE_BUTTON_LEFT && data->pushed )
	    {
		data->pushed = 0;
		set_focus_win( data->prev_focus_win, wm );
		win->color = data->color;
		draw_window( win, wm );
		retval = 1;
	    }
	    break;
	case EVT_FOCUS:
	    data->prev_focus_win = wm->prev_focus_win;
	    retval = 1;
	    break;
	case EVT_UNFOCUS:
	    if( data->pushed )
	    {
		data->pushed = 0;
		win->color = data->color;
		draw_window( win, wm );
		retval = 1;
	    }
	    break;
	case EVT_DRAW:
	    win_draw_lock( win, wm );
	    {
		int cx = win->xsize / 2;
		int cy = win->ysize / 2;
		COLOR color = win->color;
		if( data->time )
		{
		    int h = time_hours();
		    int m = time_minutes();
		    UTF8_CHAR s[ 6 ];
		    s[ 0 ] = h / 10 + '0';
		    s[ 1 ] = h % 10 + '0';
		    s[ 2 ] = ':';
		    s[ 3 ] = m / 10 + '0';
		    s[ 4 ] = m % 10 + '0';
		    s[ 5 ] = 0;

		    int time_xsize = string_size( win, s, wm ) + 2;
		    win_draw_box( win, time_xsize, 0, win->xsize, win->ysize, color, wm );
		    win_draw_box( win, 0, 0, 2, win->ysize, color, wm );
		    
		    int y = ( win->ysize - char_y_size( win, wm ) ) / 2;
		    win_draw_string( win, s, 2, y, blend( wm->color2, color, 130 ), color, wm );
		    win_draw_box( win, 0, 0, time_xsize, y, color, wm );
		    win_draw_box( win, 0, y + char_y_size( win, wm ), time_xsize, win->ysize, color, wm );
		}
		else 
		{
		    win_draw_box( win, 0, 0, win->xsize, win->ysize, color, wm );
		}
		if( 1 )
		{
		    win_draw_box( win, cx - 2, cy - 2, 2, 2, BORDER_COLOR( color ), wm );
		    win_draw_box( win, cx + 1, cy - 2, 2, 2, BORDER_COLOR( color ), wm );
		    win_draw_box( win, cx - 2, cy + 1, 2, 2, BORDER_COLOR( color ), wm );
		    win_draw_box( win, cx + 1, cy + 1, 2, 2, BORDER_COLOR( color ), wm );
		}
		int xx = 0;
		int yy = 0;
		if( data->vert ) yy = 1; else xx = 1;
		win_draw_frame( win, -xx, -yy, win->xsize + xx * 2, win->ysize + yy * 2, BORDER_COLOR( color ), wm );
	    }
	    win_draw_unlock( win, wm );
	    retval = 1;
	    break;
	case EVT_TOUCHBEGIN:
	case EVT_TOUCHEND:
	case EVT_TOUCHMOVE:
	    retval = 1;
	    break;
    }
    return retval;
}

void bind_divider_to( WINDOWPTR win, WINDOWPTR bind_to, int bind_num, window_manager* wm )
{
    divider_data* data = (divider_data*)win->data;
    data->binds[ bind_num ] = bind_to;
}

void set_divider_push( WINDOWPTR win, WINDOWPTR push_win, window_manager* wm )
{
    divider_data* data = (divider_data*)win->data;
    data->push_win = push_win;
}

void set_divider_vscroll_parameters( WINDOWPTR win, int min, int max, window_manager* wm )
{
    divider_data* data = (divider_data*)win->data;
    data->vscroll_min = min;
    data->vscroll_max = max;
}

int get_divider_vscroll_value( WINDOWPTR win, window_manager* wm )
{
    divider_data* data = (divider_data*)win->data;
    return data->vscroll_cur;
}

struct label_data 
{
    WINDOWPTR prev_focus_win;
};

int label_handler( sundog_event* evt, window_manager* wm )
{
    if( window_handler_check_data( evt, wm ) ) return 0;
    int retval = 0;
    WINDOWPTR win = evt->win;
    label_data* data = (label_data*)win->data;
    switch( evt->type )
    {
	case EVT_GETDATASIZE:
	    retval = sizeof( label_data );
	    break;
	case EVT_AFTERCREATE:
	    data->prev_focus_win = 0;
	    break;
	case EVT_FOCUS:
	    data->prev_focus_win = wm->prev_focus_win;
	    retval = 1;
	    break;
	case EVT_MOUSEBUTTONUP:
	    set_focus_win( data->prev_focus_win, wm );
	    retval = 1;
	    break;
	case EVT_DRAW:
	    win_draw_lock( win, wm );
	    win_draw_box( win, 0, 0, win->xsize, win->ysize, win->parent->color, wm );
	    //Draw name:
	    {
		int ty = ( win->ysize - char_y_size( win, wm ) ) / 2;
		win_draw_string( win, win->name, 0, ty, blend( wm->color2, win->parent->color, LABEL_TRANSP ), win->parent->color, wm );
	    }
	    retval = 1;
	    win_draw_unlock( win, wm );
	    break;
	case EVT_TOUCHBEGIN:
	case EVT_TOUCHEND:
	case EVT_TOUCHMOVE:
	    retval = 1;
	    break;
    }
    return retval;
}

int TEXT_ZOOM = 256;
bool TEXT_NO_VIRTUAL_KEYBOARD = 0;
bool TEXT_CALL_HANDLER_ON_ANY_CHANGES = 0;
bool TEXT_RO = 0;
int CREATE_NUMERIC_TEXT = 0;
int CREATE_NUMERIC_TEXT_MIN = 0;
int CREATE_NUMERIC_TEXT_MAX = 0;
bool NUMERIC_TEXT_HIDE_ZERO = 0;

struct text_data 
{
    WINDOWPTR this_window;
    WINDOWPTR clear_button;
    UTF32_CHAR* text;
    UTF8_CHAR* output_str;
    int cur_pos;
    int active;
    int call_handler_on_any_changes;
    int no_virtual_kbd;
    int zoom;
    int numeric;
    int min;
    int max;
    bool hide_zero;
    int ro;
    WINDOWPTR prev_focus_win;
    UTF8_CHAR show_char;
};

int text_dec_handler( void* user_data, WINDOWPTR win, window_manager* wm )
{
    text_data* data = (text_data*)user_data;
    int val = text_get_value( data->this_window, wm );
    val--;
    text_set_value( data->this_window, val, wm );
    draw_window( data->this_window, wm );
    if( data->this_window->action_handler )
        data->this_window->action_handler( data->this_window->handler_data, data->this_window, wm );
    return 0;
}

int text_inc_handler( void* user_data, WINDOWPTR win, window_manager* wm )
{
    text_data* data = (text_data*)user_data;
    int val = text_get_value( data->this_window, wm );
    val++;
    text_set_value( data->this_window, val, wm );
    draw_window( data->this_window, wm );
    if( data->this_window->action_handler )
        data->this_window->action_handler( data->this_window->handler_data, data->this_window, wm );
    return 0;
}

int text_clear_handler( void* user_data, WINDOWPTR win, window_manager* wm )
{
    text_data* data = (text_data*)user_data;
    text_set_text( data->this_window, "", wm );
    draw_window( data->this_window, wm );
    if( data->this_window->action_handler && data->call_handler_on_any_changes )
        data->this_window->action_handler( data->this_window->handler_data, data->this_window, wm );
    return 0;
}

void text_show_clear_button( text_data* data )
{
    if( data->ro == 0 && data->clear_button && data->active )
    {
	if( data->clear_button->flags & WIN_FLAG_ALWAYS_INVISIBLE )
        {
            data->clear_button->flags &= ~WIN_FLAG_ALWAYS_INVISIBLE;
            show_window( data->clear_button, g_wm );
            recalc_regions( g_wm );
        }
    }
}

void text_hide_clear_button( text_data* data )
{
    if( data->ro == 0 && data->clear_button )
    {
	if( !( data->clear_button->flags & WIN_FLAG_ALWAYS_INVISIBLE ) )
        {
            data->clear_button->flags |= WIN_FLAG_ALWAYS_INVISIBLE;
            hide_window( data->clear_button, g_wm );
            recalc_regions( g_wm );
        }
    }
}

int text_handler( sundog_event* evt, window_manager* wm )
{
    if( window_handler_check_data( evt, wm ) ) return 0;
    int retval = 0;
    WINDOWPTR win = evt->win;
    text_data* data = (text_data*)win->data;
    int p, i;
    UTF32_CHAR c;
    int xp;
    int rx = evt->x - win->screen_x;
    int changes = 0;
    switch( evt->type )
    {
	case EVT_GETDATASIZE:
	    retval = sizeof( text_data );
	    break;
	case EVT_AFTERCREATE:
	    data->this_window = win;
	    data->output_str = 0;
	    data->text = (UTF32_CHAR*)MEM_NEW2( 32 * sizeof( UTF32_CHAR ) );
	    mem_zero( data->text );
	    data->cur_pos = 0;
	    data->active = 0;
	    data->call_handler_on_any_changes = TEXT_CALL_HANDLER_ON_ANY_CHANGES;
	    data->no_virtual_kbd = TEXT_NO_VIRTUAL_KEYBOARD;
	    data->zoom = TEXT_ZOOM;
	    data->numeric = CREATE_NUMERIC_TEXT;
	    data->min = CREATE_NUMERIC_TEXT_MIN;
	    data->max = CREATE_NUMERIC_TEXT_MAX;
	    data->hide_zero = NUMERIC_TEXT_HIDE_ZERO;
	    data->ro = TEXT_RO;
	    data->show_char = 0;
	    data->clear_button = 0;
	    if( data->numeric )
	    {
		CREATE_FLAT_BUTTON = 1;
		CREATE_BUTTON_WITH_AUTOREPEAT = 1;
		WINDOWPTR b = new_window( "-", 1, 1, 1, 1, win->color, win, button_handler, wm );
		set_window_controller( b, 0, wm, CPERC, (WCMD)100, CSUB, (WCMD)SCROLLBAR_SIZE * 2, CEND );
		set_window_controller( b, 1, wm, (WCMD)0, CEND );
		set_window_controller( b, 2, wm, CPERC, (WCMD)100, CSUB, (WCMD)SCROLLBAR_SIZE, CEND );
	        set_window_controller( b, 3, wm, CPERC, (WCMD)100, CEND );
		set_handler( b, text_dec_handler, data, wm );
		CREATE_FLAT_BUTTON = 1;
		CREATE_BUTTON_WITH_AUTOREPEAT = 1;
		b = new_window( "+", 1, 1, 1, 1, win->color, win, button_handler, wm );
		set_window_controller( b, 0, wm, CPERC, (WCMD)100, CSUB, (WCMD)SCROLLBAR_SIZE, CEND );
		set_window_controller( b, 1, wm, (WCMD)0, CEND );
		set_window_controller( b, 2, wm, CPERC, (WCMD)100, CEND );
	        set_window_controller( b, 3, wm, CPERC, (WCMD)100, CEND );
		set_handler( b, text_inc_handler, data, wm );
	    }
	    else
	    {
		CREATE_FLAT_BUTTON = 1;
		data->clear_button = new_window( "x", 1, 1, 1, 1, win->color, win, button_handler, wm );
		data->clear_button->flags |= WIN_FLAG_ALWAYS_INVISIBLE;
		set_window_controller( data->clear_button, 0, wm, CPERC, (WCMD)100, CSUB, (WCMD)SCROLLBAR_SIZE, CEND );
		set_window_controller( data->clear_button, 1, wm, (WCMD)0, CEND );
		set_window_controller( data->clear_button, 2, wm, CPERC, (WCMD)100, CEND );
	        set_window_controller( data->clear_button, 3, wm, CPERC, (WCMD)100, CEND );
		set_handler( data->clear_button, text_clear_handler, data, wm );
	    }
	    data->prev_focus_win = 0;
	    TEXT_ZOOM = 256;
	    TEXT_NO_VIRTUAL_KEYBOARD = 0;
	    TEXT_CALL_HANDLER_ON_ANY_CHANGES = 0;
	    TEXT_RO = 0;
	    CREATE_NUMERIC_TEXT = 0;
	    CREATE_NUMERIC_TEXT_MIN = 0;
	    CREATE_NUMERIC_TEXT_MAX = 0;
	    NUMERIC_TEXT_HIDE_ZERO = 0;
	    retval = 1;
	    break;
	case EVT_FOCUS:
	    if( data->ro )
	    {
		retval = 1;
		break;
	    }
	    data->active = 1;
	    retval = 1;
	    break;
	case EVT_UNFOCUS:
	    if( data->ro )
	    {
		retval = 1;
		break;
	    }
	    if( data->active )
	    {
		data->active = 0;
		draw_window( win, wm );
		retval = 1;
	    }
	    break;
	case EVT_MOUSEMOVE:
	case EVT_MOUSEBUTTONDOWN:
	    if( data->ro )
	    {
		retval = 1;
		break;
	    }
	    if( evt->key == MOUSE_BUTTON_LEFT )
	    {
		win_draw_lock( win, wm );
		wbd_lock( win );
		wm->cur_font_zoom = data->zoom;
		data->prev_focus_win = wm->prev_focus_win;
		data->active = 1;
		xp = 0;
		for( p = 0; p < mem_get_size( data->text ) / (int)sizeof( UTF32_CHAR ); p++ )
		{
		    c = data->text[ p ];
		    int charx = wbd_char_x_size( c, wm );
		    if( c == 0 ) charx = 600;
		    if( rx >= xp && rx < xp + charx )
		    {
		        data->cur_pos = p;
		        break;
		    }
		    xp += charx;
		    if( data->text[ p ] == 0 ) break;
		}
		wm->cur_font_zoom = 256;
		wbd_unlock();
		win_draw_unlock( win, wm );
		draw_window( win, wm );
	    }
	    retval = 1;
	    break;
	case EVT_MOUSEBUTTONUP:
	    if( data->ro )
	    {
		retval = 1;
		break;
	    }
	    if( evt->key == MOUSE_BUTTON_LEFT )
	    {
		if( data->active ) 
		{
		    if( data->no_virtual_kbd == 0 )
			show_keyboard_for_text_window( win, wm );
		}
	    }
	    retval = 1;
	    break;
	case EVT_BUTTONDOWN:
	    if( data->active == 0 ) 
	    {
		retval = 0;
		break;
	    }
	    if( data->ro )
	    {
		retval = 0;
		break;
	    }
	    if( evt->flags & EVT_FLAG_CTRL )
	    {
		retval = 0;
		break;
	    }
	    
	    if( evt->key == KEY_ENTER || evt->key == KEY_ESCAPE )
	    {
		data->active = 0;
		set_focus_win( data->prev_focus_win, wm );
		draw_window( win, wm );
		if( evt->key == KEY_ENTER && win->action_handler )
		    win->action_handler( win->handler_data, win, wm );
		retval = 1;
		break;
	    }
	    
	    c = 0;
	    if( evt->key >= 0x20 && evt->key <= 0x7E ) c = evt->key;
	    if( evt->flags & EVT_FLAG_SHIFT )
	    {
		if( evt->key >= 'a' && evt->key <= 'z' ) 
		{
		    c = evt->key - 0x20;
		}
		else
		{
		    switch( evt->key )
		    {
			case '0': c = ')'; break;
			case '1': c = '!'; break;
			case '2': c = '@'; break;
			case '3': c = '#'; break;
			case '4': c = '$'; break;
			case '5': c = '%'; break;
			case '6': c = '^'; break;
			case '7': c = '&'; break;
			case '8': c = '*'; break;
			case '9': c = '('; break;
			case '[': c = '{'; break;
			case ']': c = '}'; break;
			case ';': c = ':'; break;
			case  39: c = '"'; break;
			case ',': c = '<'; break;
			case '.': c = '>'; break;
			case '/': c = '?'; break;
			case '-': c = '_'; break;
			case '=': c = '+'; break;
			case  92: c = '|'; break;
			case '`': c = '~'; break;
		    }
		}
	    }
	    
	    {
	        UTF32_CHAR* text_buf = data->text;
	        int text_size = mem_get_size( text_buf ) / (int)sizeof( UTF32_CHAR );
	    
		if( c == 0 )
	        {
		    if( evt->key == KEY_BACKSPACE )
		    {
			if( data->cur_pos >= 1 )
			{
			    data->cur_pos--;
			    for( i = data->cur_pos; i < text_size - 1; i++ ) 
				text_buf[ i ] = text_buf[ i + 1 ];
			}
			changes = 1;
		    }
		    else
		    if( evt->key == KEY_DELETE )
		    {
			if( data->text[ data->cur_pos ] != 0 )
			    for( i = data->cur_pos; i < text_size - 1; i++ ) 
			        text_buf[ i ] = text_buf[ i + 1 ];
			changes = 1;
		    }
		    else
		    if( evt->key == KEY_LEFT )
		    {
			if( data->cur_pos >= 1 )
			{
			    data->cur_pos--;
			}
		    }
		    else
		    if( evt->key == KEY_RIGHT )
		    {
			if( text_buf[ data->cur_pos ] != 0 )
			    data->cur_pos++;
		    }
		    else
		    if( evt->key == KEY_END )
		    {
			for( i = 0; i < text_size; i++ ) 
			    if( data->text[ i ] == 0 ) break;
			data->cur_pos = i;
		    }
		    else
		    if( evt->key == KEY_HOME )
		    {
			data->cur_pos = 0;
		    }
		    else 
		    {
			retval = 0;
			break;
		    }
		}
		else
		{
		    //Add new char:
		    if( mem_strlen_utf32( (const UTF32_CHAR*)data->text ) + 16 > text_size )
		    {
			data->text = (UTF32_CHAR*)mem_resize( data->text, ( text_size + 16 ) * sizeof( UTF32_CHAR ) );
			mem_set( &data->text[ text_size ], 16 * sizeof( UTF32_CHAR ), 0 );
			text_size += 16;
			text_buf = data->text;
		    }
		    for( i = text_size - 1; i >= data->cur_pos + 1; i-- )
			text_buf[ i ] = text_buf[ i - 1 ];
		    text_buf[ data->cur_pos ] = c;
		    data->cur_pos++;
		    changes = 1;
		}
	    }

	    draw_window( win, wm );
	    
	    if( data->call_handler_on_any_changes && changes && win->action_handler )
		win->action_handler( win->handler_data, win, wm );
	    retval = 1;
	    break;
	case EVT_DRAW:
	    if( data->text && data->text[ 0 ] != 0 )
	    {
		text_show_clear_button( data );
	    }
	    else
	    {
    		text_hide_clear_button( data );
	    }
	    win_draw_lock( win, wm );
	    wbd_lock( win );
	    wm->cur_font_zoom = data->zoom;
	    //Fill window:
	    wm->cur_transparency = 256;
	    wm->cur_font_color = wm->color3;
	    draw_box( 0, 0, win->xsize, win->ysize, win->color, wm );
	    //Draw text and cursor:
	    xp = 0;
	    if( data->ro )
	    {
		int txsize = 0;
		for( p = 0; p < mem_get_size( data->text ) / (int)sizeof( UTF32_CHAR ); p++ )
		{
		    c = data->text[ p ];
		    if( c == 0 ) break;
		    txsize += wbd_char_x_size( c, wm );
		}
		if( txsize + xp >= ( win->xsize - 1 ) )
		{
		    xp -= txsize + xp - ( win->xsize - 1 );
		}
	    }
	    {
		int ty = ( win->ysize - wbd_char_y_size( wm ) ) / 2;
		for( p = 0; p < mem_get_size( data->text ) / (int)sizeof( UTF32_CHAR ); p++ )
		{
		    c = data->text[ p ];
		    if( c == 0 ) c = ' ';
		    int charx = wbd_char_x_size( c, wm );
		    if( data->cur_pos == p && data->active )
		    {
			draw_box( xp, 0, charx, win->ysize, blend( wm->color3, win->color, 128 ), wm );
		    }
		    draw_char( c, xp, ty, wm );
		    xp += charx;
		    if( data->text[ p ] == 0 ) break;
		}
		if( p == 0 && !data->active )
		{
		    draw_string( "...", 0, ty, wm );
		}
	    }
	    if( data->show_char )
	    {
		//Show some symbol (defined in user request):
		wm->cur_transparency = 160;
		draw_box( 0, 0, win->xsize, win->ysize, wm->color0, wm );
		wm->cur_transparency = 256;
		UTF8_CHAR sstr[ 8 ];
		sstr[ 0 ] = data->show_char;
		sstr[ 1 ] = 0;
		int len = wbd_string_size( (const UTF8_CHAR*)sstr, wm );
		wm->cur_font_color = wm->color3;
		draw_string( 
		    sstr, 
		    ( win->xsize - len ) / 2, 
		    ( win->ysize - wbd_char_y_size( wm ) ) / 2,
		    wm );
		data->show_char = 0;
	    }
	    wm->cur_font_zoom = 256;
	    wbd_draw();
	    wbd_unlock();
	    win_draw_unlock( win, wm );
	    retval = 1;
	    break;
	case EVT_BEFORECLOSE:
	    hide_keyboard_for_text_window( wm );
	    if( data->output_str ) mem_free( data->output_str );
	    if( data->text ) mem_free( data->text );
	    retval = 1;
	    break;
	case EVT_USERDEFINED1:
	    {
		data->show_char = (UTF8_CHAR)evt->key;
		draw_window( win, wm );
	    }
	    retval = 1;
	    break;
	case EVT_TOUCHBEGIN:
	case EVT_TOUCHEND:
	case EVT_TOUCHMOVE:
	    retval = 1;
	    break;
    }
    return retval;
}

void text_set_text( WINDOWPTR win, const UTF8_CHAR* text, window_manager* wm )
{
    if( win && text )
    {
	int len = mem_strlen( text ) + 1;
	UTF32_CHAR* ts = (UTF32_CHAR*)MEM_NEW( HEAP_DYNAMIC, len * sizeof( UTF32_CHAR ) );
	utf8_to_utf32( ts, len, text );
	int len2 = mem_strlen_utf32( ts ) + 1;
	text_data* data = (text_data*)win->data;
	if( mem_get_size( data->text ) / (int)sizeof( UTF32_CHAR ) < len2 )
	{
	    //Resize text buffer:
	    data->text = (UTF32_CHAR*)mem_resize( data->text, len2 * sizeof( UTF32_CHAR ) );
	}
	mem_copy( data->text, ts, len2 * sizeof( UTF32_CHAR ) );
	if( data->cur_pos > len2 - 1 )
	    data->cur_pos = len2 - 1;
	text_hide_clear_button( data );
	draw_window( win, wm );
	mem_free( ts );
    }
}

void text_set_cursor_position( WINDOWPTR win, int cur_pos, window_manager* wm )
{
    if( win )
    {
	text_data* data = (text_data*)win->data;
	data->cur_pos = cur_pos;
	draw_window( win, wm );
    }
}

void text_set_value( WINDOWPTR win, int val, window_manager* wm )
{
    if( win )
    {
	text_data* data = (text_data*)win->data;
	if( val < data->min ) val = data->min;
	if( val > data->max ) val = data->max;
	UTF8_CHAR ts[ 16 ];
	if( data->numeric == 1 )
	    int_to_string( val, ts );
	else
	    hex_int_to_string( val, ts );
	if( data->hide_zero && val == 0 )
	{
	    ts[ 0 ] = 0;
	}
	text_set_text( win, (const UTF8_CHAR*)ts, wm );
    }
}

UTF8_CHAR* text_get_text( WINDOWPTR win, window_manager* wm )
{
    if( win )
    {
	text_data* data = (text_data*)win->data;
	if( data->output_str ) mem_free( data->output_str );
	data->output_str = (UTF8_CHAR*)MEM_NEW( HEAP_DYNAMIC, mem_get_size( data->text ) );
	utf32_to_utf8( data->output_str, mem_get_size( data->text ), data->text );
	return data->output_str;
    }
    return 0;
}

int text_get_cursor_position( WINDOWPTR win, window_manager* wm )
{
    if( win )
    {
	text_data* data = (text_data*)win->data;
	return data->cur_pos;
    }
    return 0;
}

int text_get_value( WINDOWPTR win, window_manager* wm )
{
    int val = 0;
    if( win )
    {
	text_data* data = (text_data*)win->data;
	UTF8_CHAR* s = text_get_text( win, wm );
	if( data->numeric == 1 )
	    val = string_to_int( s );
	else
	    val = hex_string_to_int( s );
	if( val < data->min ) val = data->min;
	if( val > data->max ) val = data->max;
    }
    return val;
}

void text_changed( WINDOWPTR win, window_manager* wm )
{
    if( win )
    {
	text_data* data = (text_data*)win->data;
	if( data->this_window->action_handler )
    	    data->this_window->action_handler( data->this_window->handler_data, win, wm );
    }
}

void text_set_zoom( WINDOWPTR win, int zoom, window_manager* wm )
{
    if( win )
    {
	text_data* data = (text_data*)win->data;
        data->zoom = zoom;
    }
}

sundog_image* CREATE_BITMAP_BUTTON = 0;
int CREATE_BITMAP_BUTTON_XSIZE = 0;
int CREATE_BITMAP_BUTTON_YSIZE = 0;
int CREATE_BITMAP_BUTTON_XOFF1 = 0;
int CREATE_BITMAP_BUTTON_YOFF1 = 0;
int CREATE_BITMAP_BUTTON_XOFF2 = 0;
int CREATE_BITMAP_BUTTON_YOFF2 = 0;
const UTF8_CHAR* CREATE_BUTTON_WITH_TEXT_MENU = 0;
bool CREATE_BUTTON_WITH_AUTOREPEAT = 0;
char CREATE_FLAT_BUTTON = 0;
int (*BUTTON_END_HANDLER)( void*, WINDOWPTR ) = 0;
void* BUTTON_END_HANDLER_DATA = 0;
uchar BUTTON_FLAGS = 0;

struct button_data
{
    WINDOWPTR this_window;
    sundog_image* img;
    int img_xsize;
    int img_ysize;
    int img_xoff1;
    int img_yoff1;
    int img_xoff2;
    int img_yoff2;
    const UTF8_CHAR* text_menu;
    int menu_val;
    bool autorepeat;
    int timer;
    ticks_t autorepeat_base_time;
    uchar flags;
    bool pushed;
    bool pen_inside;
    char flat;
    int add_radius;
    
    UTF8_CHAR* text;
    COLOR text_color;
    uchar text_transp;

    int (*end_handler)( void*, WINDOWPTR );
    void* end_handler_data;

    WINDOWPTR prev_focus_win;
};

void button_timer( void* vdata, sundog_timer* timer, window_manager* wm )
{
    //Autorepeat timer
    button_data* data = (button_data*)vdata;
    if( data->pushed && data->pen_inside )
    {
	if( data->this_window->action_handler )
	{
	    data->flags |= BUTTON_FLAG_HANDLER_CALLED_FROM_TIMER;
	    data->this_window->action_handler( data->this_window->handler_data, data->this_window, wm );
	    data->flags &= ~BUTTON_FLAG_HANDLER_CALLED_FROM_TIMER;
	}
    }
    ticks_t t = time_ticks();
    while( 1 )
    {
	if( t < data->autorepeat_base_time + time_ticks_per_second() * 3 )
	{
	    timer->delay = time_ticks_per_second() / 20;
	    break;
	}
	if( t < data->autorepeat_base_time + time_ticks_per_second() * 6 )
	{
	    timer->delay = time_ticks_per_second() / 70;
	    break;
	}
	if( t < data->autorepeat_base_time + time_ticks_per_second() * 9 )
	{
	    timer->delay = time_ticks_per_second() / 260;
	    break;
	}
	timer->delay = time_ticks_per_second() / 1000;
	break;
    }
}

int button_handler( sundog_event* evt, window_manager* wm )
{
    if( window_handler_check_data( evt, wm ) ) return 0;
    int retval = 0;
    WINDOWPTR win = evt->win;
    button_data* data = (button_data*)win->data;
    COLOR col;
    int rx = evt->x - win->screen_x;
    int ry = evt->y - win->screen_y;
    switch( evt->type )
    {
	case EVT_GETDATASIZE:
	    retval = sizeof( button_data );
	    break;
	case EVT_AFTERCREATE:
	    data->this_window = win;
	    data->img = 0;
	    data->text_menu = CREATE_BUTTON_WITH_TEXT_MENU;
	    data->autorepeat = CREATE_BUTTON_WITH_AUTOREPEAT;
	    data->flat = CREATE_FLAT_BUTTON;
	    data->timer = -1;
	    data->flags = BUTTON_FLAGS;
	    data->pushed = 0;
	    data->text = 0;
	    data->text_color = wm->color3;
	    data->text_transp = 255;
	    data->end_handler = BUTTON_END_HANDLER;
	    data->end_handler_data = BUTTON_END_HANDLER_DATA;
	    if( wm->control_type == TOUCHCONTROL )
		data->add_radius = SCROLLBAR_SIZE / 2;
	    else
		data->add_radius = 0;
	    if( CREATE_BITMAP_BUTTON )
	    {
		data->img = CREATE_BITMAP_BUTTON;
		data->img_xsize = CREATE_BITMAP_BUTTON_XSIZE;
		data->img_ysize = CREATE_BITMAP_BUTTON_YSIZE;
		data->img_xoff1 = CREATE_BITMAP_BUTTON_XOFF1;
		data->img_yoff1 = CREATE_BITMAP_BUTTON_YOFF1;
		data->img_xoff2 = CREATE_BITMAP_BUTTON_XOFF2;
		data->img_yoff2 = CREATE_BITMAP_BUTTON_YOFF2;
	    }
	    if( data->flags & BUTTON_FLAG_SHOW_VALUE )
	    {
		win->font = 0;
		data->text_color = wm->color2;
		data->menu_val = 0;
	    }
	    data->prev_focus_win = 0;
	    CREATE_BITMAP_BUTTON = 0;
	    CREATE_BUTTON_WITH_TEXT_MENU = 0;
	    CREATE_BUTTON_WITH_AUTOREPEAT = 0;
	    CREATE_FLAT_BUTTON = 0;
	    BUTTON_END_HANDLER = 0;
	    BUTTON_END_HANDLER_DATA = 0;
	    BUTTON_FLAGS = 0;
	    retval = 1;
	    break;
	case EVT_BEFORECLOSE:
	    mem_free( data->text );
	    retval = 1;
	    break;
	case EVT_FOCUS:
	    data->prev_focus_win = wm->prev_focus_win;
	    retval = 1;
	    break;
	case EVT_MOUSEBUTTONDOWN:
	    if( evt->key == MOUSE_BUTTON_LEFT )
	    {
		if( rx >= 0 && rx < win->xsize &&
		    ry >= 0 && ry < win->ysize )
		{
		    data->pushed = 1;
		    data->pen_inside = 1;
		    if( data->autorepeat )
		    {
			if( data->timer != -1 )
			{
			    remove_timer( data->timer, wm );
			}
			int autorepeat_delay = time_ticks_per_second() / 2;
			data->timer = add_timer( button_timer, (void*)data, autorepeat_delay, wm );
			data->autorepeat_base_time = time_ticks() + autorepeat_delay;
		    }
		    draw_window( win, wm );
		    retval = 1;
		}
	    }
	    break;
	case EVT_MOUSEMOVE:
	    if( data->pushed )
	    {
		if( rx >= -data->add_radius && rx < win->xsize + data->add_radius &&
		    ry >= -data->add_radius && ry < win->ysize + data->add_radius )
		{
		    data->pen_inside = 1;
		}
		else
		{
		    data->pen_inside = 0;
		}
	    }
	    retval = 1;
	    break;
	case EVT_MOUSEBUTTONUP:
	    if( evt->key == MOUSE_BUTTON_LEFT && data->pushed )
	    {
		data->pushed = 0;
		set_focus_win( data->prev_focus_win, wm );
		if( rx >= -data->add_radius && rx < win->xsize + data->add_radius &&
		    ry >= -data->add_radius && ry < win->ysize + data->add_radius )
		{
		    if( data->text_menu )
		    {
		        //Show popup with text menu:
		        win->action_result = popup_menu( win->name, data->text_menu, win->screen_x, win->screen_y, wm->menu_color );
			if( win->action_result >= 0 )
			{
			    data->menu_val = win->action_result;
			}
		    }
		}
		if( data->autorepeat && data->timer != -1 )
		{
		    remove_timer( data->timer, wm );
		    data->timer = -1;
		}
		draw_window( win, wm );
		int (*end_handler)( void*, WINDOWPTR ) = data->end_handler;
		void* end_handler_data = data->end_handler_data;
		if( rx >= -data->add_radius && rx < win->xsize + data->add_radius &&
		    ry >= -data->add_radius && ry < win->ysize + data->add_radius &&
		    win->action_handler )
		{
		    win->action_handler( win->handler_data, win, wm );
		}
		if( end_handler )
		{
		    end_handler( end_handler_data, win );
		}
		retval = 1;
	    }
	    break;
	case EVT_UNFOCUS:
	    if( data->pushed )
	    {
		data->pushed = 0;
		if( data->autorepeat && data->timer != -1 )
		{
		    remove_timer( data->timer, wm );
		    data->timer = -1;
		}
		draw_window( win, wm );
		if( data->end_handler )
		{
		    data->end_handler( data->end_handler_data, win );
		}
		retval = 1;
	    }
	    break;
	case EVT_DRAW:
	    if( data->flags & BUTTON_FLAG_AUTOBACKCOLOR ) 
		col = win->parent->color;
	    else
		col = win->color;
	    win_draw_lock( win, wm );
	    if( data->pushed )
		col = PUSHED_COLOR( col );

	    if( data->img )
	    {
		wbd_lock( win );
		
		draw_box( 0, 0, win->xsize, win->ysize, col, wm );
		
		//Draw bitmap:
		int xoff;
		int yoff;
		if( data->pushed )
		{
		    xoff = data->img_xoff2;
		    yoff = data->img_yoff2;
		}
		else
		{
		    xoff = data->img_xoff1;
		    yoff = data->img_yoff1;
		}
		wm->cur_transparency = data->text_transp;
		draw_image_scaled(
		    xoff << IMG_PREC,
		    yoff << IMG_PREC,
		    data->img_xsize << IMG_PREC,
		    data->img_ysize << IMG_PREC,
		    ( win->xsize - data->img_xsize ) / 2,
		    ( win->ysize - data->img_ysize ) / 2 + data->pushed,
		    data->img_xsize,
		    data->img_ysize,
		    data->img,
		    wm );
		wm->cur_transparency = 255;
		    
		//Draw border:
		if( data->flat )
		{
		    switch( data->flat )
		    {
			case 1:
			    break;
			case 2:
			    draw_line( 0, 0, win->xsize, 0, wm->color1, wm );
			    draw_line( win->xsize - 1, 0, win->xsize - 1, win->ysize, wm->color1, wm );
			    break;
			case 3:
			    draw_line( 0, win->ysize - 1, win->xsize, win->ysize - 1, wm->color1, wm );
			    draw_line( win->xsize - 1, 0, win->xsize - 1, win->ysize, wm->color1, wm );
			    break;
		    }
		}
		else
		{
		    draw_frame( 0, 0, win->xsize - 1, win->ysize - 1, BORDER_COLOR( col ), wm );
		}
		
		wbd_draw();
		wbd_unlock();
	    }
	    else
	    {
		win_draw_box( win, 0, 0, win->xsize, win->ysize, col, wm );
		
		COLOR text_color = blend( col, data->text_color, data->text_transp );
		
		UTF8_CHAR* text;
		if( data->text )
		    text = data->text;
		else
		    text = (UTF8_CHAR*)win->name;
		if( text )
		{
		    //Draw name:
		    int tx;
		    int ty;
		    if( data->flags & BUTTON_FLAG_SHOW_VALUE )
		    {
			tx = 1;
			ty = win->ysize / 2 - char_y_size( win, wm );
			win_draw_string( win, text, tx, ty, text_color, col, wm );
			int text_xsize = string_size( win, text, wm );
			win_draw_string( win, ":", tx + text_xsize, ty, text_color, col, wm );
			
			//Draw value:
			if( data->text_menu )
			{
			    UTF8_CHAR val[ 64 ];
			    val[ 0 ] = 0;
			    val[ 63 ] = 0;
			    int i = 0;
			    UTF8_CHAR* m = (UTF8_CHAR*)data->text_menu;
			    for( ; *m != 0; m++ )
			    {
				if( *m == 0xA || *m == 0xD ) 
				{
				    i++;
				    continue;
				}
				if( i == data->menu_val )
				{
				    for( int i2 = 0; i2 < 63; i2++ )
				    {
					val[ i2 ] = *m;
					if( *m == 0 || *m == 0xD || *m == 0xA ) 
					{
					    val[ i2 ] = 0;
					    break;
					}
					m++;
				    }
				    break;
				}
			    }
			    ty = win->ysize / 2 + 1;
			    win_draw_string( win, val, tx, ty, blend( wm->color2, wm->blue, 64 ), col, wm );
			}
			
		    }
		    else
		    {
			int xx = string_size( win, text, wm );
			tx = ( win->xsize - xx ) / 2;
			ty = ( win->ysize - char_y_size( win, wm ) ) / 2;
			if( xx > win->xsize )
			{
			    if( data->flags & BUTTON_FLAG_LEFT_ALIGNMENT_ON_OVERFLOW )
			    {
				tx = 1;
			    }
			}
			win_draw_string( win, text, tx, ty + data->pushed, text_color, col, wm );
		    }
		}
		
		//Draw border:
		if( data->flat )
		{
		    switch( data->flat )
		    {
			case 2:
			    win_draw_line( win, 0, 0, win->xsize, 0, wm->color1, wm );
			    win_draw_line( win, win->xsize - 1, 0, win->xsize - 1, win->ysize, wm->color1, wm );
			    break;
			case 3:
			    win_draw_line( win, 0, win->ysize - 1, win->xsize, win->ysize - 1, wm->color1, wm );
			    win_draw_line( win, win->xsize - 1, 0, win->xsize - 1, win->ysize, wm->color1, wm );
			    break;
		    }
		}
		else
		{
		    win_draw_frame( win, 0, 0, win->xsize, win->ysize, BORDER_COLOR( col ), wm );
		}
	    }
	    retval = 1;
	    win_draw_unlock( win, wm );
	    break;
	case EVT_TOUCHBEGIN:
	case EVT_TOUCHEND:
	case EVT_TOUCHMOVE:
	    retval = 1;
	    break;
    }
    return retval;
}

void button_set_menu( WINDOWPTR win, const UTF8_CHAR* menu, window_manager* wm )
{
    if( win )
    {
        button_data* data = (button_data*)win->data;
	data->text_menu = menu;
    }
}

void button_set_menu_val( WINDOWPTR win, int val, window_manager* wm )
{
    if( win )
    {
        button_data* data = (button_data*)win->data;
	data->menu_val = val;
    }
}

int button_get_menu_val( WINDOWPTR win, window_manager* wm )
{
    if( win )
    {
        button_data* data = (button_data*)win->data;
	return data->menu_val;
    }
    return -1;
}

void button_set_text( WINDOWPTR win, UTF8_CHAR* text, window_manager* wm )
{
    if( win )
    {
	button_data* data = (button_data*)win->data;
	mem_free( data->text );
	if( text )
	{
	    data->text = (UTF8_CHAR*)MEM_NEW2( mem_strlen( text ) + 1 );
	    data->text[ 0 ] = 0;
	    mem_strcat( data->text, text );
	}
	else 
	{
	    data->text = 0;
	}
    }
}

void button_set_text_color( WINDOWPTR win, COLOR c, window_manager* wm )
{
    if( win )
    {
        button_data* data = (button_data*)win->data;
	data->text_color = c;
    }
}

void button_set_text_transp( WINDOWPTR win, uchar transp, window_manager* wm )
{
    if( win )
    {
        button_data* data = (button_data*)win->data;
	data->text_transp = transp;
    }
}

int button_get_text_transp( WINDOWPTR win, window_manager* wm )
{
    if( win )
    {
        button_data* data = (button_data*)win->data;
	return data->text_transp;
    }
    return 0;
}

void button_set_bitmap( 
    WINDOWPTR win, 
    sundog_image* img, 
    int img_xsize,
    int img_ysize, 
    int img_xoff1,
    int img_yoff1,
    int img_xoff2,
    int img_yoff2
)
{
    if( win )
    {
        button_data* data = (button_data*)win->data;
	data->img = img;
	data->img_xsize = img_xsize;
	data->img_ysize = img_ysize;
	data->img_xoff1 = img_xoff1;
	data->img_yoff1 = img_yoff1;
	data->img_xoff2 = img_xoff2;
	data->img_yoff2 = img_yoff2;
    }
}

void button_set_flags( WINDOWPTR win, uchar flags )
{
    if( win )
    {
	button_data* data = (button_data*)win->data;
	data->flags = flags;
    }
}

uchar button_get_flags( WINDOWPTR win )
{
    if( win )
    {
	button_data* data = (button_data*)win->data;
	return data->flags;
    }
    else 
    {
	return 0;
    }
}

bool CREATE_VERTICAL_SCROLLBAR = 0;
bool CREATE_REVERSE_SCROLLBAR = 0;
bool CREATE_COMPACT_SCROLLBAR = 0;
bool CREATE_FLAT_SCROLLBAR = 0;
int CREATE_HEX_SCROLLBAR = 0;
int (*SCROLLBAR_BEGIN_HANDLER)( void*, WINDOWPTR, int ) = 0;
int (*SCROLLBAR_END_HANDLER)( void*, WINDOWPTR, int ) = 0;
void* SCROLLBAR_BEGIN_HANDLER_DATA = 0;
void* SCROLLBAR_END_HANDLER_DATA = 0;
const UTF8_CHAR text_up[ 2 ] = { 0x11, 0 };
const UTF8_CHAR text_down[ 2 ] = { 0x12, 0 };
const UTF8_CHAR text_left[ 2 ] = { 0x13, 0 };
const UTF8_CHAR text_right[ 2 ] = { 0x14, 0 };

struct scrollbar_data
{
    WINDOWPTR this_window;
    WINDOWPTR but1;
    WINDOWPTR but2;
    bool but1_inc;
    bool vert;
    bool rev;
    bool compact_mode;
    bool flat;
    UTF8_CHAR* name;
    UTF8_CHAR* values;
    int max_value;
    int page_size;
    int step_size;
    int cur;
    bool bar_selected;
    int drag_start_x;
    int drag_start_y;
    int drag_start_val;
    int show_offset;

    int working_area;
    int min_work_area;    
    int button_size;
    WINDOWPTR expanded;

    UTF8_CHAR val_ts[ 64 ];
    UTF8_CHAR* name_str;
    UTF8_CHAR* value_str;
    
    int cur_x;
    int cur_y;

    int pos;
    int bar_size;
    float one_pixel_size;
    int move_region;

    int hex;
    
    bool begin;
    int	(*begin_handler)( void*, WINDOWPTR, int );
    int	(*end_handler)( void*, WINDOWPTR, int );
    void* begin_handler_data;
    void* end_handler_data;

    WINDOWPTR prev_focus_win;
};

int scrollbar_button_end( void* user_data, WINDOWPTR win )
{
    scrollbar_data* data = (scrollbar_data*)user_data;
    if( data->begin )
    {
	if( data->end_handler )
	    data->end_handler( data->end_handler_data, data->this_window, scrollbar_get_value( data->this_window, g_wm ) );
	data->begin = 0;
    }
    return 0;
}

int scrollbar_button( void* user_data, WINDOWPTR win, window_manager* wm )
{
    scrollbar_data* data = (scrollbar_data*)user_data;
    if( data->begin == 0 )
    {
	data->begin = 1;
	if( data->begin_handler )
	    data->begin_handler( data->begin_handler_data, data->this_window, scrollbar_get_value( data->this_window, wm ) );
    }
    bool inc = 0;
    if( data->but1 == win && data->but1_inc == 1 ) inc = 1;
    if( data->but2 == win && data->but1_inc == 0 ) inc = 1;
    if( inc )
    {
	data->cur += data->step_size;
	if( data->cur < 0 ) data->cur = 0;
	if( data->cur > data->max_value ) data->cur = data->max_value;
    }
    else 
    {
	data->cur -= data->step_size;
	if( data->cur < 0 ) data->cur = 0;
	if( data->cur > data->max_value ) data->cur = data->max_value;
    }
    draw_window( data->this_window, wm );
    if( data->this_window->action_handler )
	data->this_window->action_handler( data->this_window->handler_data, data->this_window, wm );
    return 0;
}

void draw_scrollbar_horizontal_selection( WINDOWPTR win, int x )
{
    int xsize = 12;

    g_wm->cur_transparency = 256;

    sundog_polygon p;
    sundog_vertex v[ 4 ];
    v[ 0 ].x = x;
    v[ 0 ].y = 0;
    v[ 0 ].c = g_wm->color3;
    v[ 0 ].t = 64;
    v[ 1 ].x = x + xsize;
    v[ 1 ].y = 0;
    v[ 1 ].t = 0;
    v[ 2 ].x = x + xsize;
    v[ 2 ].y = win->ysize;
    v[ 2 ].t = 0;
    v[ 3 ].x = x;
    v[ 3 ].y = win->ysize;
    v[ 3 ].t = 64;
    p.vnum = 4;
    p.v = v;
    g_wm->cur_flags = WBD_FLAG_ONE_COLOR;
    draw_polygon( &p );

    v[ 1 ].x = x - xsize;
    v[ 2 ].x = x - xsize;
    draw_polygon( &p );
}

void draw_scrollbar_vertical_selection( WINDOWPTR win, int y )
{
    int ysize = 12;
    
    g_wm->cur_transparency = 256;
    
    sundog_polygon p;
    sundog_vertex v[ 4 ];
    v[ 0 ].x = 0;
    v[ 0 ].y = y;
    v[ 0 ].c = g_wm->color3;
    v[ 0 ].t = 64;
    v[ 1 ].x = 0;
    v[ 1 ].y = y + ysize;
    v[ 1 ].t = 0;
    v[ 2 ].x = win->xsize;
    v[ 2 ].y = y + ysize;
    v[ 2 ].t = 0;
    v[ 3 ].x = win->xsize;
    v[ 3 ].y = y;
    v[ 3 ].t = 64;
    p.vnum = 4;
    p.v = v;
    g_wm->cur_flags = WBD_FLAG_ONE_COLOR;
    draw_polygon( &p );
    
    v[ 1 ].y = y - ysize;
    v[ 2 ].y = y - ysize;
    draw_polygon( &p );
}

void scrollbar_calc_button_size( WINDOWPTR win )
{
    scrollbar_data* data = (scrollbar_data*)win->data;
    if( data->vert )
    {
	data->button_size = data->but1->ysize;
    }
    else
    {
	data->button_size = data->but1->xsize;
    }
}

void scrollbar_calc_working_area( WINDOWPTR win )
{
    scrollbar_data* data = (scrollbar_data*)win->data;

    scrollbar_calc_button_size( win );
		
    data->working_area = 0;
    if( data->vert )
        data->working_area = win->ysize - data->button_size * 2;
    else
        data->working_area = win->xsize - data->button_size * 2;
        
    if( data->compact_mode )
    {
	if( data->min_work_area == 0 )
	    data->min_work_area = SCROLLBAR_SIZE * 2; //string_size( win, "88.###", g_wm );
	if( data->working_area < data->min_work_area )
	{
	    if( ( data->but1->flags & WIN_FLAG_ALWAYS_INVISIBLE ) == 0 )
	    {
	        hide_window( data->but1, g_wm );
	        hide_window( data->but2, g_wm );
	        data->but1->flags |= WIN_FLAG_ALWAYS_INVISIBLE;
	        data->but2->flags |= WIN_FLAG_ALWAYS_INVISIBLE;
	        recalc_regions( g_wm );
	    }
	    data->working_area += data->button_size * 2;
	    data->button_size = 0;
	}
	else
	{
	    if( data->but1->flags & WIN_FLAG_ALWAYS_INVISIBLE )
	    {
	        data->but1->flags &= ~WIN_FLAG_ALWAYS_INVISIBLE;
	        data->but2->flags &= ~WIN_FLAG_ALWAYS_INVISIBLE;
	        show_window( data->but1, g_wm );
	        show_window( data->but2, g_wm );
	        recalc_regions( g_wm );
	    }
	}
    }
}

void scrollbar_get_strings( WINDOWPTR win )
{
    scrollbar_data* data = (scrollbar_data*)win->data;
    
    data->name_str = data->name;
    data->value_str = data->val_ts;
    data->value_str[ 0 ] = 0;
    
    if( data->compact_mode )
    {
	UTF8_CHAR text_val[ 32 ];
	text_val[ 0 ] = 0;
	text_val[ 31 ] = 0;
	if( data->values )
	{
	    const UTF8_CHAR* vv = data->values;
	    int v = 0;
	    while( 1 )
	    {
	        if( v == data->cur ) break;
	        if( *vv == '/' ) v++;
	        if( *vv == 0 ) break;
	        vv++;
	    }
	    for( int i = 0; i < 31; i++ )
	    {
	        text_val[ i ] = vv[ i ];
	        if( text_val[ i ] == '/' ) text_val[ i ] = 0;
	        if( text_val[ i ] == 1 ) text_val[ i ] = '/';
	        if( text_val[ i ] == 0 ) break;
	    }
	}
	if( data->hex && data->max_value )
	{
	    uint val;
	    if( data->hex == 1 )
	    {
	        val = ( data->cur * 32768 ) / data->max_value;
	        if( ( data->cur * 32768 ) % data->max_value ) val++;
	        if( val > 32768 ) val = 32768;
	    }
	    else
	        val = data->cur + data->show_offset;
	    if( val > 9 )
	    {
		if( text_val[ 0 ] )
	    	    sprintf( data->value_str, "%s (%x)", text_val, (unsigned int)val );
	    	else
	    	    sprintf( data->value_str, "%d (%x)", data->cur + data->show_offset, (unsigned int)val );
	    }
	    else
	    {
		if( text_val[ 0 ] )
	    	    sprintf( data->value_str, "%s (%x)", text_val, (unsigned int)val );
	    	else
	    	    sprintf( data->value_str, "%d", data->cur + data->show_offset );
	    }
	}
	else
	{
	    if( text_val[ 0 ] )
		sprintf( data->value_str, "%s", text_val );
	    else
		sprintf( data->value_str, "%d", data->cur + data->show_offset );
	}
    }
}

struct scrollbar_expanded_data
{
    WINDOWPTR this_window;
    WINDOWPTR ctl;
};

int scrollbar_expanded_handler( sundog_event* evt, window_manager* wm )
{
    if( window_handler_check_data( evt, wm ) ) return 0;
    int retval = 0;
    WINDOWPTR win = evt->win;
    scrollbar_expanded_data* data = (scrollbar_expanded_data*)win->data;
    switch( evt->type )
    {
	case EVT_GETDATASIZE:
	    retval = sizeof( scrollbar_expanded_data );
	    break;
	case EVT_AFTERCREATE:
	    data->this_window = win;
	    data->ctl = 0;
	    retval = 1;
	    break;
	case EVT_DRAW:
	    {
		scrollbar_data* cdata = (scrollbar_data*)data->ctl->data;
		scrollbar_get_strings( data->ctl );
		
		wbd_lock( win );

		wm->cur_transparency = 256;		
		draw_box( 0, 0, win->xsize, win->ysize, win->color, wm );
		
		int xc = 0;
                if( cdata->max_value != 0 )
                    xc = ( ( win->xsize - INTERELEMENT_SPACE * 2 ) * ( ( cdata->cur << 10 ) / cdata->max_value ) ) >> 10;
		draw_box( INTERELEMENT_SPACE, INTERELEMENT_SPACE, xc, win->ysize - INTERELEMENT_SPACE * 2, blend( wm->color2, win->color, 128 ), wm );
		
		int ychar = wbd_char_y_size( wm );

		wm->cur_font_color = wm->color3;

		if( cdata->name_str )
		{
		    draw_string( cdata->name_str, INTERELEMENT_SPACE, win->ysize / 2 - ychar, wm );
		}

		if( cdata->value_str )
		{
		    draw_string( cdata->value_str, INTERELEMENT_SPACE, win->ysize / 2, wm );
		}
		
		//Show center:
		wm->cur_transparency = 50;
		int cx = win->xsize / 2 - 1;
		draw_box( cx - 1, win->ysize - INTERELEMENT_SPACE - 1, 3, 1, wm->color3, wm );
		draw_box( cx, win->ysize - INTERELEMENT_SPACE - 2, 1, 1, wm->color3, wm );
                		
		wbd_draw();
    		wbd_unlock();
	    }
	    retval = 1;
	    break;
	case EVT_MOUSEBUTTONDOWN:
        case EVT_MOUSEMOVE:
    	    hide_window( win, wm );
    	    retval = 1;
    	    break;
    }
    return retval;
}

int scrollbar_handler( sundog_event* evt, window_manager* wm )
{
    if( window_handler_check_data( evt, wm ) ) return 0;
    int retval = 0;
    WINDOWPTR win = evt->win;
    scrollbar_data* data = (scrollbar_data*)win->data;
    int rx = evt->x - win->screen_x;
    int ry = evt->y - win->screen_y;
    switch( evt->type )
    {
	case EVT_GETDATASIZE:
	    retval = sizeof( scrollbar_data );
	    break;
	case EVT_AFTERCREATE:
	    data->this_window = win;
	    data->flat = CREATE_FLAT_SCROLLBAR;
	    data->begin_handler = SCROLLBAR_BEGIN_HANDLER;
	    data->end_handler = SCROLLBAR_END_HANDLER;
	    data->begin_handler_data = SCROLLBAR_BEGIN_HANDLER_DATA;
	    data->end_handler_data = SCROLLBAR_END_HANDLER_DATA;
	    data->begin = 0;
	    data->rev = CREATE_REVERSE_SCROLLBAR;
	    data->compact_mode = CREATE_COMPACT_SCROLLBAR;

	    if( data->compact_mode )
		data->flat = 1;

	    if( CREATE_VERTICAL_SCROLLBAR )
	    {
		data->vert = 1;
		CREATE_BUTTON_WITH_AUTOREPEAT = 1;
		CREATE_FLAT_BUTTON = 1;
		BUTTON_END_HANDLER = scrollbar_button_end;
		BUTTON_END_HANDLER_DATA = data;
		data->but1 = new_window( text_up, 0, 0, 1, 1, win->color, win, button_handler, wm );
		CREATE_BUTTON_WITH_AUTOREPEAT = 1;
		CREATE_FLAT_BUTTON = 1;
		BUTTON_END_HANDLER = scrollbar_button_end;
		BUTTON_END_HANDLER_DATA = data;
		data->but2 = new_window( text_down, 0, 0, 1, 1, win->color, win, button_handler, wm );
		set_window_controller( data->but2, 0, wm, (WCMD)0, CEND );
		set_window_controller( data->but2, 1, wm, CPERC, (WCMD)100, CEND );
		set_window_controller( data->but2, 2, wm, CPERC, (WCMD)100, CEND );
		set_window_controller( data->but1, 0, wm, (WCMD)0, CEND );
		set_window_controller( data->but1, 2, wm, CPERC, (WCMD)100, CEND );
		if( data->compact_mode )
		{
		    set_window_controller( data->but2, 3, wm, CWIN, (WCMD)win, CXSIZE, CMAXVAL, (WCMD)SCROLLBAR_SIZE, CPUTR0, CPERC, (WCMD)100, CSUB, CR0, CEND );
		    set_window_controller( data->but1, 1, wm, CWIN, (WCMD)win, CXSIZE, CMAXVAL, (WCMD)SCROLLBAR_SIZE, CPUTR0, CPERC, (WCMD)100, CSUB, CR0, CSUB, CR0, CEND );
		    set_window_controller( data->but1, 3, wm, CWIN, (WCMD)win, CXSIZE, CMAXVAL, (WCMD)SCROLLBAR_SIZE, CPUTR0, CPERC, (WCMD)100, CSUB, CR0, CEND );
		}
		else
		{
		    set_window_controller( data->but2, 3, wm, CPERC, (WCMD)100, CSUB, (WCMD)SCROLLBAR_SIZE, CEND );
		    set_window_controller( data->but1, 1, wm, CPERC, (WCMD)100, CSUB, (WCMD)SCROLLBAR_SIZE * 2, CEND );
		    set_window_controller( data->but1, 3, wm, CPERC, (WCMD)100, CSUB, (WCMD)SCROLLBAR_SIZE, CEND );
		}
		if( CREATE_REVERSE_SCROLLBAR )
		{
		    data->but1_inc = 1;
		}
		else
		{
		    data->but1_inc = 0;
		}
	    }
	    else
	    {
		data->vert = 0;
		CREATE_BUTTON_WITH_AUTOREPEAT = 1;
		CREATE_FLAT_BUTTON = 1;
		BUTTON_END_HANDLER = scrollbar_button_end;
		BUTTON_END_HANDLER_DATA = data;
		data->but1 = new_window( text_right, 0, 0, 1, 1, win->color, win, button_handler, wm );
		CREATE_BUTTON_WITH_AUTOREPEAT = 1;
		CREATE_FLAT_BUTTON = 1;
		BUTTON_END_HANDLER = scrollbar_button_end;
		BUTTON_END_HANDLER_DATA = data;
		data->but2 = new_window( text_left, 0, 0, 1, 1, win->color, win, button_handler, wm );
		set_window_controller( data->but1, 0, wm, CPERC, (WCMD)100, CEND );
		set_window_controller( data->but1, 1, wm, CPERC, (WCMD)0, CEND );
		set_window_controller( data->but1, 3, wm, CPERC, (WCMD)100, CEND );
		set_window_controller( data->but2, 1, wm, (WCMD)0, CEND );
		set_window_controller( data->but2, 3, wm, CPERC, (WCMD)100, CEND );
		if( data->compact_mode )
		{
		    set_window_controller( data->but1, 2, wm, CWIN, (WCMD)win, CYSIZE, CMAXVAL, (WCMD)SCROLLBAR_SIZE, CPUTR0, CPERC, (WCMD)100, CSUB, CR0, CEND );
		    set_window_controller( data->but2, 0, wm, CWIN, (WCMD)win, CYSIZE, CMAXVAL, (WCMD)SCROLLBAR_SIZE, CPUTR0, CPERC, (WCMD)100, CSUB, CR0, CSUB, CR0, CEND );
		    set_window_controller( data->but2, 2, wm, CWIN, (WCMD)win, CYSIZE, CMAXVAL, (WCMD)SCROLLBAR_SIZE, CPUTR0, CPERC, (WCMD)100, CSUB, CR0, CEND );
		}
		else
		{
		    set_window_controller( data->but1, 2, wm, CPERC, (WCMD)100, CSUB, (WCMD)SCROLLBAR_SIZE, CEND );
		    set_window_controller( data->but2, 0, wm, CPERC, (WCMD)100, CSUB, (WCMD)SCROLLBAR_SIZE * 2, CEND );
		    set_window_controller( data->but2, 2, wm, CPERC, (WCMD)100, CSUB, (WCMD)SCROLLBAR_SIZE, CEND );
		}
		if( CREATE_REVERSE_SCROLLBAR )
		{
		    data->but1_inc = 0;
		}
		else
		{
		    data->but1_inc = 1;
		}
	    }
	    scrollbar_calc_button_size( win );
	    data->working_area = 0;
	    data->min_work_area = 0;
	    data->expanded = 0;
	    set_handler( data->but1, scrollbar_button, data, wm );
	    set_handler( data->but2, scrollbar_button, data, wm );
	    if( data->compact_mode )
	    {
		data->but1->color = win->color;
		data->but2->color = win->color;
	    }
	    button_set_text_transp( data->but1, 150, wm );
	    button_set_text_transp( data->but2, 150, wm );
	    button_set_text_color( data->but1, wm->color2, wm );
	    button_set_text_color( data->but2, wm->color2, wm );
	    
	    data->name = 0;
	    data->values = 0;
	    data->cur = 0;
	    data->max_value = 0;
	    data->page_size = 0;
	    data->step_size = 1;
	    data->bar_selected = 0;
	    data->drag_start_x = 0;
	    data->drag_start_y = 0;
	    data->show_offset = 0;
	    data->one_pixel_size = 0;
	    
	    data->hex = CREATE_HEX_SCROLLBAR;
	    
	    data->prev_focus_win = 0;
	    	    
	    CREATE_VERTICAL_SCROLLBAR = 0;
	    CREATE_REVERSE_SCROLLBAR = 0;
	    CREATE_COMPACT_SCROLLBAR = 0;
	    CREATE_FLAT_SCROLLBAR = 0;
	    CREATE_HEX_SCROLLBAR = 0;
	    SCROLLBAR_BEGIN_HANDLER = 0;
	    SCROLLBAR_END_HANDLER = 0;
	    SCROLLBAR_BEGIN_HANDLER_DATA = 0;
	    SCROLLBAR_END_HANDLER_DATA = 0;
	    
	    retval = 1;
	    break;
	case EVT_BEFORECLOSE:
	    if( data->name )
	    {
		mem_free( data->name );
		data->name = 0;
	    }
	    if( data->values )
	    {
		mem_free( data->values );
		data->values = 0;
	    }
	    retval = 1;
	    break;
	case EVT_MOUSEBUTTONDOWN:
	case EVT_MOUSEMOVE:
	    {
		int inc = 0;
        	if( evt->key & MOUSE_BUTTON_SCROLLUP ) inc = 1;
        	if( evt->key & MOUSE_BUTTON_SCROLLDOWN ) inc = -1;
        	if( inc )
        	{
        	    if( data->vert ) inc = -inc;
		    if( data->begin == 0 )
		    {
		    	data->begin = 1;
			if( data->begin_handler )
			    data->begin_handler( data->begin_handler_data, data->this_window, scrollbar_get_value( data->this_window, wm ) );
		    }
		    if( inc > 0 )
		    {
			data->cur += data->step_size;
			if( data->cur < 0 ) data->cur = 0;
			if( data->cur > data->max_value ) data->cur = data->max_value;
		    }
		    else 
		    {
			data->cur -= data->step_size;
			if( data->cur < 0 ) data->cur = 0;
			if( data->cur > data->max_value ) data->cur = data->max_value;
		    }
		    draw_window( data->this_window, wm );
		    if( data->this_window->action_handler )
			data->this_window->action_handler( data->this_window->handler_data, data->this_window, wm );
		    if( data->begin )
		    {
		    	if( data->end_handler )
			    data->end_handler( data->end_handler_data, data->this_window, scrollbar_get_value( data->this_window, g_wm ) );
			data->begin = 0;
		    }
        	}
            }
	    if( evt->key == MOUSE_BUTTON_LEFT )
	    {
		if( evt->type == EVT_MOUSEBUTTONDOWN )
		{
		    if( data->begin == 0 )
		    {
			data->begin = 1;
			if( data->begin_handler )
			    data->begin_handler( data->begin_handler_data, win, scrollbar_get_value( win, wm ) );
		    }
		}
		
		int value_changed = 1;

		data->cur_x = rx;
		data->cur_y = ry;
		if( data->compact_mode == 0 )
		{
		    //Normal mode:
		    if( evt->type == EVT_MOUSEBUTTONDOWN && 
			rx >= 0 && rx < win->xsize && ry >= 0 && ry < win->ysize )
		    {
			if( data->vert )
			{
			    data->bar_selected = 1;
			    data->drag_start_y = ry;
			    data->drag_start_val = data->cur;
			}
			else
			{
			    data->bar_selected = 1;
			    data->drag_start_x = rx;
			    data->drag_start_val = data->cur;
			}
		    }
		    if( evt->type == EVT_MOUSEMOVE )
		    {
			//Move:
			if( data->bar_selected )
			{
			    int d = 0;
			    if( data->vert ) 
				d = ry - data->drag_start_y;
			    else
				d = rx - data->drag_start_x;
			    if( data->rev )
				data->cur = data->drag_start_val - (int)( (float)d * data->one_pixel_size );
			    else
				data->cur = data->drag_start_val + (int)( (float)d * data->one_pixel_size );
			}
		    }
		}
		else
		{
		    //Compact mode:
		    if( evt->type == EVT_MOUSEBUTTONDOWN &&
			rx >= 0 && ry >= 0 &&
			rx < win->xsize && ry < win->ysize )
		    {
			if( data->but1 && data->but1->visible == 0 )
			{
			    int new_xsize = SCROLLBAR_SIZE * 10;
			    int new_ysize = SCROLLBAR_SIZE * 2;
			    if( new_xsize > wm->screen_xsize )
				new_xsize = wm->screen_xsize;
			    int new_x = win->screen_x + ( win->xsize - new_xsize ) / 2;
			    int new_y = win->screen_y + ( win->ysize - new_ysize ) / 2;
			    if( new_x + new_xsize > wm->screen_xsize ) new_x -= new_x + new_xsize - wm->screen_xsize;
			    if( new_y + new_ysize > wm->screen_ysize ) new_y -= new_y + new_ysize - wm->screen_ysize;
			    if( new_x < 0 ) new_x = 0;
			    if( new_y < 0 ) new_y = 0;
			    data->expanded = new_window( "Expanded controller", new_x, new_y, new_xsize, new_ysize, wm->color0, wm->root_win, scrollbar_expanded_handler, wm );
			    scrollbar_expanded_data* edata = (scrollbar_expanded_data*)data->expanded->data;
			    edata->ctl = win;
			    show_window( data->expanded, wm );
			    recalc_regions( wm );
			    draw_window( data->expanded, wm );
			}
			data->bar_selected = 1;
			data->drag_start_x = rx;
			data->drag_start_y = ry;
			data->drag_start_val = data->cur;
		    }
		    if( data->bar_selected == 1 && evt->type == EVT_MOUSEMOVE )
		    {
			if( data->working_area > 1 )
			{
			    int new_val;
			    int dx = rx - data->drag_start_x;
			    int dy = ry - data->drag_start_y;
			    int s;
			    if( data->expanded == 0 )
			    {
				s = data->working_area - 1;
				new_val = ( ( dx << 12 ) / s ) * data->max_value;
			    }
			    else
			    {
				s = data->expanded->xsize - INTERELEMENT_SPACE * 2;
				if( data->values )
				{
				    if( s / ( SCROLLBAR_SIZE / 2 ) > data->max_value + 1 )
					s = ( data->max_value + 1 ) * ( SCROLLBAR_SIZE / 2 );
				}
				new_val = ( ( dx << 12 ) / s ) * data->max_value;
				new_val -= ( ( dy << 12 ) / s ) * data->max_value;
			    }
			    new_val >>= 12;
			    new_val += data->drag_start_val;
			    if( data->expanded && data->max_value + 1 > s )
			    {
				int c = data->max_value / 2 - 1;
				if( new_val < data->max_value / 2 )
				{
				    new_val = ( ( c - ( c - new_val ) ) * ( data->max_value / 2 ) ) / c;
				    if( new_val > data->max_value / 2 ) 
					new_val = data->max_value / 2;
				}
			    }
			    if( data->cur == new_val ) 
				value_changed = 0;
			    else
				data->cur = new_val;
			}
		    }
		    else
		    {
			value_changed = 0;
		    }
		}
		//Bounds control:
		if( data->cur < 0 ) data->cur = 0;
		if( data->cur > data->max_value ) data->cur = data->max_value;
		//Redraw it:
		draw_window( win, wm );
		if( data->expanded )
		    draw_window( data->expanded, wm );
		//User handler:
		if( win->action_handler && value_changed )
		    win->action_handler( win->handler_data, win, wm );
	    }
	    retval = 1;
	    break;
	case EVT_FOCUS:
	    data->prev_focus_win = wm->prev_focus_win;
	    retval = 1;
	    break;
	case EVT_UNFOCUS:
	    if( data->bar_selected )
	    {
		data->bar_selected = 0;
		if( data->expanded )
                {
                    remove_window( data->expanded, wm );
                    data->expanded = 0;
                    recalc_regions( wm );
                    draw_window( wm->root_win, wm );
                }
		draw_window( win, wm );
		if( data->begin )
		{
		    data->begin = 0;
		    if( data->end_handler )
			data->end_handler( data->end_handler_data, data->this_window, scrollbar_get_value( data->this_window, g_wm ) );
		}
		retval = 1;
	    }
	    break;
	case EVT_MOUSEBUTTONUP:
	    if( evt->key == MOUSE_BUTTON_LEFT )
	    {
		data->bar_selected = 0;
		if( data->expanded )
		{
		    remove_window( data->expanded, wm );
		    data->expanded = 0;
		    recalc_regions( wm );
		    draw_window( wm->root_win, wm );
		}
		set_focus_win( data->prev_focus_win, wm );
		draw_window( win, wm );
		//User handler:
		if( win->action_handler )
		    win->action_handler( win->handler_data, win, wm );
		if( data->begin )
		{
		    data->begin = 0;
		    if( data->end_handler )
			data->end_handler( data->end_handler_data, data->this_window, scrollbar_get_value( data->this_window, g_wm ) );
		}
	    }
	    retval = 1;
	    break;
	case EVT_DRAW:
	    {
		scrollbar_calc_working_area( win );
		scrollbar_get_strings( win );
		
		UTF8_CHAR* name_str = data->name_str;
		if( data->working_area < data->min_work_area )
	        {
    		    if( name_str )
		    {
	    		int p = 0;
	    		while( 1 )
	    		{
			    int c = name_str[ p ];
			    if( c == 0 ) break;
			    if( c == '.' )
			    {
		    		name_str += p + 1;
				break;
			    }
			    if( ( c >= '0' && c <= '9' ) || ( c >= 'A' && c <= 'F' ) )
			    {
			    }
			    else break;
			    p++;
			}
		    }
		}

		wbd_lock( win );

		if( data->compact_mode == 0 )
		{
		    //Normal mode:
		    
		    if( data->max_value == 0 || data->page_size == 0 )
		    {
			data->bar_size = data->working_area;
			data->pos = 0;
		    }
		    else
		    {
			//Calculate move-region (in pixels)
			float ppage;
			if( data->max_value + 1 == 0 ) 
			    ppage = 1.0F;
			else
			    ppage = (float)data->page_size / (float)( data->max_value + 1 );
			if( ppage == 0 ) ppage = 1.0F;
			data->move_region = (int)( (float)data->working_area / ( ppage + 1.0F ) );
			if( data->move_region == 0 ) data->move_region = 1;

			//Caclulate slider size (in pixels)
			data->bar_size = data->working_area - data->move_region;

			//Calculate one pixel size
			data->one_pixel_size = (float)( data->max_value + 1 ) / (float)data->move_region;

			//Calculate slider position (in pixels)
			data->pos = (int)( ( (float)data->cur / (float)data->max_value ) * (float)data->move_region );

			if( data->bar_size < 2 ) data->bar_size = 2;
			if( data->rev )
			    data->pos = ( data->working_area - data->pos ) - data->bar_size;

			if( data->pos + data->bar_size > data->working_area )
			{
			    data->bar_size -= ( data->pos + data->bar_size ) - data->working_area;
			}
		    }

		    bool frozen = 0;
		    if( data->bar_size >= data->working_area ) frozen = 1;
		    COLOR bar_color;
		    COLOR bg_color;
		    COLOR marker_color;
		    bg_color = win->parent->color;
		    bar_color = blend( wm->color2, bg_color, 220 );
		    marker_color = blend( bar_color, bg_color, 128 );
		    if( frozen )
		    {
			button_set_text_transp( data->but1, 64, wm );
			button_set_text_transp( data->but2, 64, wm );
			draw_box( 0, 0, win->xsize, win->ysize, bg_color, wm );
		    }
		    else 
		    {
			button_set_text_transp( data->but1, 150, wm );
			button_set_text_transp( data->but2, 150, wm );
			if( data->vert )
			{
			    draw_box( 0, 0, win->xsize, data->pos, bg_color, wm );
			    draw_box( 0, data->pos + data->bar_size, win->xsize, data->working_area - ( data->pos + data->bar_size ), bg_color, wm );

			    int ss = 2;
			    draw_box( 0, 0, win->xsize, ss, marker_color, wm );
			    draw_box( 0, data->working_area - ss, win->xsize, ss, marker_color, wm );
			    
			    draw_box( 0, data->pos, win->xsize, data->bar_size, bar_color, wm );
			    draw_box( 0, data->pos - 1, win->xsize, 1, BORDER_COLOR( bar_color ), wm );
			    draw_box( 0, data->pos + data->bar_size, win->xsize, 1, BORDER_COLOR( bar_color ), wm );
			}
			else
			{
			    draw_box( 0, 0, data->pos, win->ysize, bg_color, wm );
			    draw_box( data->pos + data->bar_size, 0, data->working_area - ( data->pos + data->bar_size ), win->ysize, bg_color, wm );
			    
			    int ss = 2;
			    draw_box( 0, 0, ss, win->xsize, marker_color, wm );
			    draw_box( data->working_area - ss, 0, ss, win->xsize, marker_color, wm );
			    
			    draw_box( data->pos, 0, data->bar_size, win->ysize, bar_color, wm );
			    draw_box( data->pos - 1, 0, 1, win->ysize, BORDER_COLOR( bar_color ), wm );
			    draw_box( data->pos + data->bar_size, 0, 1, win->ysize, BORDER_COLOR( bar_color ), wm );
			}
			if( data->bar_selected )
			{
			    wm->cur_transparency = 64;
			    draw_box( 0, 0, win->xsize, win->ysize, wm->green, wm );
			    if( data->vert )
				draw_scrollbar_vertical_selection( win, data->cur_y );
			    else 
				draw_scrollbar_horizontal_selection( win, data->cur_x );
			    bg_color = blend( bg_color, wm->green, 64 );
			}
		    }
		    		    
		    data->but1->color = bg_color;
		    data->but2->color = bg_color;
		}
		else
		{
		    //Compact mode:
		    		    
		    int start_x = 0;
		    COLOR bgcolor = win->color;
		    if( data->bar_selected )
		    {
			bgcolor = blend( bgcolor, wm->green, 64 );
		    }
		    data->but1->color = bgcolor;
		    data->but2->color = bgcolor;
		    draw_box( 0, 0, win->xsize, win->ysize, bgcolor, wm );
		    
		    int ychar = wbd_char_y_size( wm );

#define SCROLL_XBORDER 0
#define SCROLL_YBORDER 0
		    
		    int xc = 0;
		    if( data->max_value != 0 )
			xc = ( ( data->working_area - SCROLL_XBORDER * 2 ) * ( ( data->cur << 10 ) / data->max_value ) ) >> 10;

		    COLOR fgcolor = wm->color2;
		    
		    wm->cur_transparency = 256;
		    if( data->but1->visible == 0 )
		    {
			COLOR lr_color = blend( fgcolor, bgcolor, 230 );
			draw_box( start_x + SCROLL_XBORDER, SCROLL_YBORDER, 1, win->ysize - SCROLL_YBORDER * 2, lr_color, wm );
			draw_box( start_x + SCROLL_XBORDER + data->working_area - 1, SCROLL_YBORDER, 1, win->ysize - SCROLL_YBORDER * 2, lr_color, wm );
		    }
		    draw_box( start_x + SCROLL_XBORDER, SCROLL_YBORDER, xc, win->ysize - SCROLL_YBORDER * 2, blend( fgcolor, bgcolor, 180 ), wm );
		    
		    if( data->bar_selected )
		    {
			//Show selection:
			draw_scrollbar_horizontal_selection( win, data->cur_x );
		    }

		    if( name_str )
		    {
			wm->cur_font_color = wm->color2;
			wm->cur_transparency = 256;
			draw_string( name_str, start_x + SCROLL_XBORDER, win->ysize / 2 - ychar, wm );
		    }
		    
		    if( data->bar_selected )
			wm->cur_font_color = wm->color3;
		    else 
			wm->cur_font_color = blend( wm->color2, wm->blue, 64 );
		    wm->cur_transparency = 256;
		    draw_string( data->value_str, start_x + SCROLL_XBORDER, win->ysize / 2 + 1, wm );

		    //Show center:
		    wm->cur_transparency = 50;
		    int cx = SCROLL_XBORDER + data->working_area / 2 - 1;
		    draw_box( cx - 1, win->ysize - SCROLL_YBORDER - 1, 3, 1, wm->color3, wm );
		    draw_box( cx, win->ysize - SCROLL_YBORDER - 2, 1, 1, wm->color3, wm );
		}
		
		wbd_draw();
    		wbd_unlock();
	    }
	    retval = 1;
	    break;
	case EVT_AFTERRESIZE:
	    scrollbar_calc_working_area( win );
	    retval = 1;
	    break;
	case EVT_TOUCHBEGIN:
	case EVT_TOUCHEND:
	case EVT_TOUCHMOVE:
	    retval = 1;
	    break;
    }
    return retval;
}

void scrollbar_set_parameters( WINDOWPTR win, int cur, int max_value, int page_size, int step_size, window_manager* wm )
{
    if( win )
    {
	scrollbar_data* data = (scrollbar_data*)win->data;
	if( data )
	{
	    data->cur = cur;
	    if( max_value >= 0 ) data->max_value = max_value;
	    if( page_size >= 0 ) data->page_size = page_size;
	    if( step_size >= 0 ) data->step_size = step_size;
	}
    }
}

void scrollbar_set_value( WINDOWPTR win, int val, window_manager* wm )
{
    if( win )
    {
	scrollbar_data* data = (scrollbar_data*)win->data;
	if( data )
	{
	    data->cur = val;
	    if( data->cur < 0 ) data->cur = 0;
	    if( data->cur > data->max_value ) data->cur = data->max_value;
	}
    }
}

int scrollbar_get_value( WINDOWPTR win, window_manager* wm )
{
    if( win )
    {
	scrollbar_data* data = (scrollbar_data*)win->data;
	if( data )
	{
	    return data->cur;
	}
    }
    return 0;
}

void scrollbar_set_name( WINDOWPTR win, const UTF8_CHAR* name, window_manager* wm )
{
    if( win )
    {
	scrollbar_data* data = (scrollbar_data*)win->data;
	if( data )
	{
	    if( name == 0 )
	    {
		mem_free( data->name );
		data->name = 0;
		return;
	    }
	    if( data->name == 0 )
	    {
		data->name = (UTF8_CHAR*)MEM_NEW( HEAP_DYNAMIC, mem_strlen( name ) + 1 );
	    }
	    else
	    {
		if( mem_strlen( name ) + 1 > mem_get_size( data->name ) )
		{
		    data->name = (UTF8_CHAR*)mem_resize( data->name, mem_strlen( name ) + 1 );
		}
	    }
	    mem_copy( data->name, name, mem_strlen( name ) + 1 );
	}
    }
}

void scrollbar_set_values( WINDOWPTR win, const UTF8_CHAR* values, window_manager* wm )
{
    if( win )
    {
	scrollbar_data* data = (scrollbar_data*)win->data;
	if( data )
	{
	    if( values == 0 )
	    {
		mem_free( data->values );
		data->values = 0;
		return;
	    }
	    size_t len = mem_strlen( values );
	    if( data->values == 0 )
	    {
		data->values = (UTF8_CHAR*)MEM_NEW( HEAP_DYNAMIC, len + 1 );
	    }
	    else
	    {
		if( len + 1 > mem_get_size( data->values ) )
		{
		    data->values = (UTF8_CHAR*)mem_resize( data->values, len + 1 );
		}
	    }
	    const UTF8_CHAR* src = values;
	    UTF8_CHAR* dest = data->values;
	    bool slash = 0;
	    while( 1 )
	    {
		UTF8_CHAR c = *src;
		*dest = c;
		if( c == '/' ) 
		{
		    if( slash )
		    {
			dest--;
			*dest = 1;
			slash = 0;
		    }
		    else
		    {
			slash = 1;
		    }
		}
		else
		{
		    slash = 0;
		}
		if( c == 0 ) break;
		dest++;
		src++;
	    }
	}
    }
}

void scrollbar_set_showing_offset( WINDOWPTR win, int offset, window_manager *wm )
{
    if( win )
    {
	scrollbar_data *data = (scrollbar_data*)win->data;
	if( data )
	{
	    data->show_offset = offset;
	}
    }
}

void scrollbar_set_hex_mode( WINDOWPTR win, int hex, window_manager *wm )
{
    if( win )
    {
	scrollbar_data *data = (scrollbar_data*)win->data;
	if( data )
	{
	    data->hex = hex;
	}
    }
}
