/*
    wbd.cpp. WBD (Window Buffer Draw) functions
    This file is part of the SunDog engine.
    Copyright (C) 2002 - 2013 Alexander Zolotov <nightradio@gmail.com>
    www.warmplace.ru
*/

//How to use it in a window handler:
//EVT_DRAW:	
//  win_draw_lock( win, wm );
//  wbd_lock( win );
//  ...
//  wbd_draw();
//  wbd_unlock();
//  win_draw_unlock( win, wm );

#include "../../core/core.h"
#include "../../main/user_code.h"
#include "../../core/debug.h"
#include "../../utils/utils.h"
#include "../wmanager.h"

#define LINE_PREC 20

void wbd_lock( WINDOWPTR win )
{
    if( win )
    {
#if defined(OPENGL) && !defined(FRAMEBUFFER)
	g_wm->cur_window = win;
	if( win->xsize == 0 ||
	    win->ysize == 0 )
	{
	    //Wrong window size:
	    g_wm->cur_window = 0;
	}
#else
	if( g_wm->g_screen == 0 )
	{
	    g_wm->g_screen = (COLORPTR)MEM_NEW( HEAP_DYNAMIC, g_wm->screen_xsize * g_wm->screen_ysize * COLORLEN );
	}
	else
	{
    	    size_t real_size = mem_get_size( g_wm->g_screen ) / COLORLEN;
    	    size_t new_size = g_wm->screen_xsize * g_wm->screen_ysize;
    	    if( new_size > real_size )
    	    {
    		//Resize:
    		g_wm->g_screen = (COLORPTR)mem_resize( g_wm->g_screen, ( new_size + ( new_size / 4 ) ) * COLORLEN );
	    }
	}
	g_wm->cur_window = win;
	if( win->xsize == 0 ||
	    win->ysize == 0 )
	{
	    //Wrong window size:
	    g_wm->cur_window = 0;
	}
#endif
	g_wm->cur_font_color = COLORMASK;
	g_wm->cur_transparency = 255;
	g_wm->cur_font_zoom = 256;
    }
}

void wbd_unlock( void )
{
    g_wm->cur_transparency = 255;
    g_wm->cur_window = 0;
}

void wbd_draw( void )
{
#if defined(OPENGL) && !defined(FRAMEBUFFER)
#else
    WINDOWPTR win = g_wm->cur_window;
    sundog_image img;
    img.xsize = g_wm->screen_xsize;
    img.ysize = g_wm->screen_ysize;
    img.flags = IMAGE_NATIVE_RGB | IMAGE_STATIC_SOURCE;
    img.data = g_wm->g_screen;
    win_draw_bitmap_ext( win, 0, 0, win->xsize, win->ysize, win->screen_x, win->screen_y, &img, g_wm );
#endif
}

void draw_line( int x1, int y1, int x2, int y2, COLOR color, window_manager* wm )
{
    if( wm->cur_transparency == 0 ) return;
    
#if defined(OPENGL) && !defined(FRAMEBUFFER)
    WINDOWPTR win = wm->cur_window;
    if( win && win->visible && win->reg && win->reg->numRects )
    {
	x1 += win->screen_x;
	y1 += win->screen_y;
	x2 += win->screen_x;
	y2 += win->screen_y;
	for( int r = 0; r < win->reg->numRects; r++ )
	{
	    int rx1 = win->reg->rects[ r ].left;
	    int rx2 = win->reg->rects[ r ].right;
	    int ry1 = win->reg->rects[ r ].top;
	    int ry2 = win->reg->rects[ r ].bottom;
	    
	    int lx1 = x1;
	    int ly1 = y1;
	    int lx2 = x2;
	    int ly2 = y2;

	    if( line_clip( &lx1, &ly1, &lx2, &ly2, rx1, ry1, rx2, ry2 ) )
	    {
		device_draw_line( lx1, ly1, lx2, ly2, color );
		g_wm->screen_changed++;
	    }
	}
    }
#else

    x1 += wm->cur_window->screen_x;
    y1 += wm->cur_window->screen_y;
    x2 += wm->cur_window->screen_x;
    y2 += wm->cur_window->screen_y;

    int screen_xsize = wm->screen_xsize;
    int screen_ysize = wm->screen_ysize;
    COLORPTR g_screen = wm->g_screen;
    
    int clip_x1 = wm->cur_window->screen_x;
    int clip_y1 = wm->cur_window->screen_y;
    int clip_x2 = clip_x1 + wm->cur_window->xsize;
    int clip_y2 = clip_y1 + wm->cur_window->ysize;
    if( clip_x1 < 0 ) clip_x1 = 0;
    if( clip_y1 < 0 ) clip_y1 = 0;
    if( clip_x2 > screen_xsize ) clip_x2 = screen_xsize;
    if( clip_y2 > screen_ysize ) clip_y2 = screen_ysize;
    
    if( x1 == x2 )
    {
	if( x1 >= clip_x1 && x1 < clip_x2 )
	{
	    g_wm->screen_changed++;
	    
	    if( y1 > y2 ) { int temp = y1; y1 = y2; y2 = temp; }
	    if( y1 < clip_y1 ) y1 = clip_y1;
	    if( y1 >= clip_y2 ) return;
	    if( y2 < clip_y1 ) return;
	    if( y2 >= clip_y2 ) y2 = clip_y2 - 1;
	    int ptr = y1 * screen_xsize + x1;
	    int transp = wm->cur_transparency;
	    if( transp >= 256 )
	    {
		for( ; y1 <= y2; y1++ )
		{
		    g_screen[ ptr ] = color;
		    ptr += screen_xsize;
		}
	    }
	    else
	    {
		for( ; y1 <= y2; y1++ )
		{
		    g_screen[ ptr ] = blend( g_screen[ ptr ], color, transp );
		    ptr += screen_xsize;
		}
	    }
	}
	return;
    }
    if( y1 == y2 )
    {
	if( y1 >= 0 && y1 < screen_ysize )
	{
	    g_wm->screen_changed++;
	    
	    if( x1 > x2 ) { int temp = x1; x1 = x2; x2 = temp; }
	    if( x1 < clip_x1 ) x1 = clip_x1;
	    if( x1 >= clip_x2 ) return;
	    if( x2 < clip_x1 ) return;
	    if( x2 >= clip_x2 ) x2 = clip_x2 - 1;
	    int ptr = y1 * screen_xsize + x1;
	    int transp = wm->cur_transparency;
	    if( transp >= 256 )
	    {
		for( ; x1 <= x2; x1++ )
		{
		    g_screen[ ptr ] = color;
		    ptr ++;
		}
	    }
	    else
	    {
		for( ; x1 <= x2; x1++ )
		{
		    g_screen[ ptr ] = blend( g_screen[ ptr ], color, transp );
		    ptr ++;
		}
	    }
	}
	return;
    }

    if( line_clip( &x1, &y1, &x2, &y2, clip_x1, clip_y1, clip_x2, clip_y2 ) )
    {
	//Draw line:
	g_wm->screen_changed++;
	bool x_inc;
	bool y_inc;
	int len_x = x2 - x1; if( len_x < 0 ) { len_x = -len_x; x_inc = 0; } else { x_inc = 1; }
	int len_y = y2 - y1; if( len_y < 0 ) { len_y = -len_y; y_inc = 0; } else { y_inc = 1; }
	int ptr = y1 * screen_xsize + x1;
	int delta;
	int v = 0, old_v = 0;
	int transp = wm->cur_transparency;
	if( len_x > len_y )
	{
	    //Horisontal:
	    if( len_x != 0 )
	    {
		delta = ( len_y << LINE_PREC ) / len_x;
		delta += ( ( 1 << LINE_PREC ) / 2 ) / len_x;
	    }
	    else
		delta = 0;
	    if( transp >= 256 )
		for( int a = 0; a <= len_x; a++ )
		{
		    g_screen[ ptr ] = color;
		    old_v = v;
		    v += delta;
		    if( x_inc ) ptr++; else ptr--;
		    if( ( old_v >> LINE_PREC ) != ( v >> LINE_PREC ) ) 
		    {
			if( y_inc )
			    ptr += screen_xsize;
			else
			    ptr -= screen_xsize;
		    }
		}
	    else
		for( int a = 0; a <= len_x; a++ )
		{
		    g_screen[ ptr ] = blend( g_screen[ ptr ], color, transp );
		    old_v = v;
		    v += delta;
		    if( x_inc ) ptr++; else ptr--;
		    if( ( old_v >> LINE_PREC ) != ( v >> LINE_PREC ) ) 
		    {
			if( y_inc )
			    ptr += screen_xsize;
			else
			    ptr -= screen_xsize;
		    }
		}
	}
	else
	{
	    //Vertical:
	    if( len_y != 0 )
	    {
		delta = ( len_x << LINE_PREC ) / len_y;
		delta += ( ( 1 << LINE_PREC ) / 2 ) / len_y;
	    }
	    else
		delta = 0;
	    if( transp >= 256 )
		for( int a = 0; a <= len_y; a++ )
		{
		    g_screen[ ptr ] = color;
		    old_v = v;
		    v += delta;
		    if( y_inc ) 
			ptr += screen_xsize;
		    else
			ptr -= screen_xsize;
		    if( ( old_v >> LINE_PREC ) != ( v >> LINE_PREC ) ) 
		    {
			if( x_inc ) ptr++; else ptr--;
		    }
		}
	    else
		for( int a = 0; a <= len_y; a++ )
		{
		    g_screen[ ptr ] = blend( g_screen[ ptr ], color, transp );
		    old_v = v;
		    v += delta;
		    if( y_inc ) 
			ptr += screen_xsize;
		    else
			ptr -= screen_xsize;
		    if( ( old_v >> LINE_PREC ) != ( v >> LINE_PREC ) ) 
		    {
			if( x_inc ) ptr++; else ptr--;
		    }
		}
	}
    }
#endif
}

void draw_line_with_transp( int x1, int y1, int t1, int x2, int y2, int t2, COLOR color )
{
    if( g_wm->cur_transparency == 0 ) return;
    
#if defined(OPENGL) && !defined(FRAMEBUFFER)
    WINDOWPTR win = g_wm->cur_window;
    if( win && win->visible && win->reg && win->reg->numRects )
    {
	x1 += win->screen_x;
	y1 += win->screen_y;
	x2 += win->screen_x;
	y2 += win->screen_y;
	for( int r = 0; r < win->reg->numRects; r++ )
	{
	    int rx1 = win->reg->rects[ r ].left;
	    int rx2 = win->reg->rects[ r ].right;
	    int ry1 = win->reg->rects[ r ].top;
	    int ry2 = win->reg->rects[ r ].bottom;
	    
	    int lx1 = x1;
	    int ly1 = y1;
	    int lt1 = t1;
	    int lx2 = x2;
	    int ly2 = y2;
	    int lt2 = t2;
	    
	    if( line_clip_with_transp( &lx1, &ly1, &lt1, &lx2, &ly2, &lt2, rx1, ry1, rx2, ry2 ) )
	    {
		device_draw_line_with_transp( lx1, ly1, lt1, lx2, ly2, lt2, color );
		g_wm->screen_changed++;
	    }
	}
    }
#else
#endif    
}

void draw_frame( int x, int y, int x_size, int y_size, COLOR color, window_manager* wm )
{
    draw_line( x, y, x + x_size, y, color, wm );
    draw_line( x + x_size, y, x + x_size, y + y_size, color, wm );
    draw_line( x + x_size, y + y_size, x, y + y_size, color, wm );
    draw_line( x, y + y_size, x, y, color, wm );
}

void draw_box( int x, int y, int x_size, int y_size, COLOR color, window_manager* wm )
{
    if( wm->cur_transparency == 0 ) return;

    WINDOWPTR win = wm->cur_window;
    if( win && win->visible && win->reg && win->reg->numRects )
    {
	x += win->screen_x;
	y += win->screen_y;
	if( win->reg->numRects )
	{
#if defined(OPENGL) && !defined(FRAMEBUFFER)
	    for( int r = 0; r < win->reg->numRects; r++ )
	    {
		int rx1 = win->reg->rects[ r ].left;
		int rx2 = win->reg->rects[ r ].right;
		int ry1 = win->reg->rects[ r ].top;
		int ry2 = win->reg->rects[ r ].bottom;

		//Control box size:
		int nx = x;
		int ny = y;
		int nxsize = x_size;
		int nysize = y_size;
		if( nx < rx1 ) { nxsize -= ( rx1 - nx ); nx = rx1; }
		if( ny < ry1 ) { nysize -= ( ry1 - ny ); ny = ry1; }
		if( nx + nxsize <= rx1 ) continue;
		if( ny + nysize <= ry1 ) continue;
		if( nx + nxsize > rx2 ) nxsize -= nx + nxsize - rx2;
		if( ny + nysize > ry2 ) nysize -= ny + nysize - ry2;
		if( nx >= rx2 ) continue;
		if( ny >= ry2 ) continue;
		if( nxsize < 0 ) continue;
		if( nysize < 0 ) continue;
        	
		//Draw it:
		device_draw_box( nx, ny, nxsize, nysize, color );
		g_wm->screen_changed++;
	    }
#else
	    if( x < 0 ) { x_size += x; x = 0; }
	    if( y < 0 ) { y_size += y; y = 0; }
	    if( x + x_size <= 0 ) return;
	    if( y + y_size <= 0 ) return;
	    
	    int screen_xsize = wm->screen_xsize;
	    int screen_ysize = wm->screen_ysize;
	    
	    if( x + x_size > screen_xsize ) x_size = screen_xsize - x;
	    if( y + y_size > screen_ysize ) y_size = screen_ysize - y;
	    if( x >= screen_xsize ) return;
	    if( y >= screen_ysize ) return;
	    if( x_size < 0 ) return;
	    if( y_size < 0 ) return;
	    
	    g_wm->screen_changed++;
	    
	    COLORPTR ptr = wm->g_screen + y * screen_xsize + x;
	    int transp = wm->cur_transparency;
	    int add = screen_xsize - x_size;
	    if( transp >= 255 )
	    {
		for( int cy = 0; cy < y_size; cy++ )
		{
		    COLORPTR size = ptr + x_size;
		    while( ptr < size ) { *ptr = color; ptr++; }
		    ptr += add;
		}
	    }
	    else
	    {
		for( int cy = 0; cy < y_size; cy++ )
		{
		    COLORPTR size = ptr + x_size;
		    while( ptr < size ) { *ptr = blend( *ptr, color, transp ); ptr++; }
		    ptr += add;
		}
	    }
#endif
	}
    }
}

void draw_hgradient( int x, int y, int xsize, int ysize, COLOR c, int t1, int t2 )
{
    if( g_wm->cur_transparency == 0 ) return;
    
    sundog_polygon p;
    sundog_vertex v[ 4 ];
    v[ 0 ].x = x;
    v[ 0 ].y = y;
    v[ 0 ].c = c;
    v[ 0 ].t = t1;
    v[ 1 ].x = x + xsize;
    v[ 1 ].y = y;
    v[ 1 ].t = t2;
    v[ 2 ].x = x + xsize;
    v[ 2 ].y = y + ysize;
    v[ 2 ].t = t2;
    v[ 3 ].x = x;
    v[ 3 ].y = y + ysize;
    v[ 3 ].t = t1;
    p.vnum = 4;
    p.v = v;
    g_wm->cur_flags = WBD_FLAG_ONE_COLOR;
    draw_polygon( &p );
}

void draw_vgradient( int x, int y, int xsize, int ysize, COLOR c, int t1, int t2 )
{
    if( g_wm->cur_transparency == 0 ) return;
    
#if defined(OPENGL) && !defined(FRAMEBUFFER)
    sundog_polygon p;
    sundog_vertex v[ 4 ];
    v[ 0 ].x = x;
    v[ 0 ].y = y;
    v[ 0 ].c = c;
    v[ 0 ].t = t1;
    v[ 1 ].x = x + xsize;
    v[ 1 ].y = y;
    v[ 1 ].t = t1;
    v[ 2 ].x = x + xsize;
    v[ 2 ].y = y + ysize;
    v[ 2 ].t = t2;
    v[ 3 ].x = x;
    v[ 3 ].y = y + ysize;
    v[ 3 ].t = t2;
    p.vnum = 4;
    p.v = v;
    g_wm->cur_flags = WBD_FLAG_ONE_COLOR;
    draw_polygon( &p );
#else
    WINDOWPTR win = g_wm->cur_window;
    if( win && win->visible && win->reg && win->reg->numRects )
    {
	x += win->screen_x;
	y += win->screen_y;
	if( win->reg->numRects )
	{
	    if( x < 0 ) { xsize += x; x = 0; }
	    if( y < 0 ) { if( ysize ) { t1 += ( -y * (t2-t1) ) / ysize; } ysize += y; y = 0; }
	    if( x + xsize <= 0 ) return;
	    if( y + ysize <= 0 ) return;
	    
	    int screen_xsize = g_wm->screen_xsize;
	    int screen_ysize = g_wm->screen_ysize;
	    
	    if( x + xsize > screen_xsize ) xsize = screen_xsize - x;
	    if( y + ysize > screen_ysize ) 
	    {
		if( ysize )
		{
		    t2 = t1 + ( ( screen_ysize - y ) * (t2-t1) ) / ysize;
		}
		ysize = screen_ysize - y;
	    }
	    if( x >= screen_xsize ) return;
	    if( y >= screen_ysize ) return;
	    if( xsize < 0 ) return;
	    if( ysize < 0 ) return;
	    
	    g_wm->screen_changed++;
	    
	    COLORPTR ptr = g_wm->g_screen + y * screen_xsize + x;
	    int transp = g_wm->cur_transparency;
	    if( transp < 255 )
	    {
		t1 *= transp; t1 /= 256;
		t2 *= transp; t2 /= 256;
	    }
	    int add = screen_xsize - xsize;
	    if( t1 >= 255 && t2 >= 255 )
	    {
		for( int cy = 0; cy < ysize; cy++ )
		{
		    COLORPTR size = ptr + xsize;
		    while( ptr < size ) { *ptr = c; ptr++; }
		    ptr += add;
		}
	    }
	    else
	    {
		int tdelta;
		if( ysize )
		    tdelta = ( ( t2 - t1 ) << 15 ) / ysize;
		else 
		    tdelta = 0;
		int t = t1 << 15;
		for( int cy = 0; cy < ysize; cy++ )
		{
		    COLORPTR size = ptr + xsize;
		    int tt = t >> 15;
		    while( ptr < size ) { *ptr = blend( *ptr, c, tt ); ptr++; }
		    t += tdelta;
		    ptr += add;
		}
	    }
	}
    }
#endif
}

void draw_glass_rect( int x, int y, int xsize, int ysize, int parts )
{
    if( g_wm->cur_transparency == 0 ) return;
    
    if( parts & 1 ) 
    {
	draw_vgradient( x, y, xsize, ysize / 1, g_wm->color0, 70, 0 );
    }
    if( parts & 2 )
    {
	draw_vgradient( x, y + 1, xsize, 2, g_wm->color3, 0, 64 );
	draw_vgradient( x, y + 3, xsize, 2, g_wm->color3, 64, 0 );
    }
    if( parts & 4 )
    {
	draw_vgradient( x, y + ysize - ysize / 4, xsize, ysize / 4, g_wm->color3, 0, 70 );
    }
}

void draw_corners( int x, int y, int xsize, int ysize, int size, int len, COLOR c )
{
    draw_box( x - size, y, size, len, c, g_wm );
    draw_box( x - size, y - size, size + len, size, c, g_wm );
    
    draw_box( x + xsize, y, size, len, c, g_wm );
    draw_box( x + xsize - len, y - size, size + len, size, c, g_wm );
    
    draw_box( x - size, y + ysize - len, size, len, c, g_wm );
    draw_box( x - size, y + ysize, size + len, size, c, g_wm );
    
    draw_box( x + xsize, y + ysize - len, size, len, c, g_wm );
    draw_box( x + xsize - len, y + ysize, size + len, size, c, g_wm );    
}

void draw_corners2( int x, int y, int xsize, int ysize, int size, int len, COLOR c )
{
    draw_box( x + xsize, y, size, len, c, g_wm );
    draw_box( x + xsize - len, y - size, size + len, size, c, g_wm );
    
    draw_box( x - size, y + ysize - len, size, len, c, g_wm );
    draw_box( x - size, y + ysize, size + len, size, c, g_wm );
}

void polygon_clipping_stage( sundog_polygon* output, sundog_polygon* input, int16 clip, bool clip_direction, bool clip_vertical )
{
    input->vnum = output->vnum;
    mem_copy( input->v, output->v, input->vnum * sizeof( sundog_vertex ) );
    
    output->vnum = 0;
    
    sundog_vertex* s = &input->v[ input->vnum - 1 ];
    bool s_inside;
    if( clip_vertical )
    {
	if( s->x >= clip ) s_inside = 1; else s_inside = 0;
    }
    else 
    {
	if( s->y >= clip ) s_inside = 1; else s_inside = 0;
    }
    s_inside ^= clip_direction;
    for( int input_v = 0; input_v < input->vnum; input_v++ )
    {
	sundog_vertex* e = &input->v[ input_v ];
	
	bool e_inside;
	if( clip_vertical )
	{
	    if( e->x >= clip ) e_inside = 1; else e_inside = 0;
	}
	else 
	{
	    if( e->y >= clip ) e_inside = 1; else e_inside = 0;
	}
	e_inside ^= clip_direction;
	
	bool add_intersection = 0;
	bool add_e = 0;
	
	if( e_inside )
	{
	    if( !s_inside )
	    {
		add_intersection = 1;
	    }
	    add_e = 1;
	}
	else 
	{
	    if( s_inside )
	    {
		add_intersection = 1;
	    }
	}
	
	sundog_vertex* output_v;
	if( add_intersection )
	{
	    output_v = &output->v[ output->vnum ];
	    if( clip_vertical )
	    {
		output_v->x = clip;
		int len = e->x - s->x;
		if( len )
		{
		    if( len < 0 ) len = -len;
		    int d = ( (int)( e->y - s->y ) * 4096 ) / len;
		    if( clip >= s->x )
		    {
			d *= ( clip - s->x );
		    }
		    else
		    {
			d *= ( s->x - clip );
		    }
		    d /= 4096;
		    output_v->y = s->y + d;
		    if( ( g_wm->cur_flags & ( WBD_FLAG_ONE_TRANSP | WBD_FLAG_ONE_COLOR ) ) != ( WBD_FLAG_ONE_TRANSP | WBD_FLAG_ONE_COLOR ) )
		    {
			int coef;
			if( clip >= s->x )
			{
			    coef = ( (int)( clip - s->x ) * 256 ) / len;
			}
			else 
			{
			    coef = ( (int)( s->x - clip ) * 256 ) / len;
			}
			if( !( g_wm->cur_flags & WBD_FLAG_ONE_COLOR ) )
			{
			    output_v->c = blend( s->c, e->c, coef );
			}
			if( !( g_wm->cur_flags & WBD_FLAG_ONE_TRANSP ) )
			{
			    int t = ( ( 256 - coef ) * s->t + coef * e->t ) / 256;
			    if( t >= 255 ) t = 255;
			    output_v->t = (uchar)t;
			}
		    }
		}
		else 
		{
		    output_v->y = s->y;
		    output_v->c = s->c;
		    output_v->t = s->t;
		}
	    }
	    else 
	    {
		output_v->y = clip;
		int len = e->y - s->y;
		if( len )
		{
		    if( len < 0 ) len = -len;
		    int d = ( (int)( e->x - s->x ) * 4096 ) / len;
		    if( clip >= s->y )
		    {
			d *= ( clip - s->y );
		    }
		    else 
		    {
			d *= ( s->y - clip );
		    }
		    d /= 4096;
		    output_v->x = s->x + d;
		    if( ( g_wm->cur_flags & ( WBD_FLAG_ONE_TRANSP | WBD_FLAG_ONE_COLOR ) ) != ( WBD_FLAG_ONE_TRANSP | WBD_FLAG_ONE_COLOR ) )
		    {
			int coef;
			if( clip >= s->y )
			{
			    coef = ( (int)( clip - s->y ) * 256 ) / len;
			}
			else 
			{
			    coef = ( (int)( s->y - clip ) * 256 ) / len;
			}
			if( !( g_wm->cur_flags & WBD_FLAG_ONE_COLOR ) )
			{
			    output_v->c = blend( s->c, e->c, coef );
			}
			if( !( g_wm->cur_flags & WBD_FLAG_ONE_TRANSP ) )
			{
			    int t = ( ( 256 - coef ) * s->t + coef * e->t ) / 256;
			    if( t >= 255 ) t = 255;
			    output_v->t = (uchar)t;
			}
		    }
		}
		else 
		{
		    output_v->x = s->x;
		    output_v->c = s->c;
		    output_v->t = s->t;
		}
	    }
	    output->vnum++;
	}
	if( add_e )
	{
	    output_v = &output->v[ output->vnum ];
	    output_v->x = e->x;
	    output_v->y = e->y;
	    output_v->c = e->c;
	    output_v->t = e->t;
	    output->vnum++;
	}
	
	s = e;
	s_inside = e_inside;
    }
}

void clip_polygon( 
    sundog_polygon* subject, 
    sundog_polygon* output, 
    sundog_polygon* input, 
    int16 offset_x, int16 offset_y,
    int16 clip_x, int16 clip_y, int16 clip_xsize, int16 clip_ysize )
{
    //Sutherland–Hodgman algorithm:
    output->vnum = subject->vnum;
    mem_copy( output->v, subject->v, output->vnum * sizeof( sundog_vertex ) );
    for( int i = 0; i < output->vnum; i++ )
    {
	output->v[ i ].x += offset_x;
	output->v[ i ].y += offset_y;
    }
    if( g_wm->cur_transparency < 255 )
    {
	if( !( g_wm->cur_flags & WBD_FLAG_ONE_TRANSP ) )
	{
	    for( int i = 0; i < output->vnum; i++ )
	    {
		int t = output->v[ i ].t;
		t *= g_wm->cur_transparency;
		t /= 256;
		output->v[ i ].t = t;
	    }
	}
    }
    polygon_clipping_stage( output, input, clip_x, 0, 1 );
    if( output->vnum == 0 ) return;
    polygon_clipping_stage( output, input, clip_y, 0, 0 );
    if( output->vnum == 0 ) return;
    polygon_clipping_stage( output, input, clip_x + clip_xsize, 1, 1 );
    if( output->vnum == 0 ) return;
    polygon_clipping_stage( output, input, clip_y + clip_ysize, 1, 0 );
    if( output->vnum == 0 ) return;
    if( g_wm->cur_flags & WBD_FLAG_ONE_COLOR )
    {
	output->v[ 0 ].c = subject->v[ 0 ].c;
    }
    if( g_wm->cur_flags & WBD_FLAG_ONE_TRANSP )
    {
	int t = subject->v[ 0 ].t;
	t *= g_wm->cur_transparency;
	t /= 256;
	output->v[ 0 ].t = t;
    }
}

void draw_triangle( sundog_vertex* v1, sundog_vertex* v2, sundog_vertex* v3 )
{
    uchar one_t = v1->t;
    COLOR one_c = v1->c;
    
    //Sort vertices:
    sundog_vertex* temp;
    if( v1->y > v2->y ) { temp = v1; v1 = v2; v2 = temp; }
    if( v1->y > v3->y ) { temp = v1; v1 = v3; v3 = temp; }
    if( v2->y > v3->y ) { temp = v2; v2 = v3; v3 = temp; }
    
    //Check:
    if( v1->y == v3->y ) return;
    bool color;
    bool transp;
    color = ( g_wm->cur_flags & WBD_FLAG_ONE_COLOR ) == 0;
    transp = ( g_wm->cur_flags & WBD_FLAG_ONE_TRANSP ) == 0;

    g_wm->screen_changed++;
    
    int dx1, dx2; //deltas...
    int dt1, dt2;
    int dr1, dg1, db1;
    int dr2, dg2, db2;
    int cx1, cx2; //current values...
    int ct1, ct2;
    int cr1, cg1, cb1;
    int cr2, cg2, cb2;
    
    //Calc deltas:
    int len = ( v2->y - v1->y );
    if( len ) 
    {
	dx1 = ( ( v2->x - v1->x ) << LINE_PREC ) / len;
	if( transp )
	{
	    dt1 = ( ( v2->t - v1->t ) << LINE_PREC ) / len;
	}
	if( color )
	{
	    dr1 = ( ( red( v2->c ) - red( v1->c ) ) << LINE_PREC ) / len;
	    dg1 = ( ( green( v2->c ) - green( v1->c ) ) << LINE_PREC ) / len;
	    db1 = ( ( blue( v2->c ) - blue( v1->c ) ) << LINE_PREC ) / len;
	}
    }
    else
    {
	dx1 = 0;
	dt1 = 0;
	dr1 = 0;
	dg1 = 0;
	db1 = 0;
    }
    len = ( v3->y - v1->y );
    if( len ) 
    {
	dx2 = ( ( v3->x - v1->x ) << LINE_PREC ) / len;
	if( transp )
	{
	    dt2 = ( ( v3->t - v1->t ) << LINE_PREC ) / len;
	}
	if( color )
	{
	    dr2 = ( ( red( v3->c ) - red( v1->c ) ) << LINE_PREC ) / len;
	    dg2 = ( ( green( v3->c ) - green( v1->c ) ) << LINE_PREC ) / len;
	    db2 = ( ( blue( v3->c ) - blue( v1->c ) ) << LINE_PREC ) / len;
	}
    }
    else
    {
	dx2 = 0;
	dt2 = 0;
	dr2 = 0;
	dg2 = 0;
	db2 = 0;
    }
    //Set start values:
    cx1 = v1->x << LINE_PREC;
    cx2 = cx1;
    if( transp )
    {
	ct1 = v1->t << LINE_PREC;
	ct2 = ct1;
    }
    else 
    {
	ct1 = one_t;
    }
    if( color )
    {
	cr1 = red( v1->c ) << LINE_PREC;
	cg1 = green( v1->c ) << LINE_PREC;
	cb1 = blue( v1->c ) << LINE_PREC;
	cr2 = cr1;
	cg2 = cg1;
	cb2 = cb1;
    }
    for( int y = v1->y; y < v2->y; y++ )
    {
	int x1 = cx1 >> LINE_PREC;
	int x2 = cx2 >> LINE_PREC;
	bool xreverse = 0;
	if( x1 > x2 ) { int temp_x = x1; x1 = x2; x2 = temp_x; xreverse = 1; }
	int xsize = x2 - x1;
	if( xsize )
	{
	    COLORPTR ptr = g_wm->g_screen + y * g_wm->screen_xsize + x1;
	    COLORPTR size = ptr + xsize;
	    if( !transp )
	    {
		if( !color )
		{
		    if( ct1 >= 255 )
		    {
			while( ptr < size ) *ptr++ = one_c;
		    }
		    else 
		    {
			while( ptr < size ) { *ptr = blend( *ptr, one_c, ct1 ); *ptr++; }
		    }
		}
		else
		{
		    int rlen = cr2 - cr1;
		    int glen = cg2 - cg1;
		    int blen = cb2 - cb1;
		    if( rlen == 0 && glen == 0 && blen == 0 )
		    {
			COLOR c = get_color( cr1 >> LINE_PREC, cg1 >> LINE_PREC, cb1 >> LINE_PREC );
			if( ct1 >= 255 )
			{
			    while( ptr < size ) *ptr++ = c;
			}
			else 
			{
			    while( ptr < size ) { *ptr = blend( *ptr, c, ct1 ); *ptr++; }
			}
		    }
		    else 
		    {
			int r;
			int g;
			int b;
			int dr = rlen / xsize;
			int dg = glen / xsize;
			int db = blen / xsize;
			if( xreverse )
			{
			    r = cr2;
			    g = cg2;
			    b = cb2;
			    dr = -dr;
			    dg = -dg;
			    db = -db;
			}
			else 
			{
			    r = cr1;
			    g = cg1;
			    b = cb1;
			}
			if( ct1 >= 255 )
			{
			    while( ptr < size ) 
			    { 
				*ptr++ = get_color( r >> LINE_PREC, g >> LINE_PREC, b >> LINE_PREC );
				r += dr;
				g += dg;
				b += db;
			    }
			}
			else 
			{
			    while( ptr < size ) 
			    {
				COLOR c = get_color( r >> LINE_PREC, g >> LINE_PREC, b >> LINE_PREC );
				*ptr = blend( *ptr, c, ct1 ); 
				*ptr++;
				r += dr;
				g += dg;
				b += db;
			    }
			}
		    }
		}
	    }
	    else 
	    {
		if( !color )
		{
		    int t;
		    int dt;
		    if( xreverse )
		    {
			t = ct2;
			dt = ( ct1 - ct2 ) / xsize;
		    }
		    else 
		    {
			t = ct1;
			dt = ( ct2 - ct1 ) / xsize;
		    }
		    while( ptr < size ) 
		    { 
			*ptr = blend( *ptr, one_c, t >> LINE_PREC );
			*ptr++; 
			t += dt;
		    }
		}
		else 
		{
		    int t;
		    int dt;
		    int r;
		    int g;
		    int b;
		    int dr;
		    int dg;
		    int db;
		    if( xreverse )
		    {
			t = ct2;
			dt = ( ct1 - ct2 ) / xsize;
			r = cr2;
			g = cg2;
			b = cb2;
			dr = ( cr1 - cr2 ) / xsize;
			dg = ( cg1 - cg2 ) / xsize;
			db = ( cb1 - cb2 ) / xsize;
		    }
		    else 
		    {
			t = ct1;
			dt = ( ct2 - ct1 ) / xsize;
			r = cr1;
			g = cg1;
			b = cb1;
			dr = ( cr2 - cr1 ) / xsize;
			dg = ( cg2 - cg1 ) / xsize;
			db = ( cb2 - cb1 ) / xsize;
		    }
		    while( ptr < size ) 
		    { 
			COLOR c = get_color( r >> LINE_PREC, g >> LINE_PREC, b >> LINE_PREC );
			*ptr = blend( *ptr, c, t >> LINE_PREC );
			*ptr++; 
			t += dt;
			r += dr;
			g += dg;
			b += db;
		    }
		}
	    }
	}
	cx1 += dx1;
	cx2 += dx2;
	if( transp )
	{
	    ct1 += dt1;
	    ct2 += dt2;
	}
	if( color )
	{
	    cr1 += dr1;
	    cg1 += dg1;
	    cb1 += db1;
	    cr2 += dr2;
	    cg2 += dg2;
	    cb2 += db2;
	}
    }
    //Calc deltas:
    len = ( v3->y - v2->y );
    if( len ) 
    {
	dx1 = ( ( v3->x - v2->x ) << LINE_PREC ) / len;
	if( transp )
	{
	    dt1 = ( ( v3->t - v2->t ) << LINE_PREC ) / len;
	}
	if( color )
	{
	    dr1 = ( ( red( v3->c ) - red( v2->c ) ) << LINE_PREC ) / len;
	    dg1 = ( ( green( v3->c ) - green( v2->c ) ) << LINE_PREC ) / len;
	    db1 = ( ( blue( v3->c ) - blue( v2->c ) ) << LINE_PREC ) / len;
	}
    }
    else 
    {
	dx1 = 0;
	dt1 = 0;
	dr1 = 0;
	dg1 = 0;
	db1 = 0;
    }
    //Set start values:
    cx1 = v2->x << LINE_PREC;
    if( transp )
    {
	ct1 = v2->t << LINE_PREC;
    }
    if( color )
    {
	cr1 = red( v2->c ) << LINE_PREC;
	cg1 = green( v2->c ) << LINE_PREC;
	cb1 = blue( v2->c ) << LINE_PREC;
    }
    for( int y = v2->y; y < v3->y; y++ )
    {
	int x1 = cx1 >> LINE_PREC;
	int x2 = cx2 >> LINE_PREC;
	bool xreverse = 0;
	if( x1 > x2 ) { int temp_x = x1; x1 = x2; x2 = temp_x; xreverse = 1; }
	int xsize = x2 - x1;
	if( xsize )
	{
	    COLORPTR ptr = g_wm->g_screen + y * g_wm->screen_xsize + x1;
	    COLORPTR size = ptr + xsize;
	    if( !transp )
	    {
		if( !color )
		{
		    if( ct1 >= 255 )
		    {
			while( ptr < size ) *ptr++ = one_c;
		    }
		    else 
		    {
			while( ptr < size ) { *ptr = blend( *ptr, one_c, ct1 ); *ptr++; }
		    }
		}
		else
		{
		    int rlen = cr2 - cr1;
		    int glen = cg2 - cg1;
		    int blen = cb2 - cb1;
		    if( rlen == 0 && glen == 0 && blen == 0 )
		    {
			COLOR c = get_color( cr1 >> LINE_PREC, cg1 >> LINE_PREC, cb1 >> LINE_PREC );
			if( ct1 >= 255 )
			{
			    while( ptr < size ) *ptr++ = c;
			}
			else 
			{
			    while( ptr < size ) { *ptr = blend( *ptr, c, ct1 ); *ptr++; }
			}
		    }
		    else 
		    {
			int r;
			int g;
			int b;
			int dr = rlen / xsize;
			int dg = glen / xsize;
			int db = blen / xsize;
			if( xreverse )
			{
			    r = cr2;
			    g = cg2;
			    b = cb2;
			    dr = -dr;
			    dg = -dg;
			    db = -db;
			}
			else 
			{
			    r = cr1;
			    g = cg1;
			    b = cb1;
			}
			if( ct1 >= 255 )
			{
			    while( ptr < size ) 
			    { 
				*ptr++ = get_color( r >> LINE_PREC, g >> LINE_PREC, b >> LINE_PREC );
				r += dr;
				g += dg;
				b += db;
			    }
			}
			else 
			{
			    while( ptr < size ) 
			    {
				COLOR c = get_color( r >> LINE_PREC, g >> LINE_PREC, b >> LINE_PREC );
				*ptr = blend( *ptr, c, ct1 ); 
				*ptr++;
				r += dr;
				g += dg;
				b += db;
			    }
			}
		    }
		}
	    }
	    else 
	    {
		if( !color )
		{
		    int t;
		    int dt;
		    if( xreverse )
		    {
			t = ct2;
			dt = ( ct1 - ct2 ) / xsize;
		    }
		    else 
		    {
			t = ct1;
			dt = ( ct2 - ct1 ) / xsize;
		    }
		    while( ptr < size ) 
		    { 
			*ptr = blend( *ptr, one_c, t >> LINE_PREC );
			*ptr++; 
			t += dt;
		    }
		}
		else 
		{
		    int t;
		    int dt;
		    int r;
		    int g;
		    int b;
		    int dr;
		    int dg;
		    int db;
		    if( xreverse )
		    {
			t = ct2;
			dt = ( ct1 - ct2 ) / xsize;
			r = cr2;
			g = cg2;
			b = cb2;
			dr = ( cr1 - cr2 ) / xsize;
			dg = ( cg1 - cg2 ) / xsize;
			db = ( cb1 - cb2 ) / xsize;
		    }
		    else 
		    {
			t = ct1;
			dt = ( ct2 - ct1 ) / xsize;
			r = cr1;
			g = cg1;
			b = cb1;
			dr = ( cr2 - cr1 ) / xsize;
			dg = ( cg2 - cg1 ) / xsize;
			db = ( cb2 - cb1 ) / xsize;
		    }
		    while( ptr < size ) 
		    { 
			COLOR c = get_color( r >> LINE_PREC, g >> LINE_PREC, b >> LINE_PREC );
			*ptr = blend( *ptr, c, t >> LINE_PREC );
			*ptr++; 
			t += dt;
			r += dr;
			g += dg;
			b += db;
		    }
		}
	    }
	}
	cx1 += dx1;
	cx2 += dx2;
	if( transp )
	{
	    ct1 += dt1;
	    ct2 += dt2;
	}
	if( color )
	{
	    cr1 += dr1;
	    cg1 += dg1;
	    cb1 += db1;
	    cr2 += dr2;
	    cg2 += dg2;
	    cb2 += db2;
	}
    }
}

sundog_vertex g_vertices1[ 8 ];
sundog_vertex g_vertices2[ 8 ];

void draw_polygon( sundog_polygon* p )
{
    if( g_wm->cur_transparency == 0 ) return;
    
    sundog_polygon input;
    sundog_polygon output;
    input.v = g_vertices1;
    output.v = g_vertices2;
    
    WINDOWPTR win = g_wm->cur_window;
    
    if( win && win->visible && win->reg && win->reg->numRects )
    {
	if( win->reg->numRects )
	{
#if defined(OPENGL) && !defined(FRAMEBUFFER)
	    for( int r = 0; r < win->reg->numRects; r++ )
	    {
		int rx1 = win->reg->rects[ r ].left;
		int rx2 = win->reg->rects[ r ].right;
		int ry1 = win->reg->rects[ r ].top;
		int ry2 = win->reg->rects[ r ].bottom;
		
		clip_polygon( p, &output, &input, win->screen_x, win->screen_y, rx1, ry1, rx2 - rx1, ry2 - ry1 );
		
		if( output.vnum >= 3 )
		{
		    if( output.vnum > 8 ) dprint( "ERROR: %d vertices!\n", output.vnum );
		    device_draw_polygon( &output );
		    g_wm->screen_changed++;
		}
	    }
#else
	    clip_polygon( p, &output, &input, win->screen_x, win->screen_y, 0, 0, g_wm->screen_xsize, g_wm->screen_ysize );
	    if( output.vnum >= 3 )
	    {
		if( output.vnum > 8 ) dprint( "ERROR: %d vertices!\n", output.vnum );
		for( int i = 2; i < output.vnum; i++ )
		{
		    draw_triangle( &output.v[ 0 ], &output.v[ i - 1 ], &output.v[ i ] );
		}
	    }
#endif
	}
    }    
}

int draw_char( UTF32_CHAR c, int x, int y, window_manager* wm )
{
    if( c < 0x10 ) return 0; //Control symbols

    WINDOWPTR win = wm->cur_window;

    //Convert from UTF32 to WIN1251:
    utf32_to_win1251( c, c );

    uchar* font;
    int char_xsize;
    int char_ysize;
    int font_num = win->font;
    if( font_num == 0 )
    {
        font = g_font0;
	char_xsize = 8;
	char_ysize = 8;
    }
    else
    {
        font = g_font1;
	char_xsize = font[ 1 + c ];
	char_ysize = font[ 0 ];
    }
    
    int char_xsize_zoom = ( char_xsize * wm->cur_font_zoom ) / 256;
    int char_ysize_zoom = ( char_ysize * wm->cur_font_zoom ) / 256;

    if( c == ' ' ) return char_xsize_zoom;

#if defined(OPENGL) && !defined(FRAMEBUFFER)
    sundog_image* font_img = g_font_img[ font_num ];
    int cxsize = g_font_cxsize[ font_num ];
    int cysize = g_font_cysize[ font_num ];

    unsigned int transp = wm->cur_transparency;
    if( transp > 255 ) transp = 255;
    font_img->transp = transp;
    font_img->color = wm->cur_font_color;

    if( win && win->visible && win->reg && win->reg->numRects )
    {
	float source_x = ( c & 15 ) * cxsize;
	float source_y = ( c / 16 ) * cysize;
	float source_xsize = char_xsize;
	float source_ysize = char_ysize;
	int dest_xsize = char_xsize_zoom;
	int dest_ysize = char_ysize_zoom;
	float pixel_xsize = source_xsize / (float)dest_xsize;
	float pixel_ysize = source_ysize / (float)dest_ysize;
	
	x += win->screen_x;
	y += win->screen_y;
	if( win->reg->numRects )
	{
	    for( int r = 0; r < win->reg->numRects; r++ )
	    {
		int rx1 = win->reg->rects[ r ].left;
		int rx2 = win->reg->rects[ r ].right;
		int ry1 = win->reg->rects[ r ].top;
		int ry2 = win->reg->rects[ r ].bottom;

		//Control box size:
		float src_xoff = 0;
		float src_yoff = 0;
		float src_xs = source_xsize;
		float src_ys = source_ysize;
		int nx = x;
		int ny = y;
		int nxsize = dest_xsize;
		int nysize = dest_ysize;
		if( nx < rx1 ) 
		{ 
		    //Destination:
		    int sub = ( rx1 - nx );
		    nxsize -= sub;
		    nx = rx1; 
		    //Source:
		    src_xoff = (float)sub * pixel_xsize;
		}
		if( ny < ry1 ) 
		{ 
		    //Destination:
		    int sub = ( ry1 - ny ); 
		    nysize -= sub;
		    ny = ry1; 
		    //Source:
		    src_yoff = (float)sub * pixel_ysize; 
		}
		if( nx + nxsize <= rx1 ) continue;
		if( ny + nysize <= ry1 ) continue;
		if( nx + nxsize > rx2 ) 
		{
		    //Destination:
		    int sub = nx + nxsize - rx2;
		    nxsize -= sub;
		    //Source:
		    src_xs -= (float)sub * pixel_xsize;
		}
		if( ny + nysize > ry2 ) 
		{
		    //Destination:
		    int sub = ny + nysize - ry2;
		    nysize -= sub;
		    //Source:
		    src_ys -= (float)sub * pixel_ysize;
		}
		if( nx >= rx2 ) continue;
		if( ny >= ry2 ) continue;
		if( nxsize < 0 ) continue;
		if( nysize < 0 ) continue;
        	
		//Draw it:
		device_draw_bitmap_scaled( 
		    nx, ny, 
		    nxsize, nysize, 
		    source_x + src_xoff, source_y + src_yoff,
		    src_xs - src_xoff, src_ys - src_yoff,
		    font_img );
		g_wm->screen_changed++;
	    }
	}
    }
#else

    x += win->screen_x;
    y += win->screen_y;

    //Bounds control:
    if( x < 0 ) return char_xsize_zoom;
    if( y < 0 ) return char_xsize_zoom;
    if( x > wm->screen_xsize - char_xsize_zoom ) return char_xsize_zoom;
    if( x >= win->screen_x + win->xsize ) return char_xsize_zoom;

    g_wm->screen_changed++;

    int transp = wm->cur_transparency;
    int screen_xsize = wm->screen_xsize;
    int screen_ysize = wm->screen_ysize;
    COLORPTR g_screen = wm->g_screen;
    COLOR cur_font_color = wm->cur_font_color;

    int p = c * char_ysize;
    int sp = y * screen_xsize + x;
    
    font += 257;

    if( transp >= 255 )
    {
	if( wm->cur_font_zoom < 512 )
	{
	    //y clipping
	    if( y + char_ysize > screen_ysize ) 
	    {
		char_ysize -= ( y + char_ysize ) - screen_ysize;
		if( char_ysize <= 0 ) return char_xsize_zoom;
	    }
	    int add = screen_xsize - char_xsize;
	    for( int cy = 0; cy < char_ysize; cy++ )
	    {
		int fpart = font[ p ];
		for( int cx = 0; cx < char_xsize; cx++ )
		{
		    if( fpart & 128 ) 
			g_screen[ sp ] = cur_font_color;
		    fpart <<= 1;
		    sp++;
		}
		sp += add;
		p++;
	    }
	}
	else
	{
	    //y clipping
	    if( y + char_ysize * 2 > screen_ysize ) 
	    {
		char_ysize -= ( ( y + char_ysize * 2 ) - screen_ysize ) / 2;
		if( char_ysize <= 0 ) return char_xsize_zoom;
	    }
	    int add = screen_xsize * 2 - char_xsize * 2;
	    for( int cy = 0; cy < char_ysize; cy++ )
	    {
		int fpart = font[ p ];
		for( int cx = 0; cx < char_xsize; cx++ )
		{
		    if( fpart & 128 ) 
		    {
			g_screen[ sp ] = cur_font_color;
			g_screen[ sp+1 ] = cur_font_color;
			g_screen[ sp+screen_xsize ] = cur_font_color;
			g_screen[ sp+screen_xsize+1 ] = cur_font_color;
		    }
		    fpart <<= 1;
		    sp += 2;
		}
		sp += add;
		p++;
	    }
	}
    }
    else
    {
	if( wm->cur_font_zoom < 512 )
	{
	    //y clipping
	    if( y + char_ysize > screen_ysize ) 
	    {
		char_ysize -= ( y + char_ysize ) - screen_ysize;
		if( char_ysize <= 0 ) return char_xsize_zoom;
	    }
	    int add = screen_xsize - char_xsize;
	    for( int cy = 0; cy < char_ysize; cy++ )
	    {
		int fpart = font[ p ];
		for( int cx = 0; cx < char_xsize; cx++ )
		{
		    if( fpart & 128 ) 
			g_screen[ sp ] = blend( g_screen[ sp ], cur_font_color, transp );
		    fpart <<= 1;
		    sp++;
		}
		sp += add;
		p++;
	    }
	}
	else
	{
	    //y clipping
	    if( y + char_ysize * 2 > screen_ysize ) 
	    {
		char_ysize -= ( ( y + char_ysize * 2 ) - screen_ysize ) / 2;
		if( char_ysize <= 0 ) return char_xsize_zoom;
	    }
	    int add = screen_xsize * 2 - char_xsize * 2;
	    for( int cy = 0; cy < char_ysize; cy++ )
	    {
		int fpart = font[ p ];
		for( int cx = 0; cx < char_xsize; cx++ )
		{
		    if( fpart & 128 ) 
		    {
			g_screen[ sp ] = blend( g_screen[ sp ], cur_font_color, transp );
			g_screen[ sp+1 ] = blend( g_screen[ sp+1 ], cur_font_color, transp );
			g_screen[ sp+screen_xsize ] = blend( g_screen[ sp+screen_xsize ], cur_font_color, transp );
			g_screen[ sp+screen_xsize+1 ] = blend( g_screen[ sp+screen_xsize+1 ], cur_font_color, transp );
		    }
		    fpart <<= 1;
		    sp += 2;
		}
		sp += add;
		p++;
	    }
	}
    }
#endif
    return char_xsize_zoom;
}

void draw_string( const UTF8_CHAR* str, int x, int y, window_manager* wm )
{
    //Bounds control:
    if( y < -wbd_char_y_size( wm ) ) return;
    if( x >= wm->cur_window->xsize ) return;

    if( wm->cur_transparency == 0 ) return;

    int start_x = x;

    while( *str )
    {
	if( *str == 0xA ) 
	{ 
	    y += wbd_char_y_size( wm ); 
	    x = start_x; 
	    str++;
	}
	else
	{
	    UTF32_CHAR c32;
	    str += utf8_to_utf32_char( str, &c32 );
	    draw_char( c32, x, y, wm );
	    x += wbd_char_x_size( c32, wm );
	}
    }
}

void draw_string_limited( const UTF8_CHAR* str, int x, int y, int limit, window_manager* wm )
{
    //Bounds control:
    if( y < -wbd_char_y_size( wm ) ) return;
    if( x >= wm->cur_window->xsize ) return;
    
    if( wm->cur_transparency == 0 ) return;

    int start_x = x;
    int ptr = 0;
    
    while( *str && ptr < limit )
    {
	if( *str == 0xA ) 
	{ 
	    y += wbd_char_y_size( wm ); 
	    x = start_x; 
	    str++;
	    ptr++;
	}
	else
	{
	    UTF32_CHAR c32;
	    int add = utf8_to_utf32_char( str, &c32 );
	    str += add;
	    ptr += add;
	    draw_char( c32, x, y, wm );
	    x += wbd_char_x_size( c32, wm );
	}
    }
}

void draw_pixel( int x, int y, COLOR color, window_manager *wm )
{
    if( wm->cur_transparency == 0 ) return;

#if defined(OPENGL) && !defined(FRAMEBUFFER)
    draw_box( x, y, 1, 1, color, wm );
#else

    x += wm->cur_window->screen_x;
    y += wm->cur_window->screen_y;

    if( x >= 0 && x < wm->screen_xsize &&
	y >= 0 && y < wm->screen_ysize )
    {
	int ptr = y * wm->screen_xsize + x;
	if( wm->cur_transparency >= 255 )
	    wm->g_screen[ ptr ] = color;
	else
	    wm->g_screen[ ptr ] = blend( wm->g_screen[ ptr ], color, wm->cur_transparency );
	g_wm->screen_changed++;
    }
#endif
}

void draw_image_scaled( 
    int src_x, int src_y, //fixed point
    int src_xsize, int src_ysize, //fixed point
    int dest_x, int dest_y, 
    int dest_xsize, int dest_ysize, 
    sundog_image* img, 
    window_manager* wm )
{
    if( img == 0 || src_xsize <= 0 || src_ysize <= 0 || dest_xsize <= 0 || dest_ysize <= 0 ) return;

    if( wm->cur_transparency == 0 ) return;

    WINDOWPTR win = wm->cur_window;

    unsigned int transp = wm->cur_transparency;
    if( transp > 255 ) transp = 255;
    img->transp = transp;

#if defined(OPENGL) && !defined(FRAMEBUFFER)
    if( win && win->visible && win->reg && win->reg->numRects )
    {
	float f_src_xsize = (float)src_xsize / (float)( (int)1 << IMG_PREC );
	float f_src_ysize = (float)src_ysize / (float)( (int)1 << IMG_PREC );
	float f_src_x = (float)src_x / (float)( (int)1 << IMG_PREC );
	float f_src_y = (float)src_y / (float)( (int)1 << IMG_PREC );
	float pixel_xsize = f_src_xsize / (float)dest_xsize;
	float pixel_ysize = f_src_ysize / (float)dest_ysize;

	dest_x += win->screen_x;
	dest_y += win->screen_y;
	if( win->reg->numRects )
	{
	    for( int r = 0; r < win->reg->numRects; r++ )
	    {
		int rx1 = win->reg->rects[ r ].left;
		int rx2 = win->reg->rects[ r ].right;
		int ry1 = win->reg->rects[ r ].top;
		int ry2 = win->reg->rects[ r ].bottom;

		//Control box size:
		float src_x2 = f_src_x;
		float src_y2 = f_src_y;
		float src_xsize2 = f_src_xsize;
		float src_ysize2 = f_src_ysize;
		int nx = dest_x;
		int ny = dest_y;
		int nxsize = dest_xsize;
		int nysize = dest_ysize;
		if( nx < rx1 ) 
		{ 
		    //Destination:
		    int sub = ( rx1 - nx );
		    nxsize -= sub;
		    nx = rx1; 
		    //Source:
		    float fsub = (float)sub * pixel_xsize;
		    src_x2 += fsub;
		    src_xsize2 -= fsub;
		}
		if( ny < ry1 ) 
		{ 
		    //Destination:
		    int sub = ( ry1 - ny ); 
		    nysize -= sub;
		    ny = ry1; 
		    //Source:
		    float fsub = (float)sub * pixel_xsize;
		    src_y2 += fsub;
		    src_ysize2 -= fsub;
		}
		if( nx + nxsize <= rx1 ) continue;
		if( ny + nysize <= ry1 ) continue;
		if( nx + nxsize > rx2 ) 
		{
		    //Destination:
		    int sub = nx + nxsize - rx2;
		    nxsize -= sub;
		    //Source:
		    src_xsize2 -= (float)sub * pixel_xsize;
		}
		if( ny + nysize > ry2 ) 
		{
		    //Destination:
		    int sub = ny + nysize - ry2;
		    nysize -= sub;
		    //Source:
		    src_ysize2 -= (float)sub * pixel_ysize;
		}
		if( nx >= rx2 ) continue;
		if( ny >= ry2 ) continue;
		if( nxsize < 0 ) continue;
		if( nysize < 0 ) continue;
        	
		//Draw it:
		device_draw_bitmap_scaled( 
		    nx, ny, 
		    nxsize, nysize, 
		    src_x2, src_y2,
		    src_xsize2, src_ysize2,
		    img );
		g_wm->screen_changed++;
	    }
	}
    }
#else

    //Calc deltas:
    int dx = src_xsize / dest_xsize;
    int dy = src_ysize / dest_ysize;
    
    COLORPTR scr = wm->g_screen;
    int screen_xsize = wm->screen_xsize;
    int screen_ysize = wm->screen_ysize;
    dest_x += win->screen_x;
    dest_y += win->screen_y;

    //Bounds control:
    if( dest_x < 0 ) 
    { 
	dest_xsize += dest_x;
	if( dest_xsize <= 0 ) return;
	src_x += -dest_x * dx;
	dest_x = 0;
    }    
    if( dest_y < 0 ) 
    { 
	dest_ysize += dest_y;
	if( dest_ysize <= 0 ) return;
	src_y += -dest_y * dy;
	dest_y = 0;
    }
    if( dest_x + dest_xsize > screen_xsize )
    {
	dest_xsize -= ( dest_x + dest_xsize ) - screen_xsize;
	if( dest_xsize <= 0 ) return;
    }
    if( dest_y + dest_ysize > screen_ysize )
    {
	dest_ysize -= ( dest_y + dest_ysize ) - screen_ysize;
	if( dest_ysize <= 0 ) return;
    }
    
    g_wm->screen_changed++;

    int dest_ptr = dest_y * screen_xsize + dest_x;
    int add = screen_xsize - dest_xsize;
    if( img->flags & IMAGE_ALPHA8 )
    {
	uchar* img_data = (uchar*)img->data;
	COLOR color = img->color;
	for( int yy = 0; yy < dest_ysize; yy++ )
	{
	    int cx = src_x;
	    int img_ptr = ( src_y >> IMG_PREC ) * img->xsize;
	    if( transp == 255 )
	    {
		for( int xx = 0; xx < dest_xsize; xx++ )
		{
		    uchar v = img_data[ img_ptr + ( cx >> IMG_PREC ) ];
		    if( v == 255 )
			scr[ dest_ptr ] = color;
		    else if( v > 0 )
			scr[ dest_ptr ] = blend( scr[ dest_ptr ], color, v );
		    dest_ptr++;
		    cx += dx;
		}
	    }
	    else
	    {
		for( int xx = 0; xx < dest_xsize; xx++ )
		{
		    uint v = img_data[ img_ptr + ( cx >> IMG_PREC ) ];
		    v = ( v * transp ) / 256;
		    if( v == 255 )
			scr[ dest_ptr ] = color;
		    else if( v > 0 )
			scr[ dest_ptr ] = blend( scr[ dest_ptr ], color, v );
		    dest_ptr++;
		    cx += dx;
		}
	    }
	    dest_ptr += add;
	    src_y += dy;
	}
    }
    else
    if( img->flags & IMAGE_NATIVE_RGB )
    {
	COLORPTR img_data = (COLORPTR)img->data;
	for( int yy = 0; yy < dest_ysize; yy++ )
	{
	    int cx = src_x;
	    int img_ptr = ( src_y >> IMG_PREC ) * img->xsize;
	    if( transp == 255 )
	    {
		for( int xx = 0; xx < dest_xsize; xx++ )
		{
		    scr[ dest_ptr ] = img_data[ img_ptr + ( cx >> IMG_PREC ) ];
		    dest_ptr++;
		    cx += dx;
		}
	    }
	    else
	    {
		for( int xx = 0; xx < dest_xsize; xx++ )
		{
		    scr[ dest_ptr ] = blend( scr[ dest_ptr ], img_data[ img_ptr + ( cx >> IMG_PREC ) ], transp );
		    dest_ptr++;
		    cx += dx;
		}
	    }
	    dest_ptr += add;
	    src_y += dy;
	}
    }
#endif
}

int wbd_char_x_size( UTF32_CHAR c, window_manager *wm )
{
    if( c < 0x10 ) return 0;
    if( wm->cur_window->font == 0 )
    {
	return ( 8 * wm->cur_font_zoom ) / 256;
    }
    else
    {
	utf32_to_win1251( c, c );
	return ( (int)g_font1[ 1 + c ] * wm->cur_font_zoom ) / 256;
    }
}

int wbd_char_y_size( window_manager *wm )
{
    if( wm->cur_window->font == 0 )
    {
	return ( 8 * wm->cur_font_zoom ) / 256;
    }
    else
    {
	return ( (int)g_font1[ 0 ] * wm->cur_font_zoom ) / 256;
    }
}

int wbd_string_size( const UTF8_CHAR *str, window_manager *wm )
{
    int size = 0;
    while( *str )
    {
	UTF32_CHAR c32;
	str += utf8_to_utf32_char( str, &c32 );
	size += wbd_char_x_size( c32, wm );
    }    
    return size;
}
